
/****************************************************************************
 *  File: main.c
 *  Version: 0.4.6c
 *  Sukuna
 *
 *  Created by No22 <gikoforth@s13.xrea.com> on Sun Jul 20 2003.
 *  Copyright (c) 2003,2004,2005,2006,2007 No22. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "sukuna.h"

void sukuna_usage() {
    printf("Usage: sukuna [switches] [programfile]\n");
    printf("  -h         Print this help message.\n");
    printf("  -q         Quiet mode.\n");
    printf("  -i <file>  Use image <file> instead of 'sukuna.si'.\n");
    printf("  -e 'word'  Evaluate 'word'.\n");
    printf("  -w level   Specify warning level.\n");
    printf("  -o level   Specify optimization level.\n");
    printf("  -m  size   Specify sukuna memory size.\n");
    printf("  -mc size   Specify code size.\n");
    printf("  -md size   Specify data size.\n");
    printf("  -mt size   Specify temporary size.\n");
    printf("  -ms size   Specify stack size.\n");
    printf("  -mr size   Specify return stack size.\n");
    printf("  -mh size   Specify heap size.\n");
}

void sukuna_setup_defalt_global_value() {
	gCodeSize = 98304;
	gDataSize = 196608;
	gStackSize = 8192;
	gRStackSize = 16384;
	gTempSize = 16384;
	gGcLimitSize = 1536000;
    gGcLimitMaxSize = 8192000;
	gQuietFlag = SNIL;
}

CELL* sukuna_copy_args(char** p, int n) {
    int i;
    CELL* tablep;
    CELL* q;
    UWORD* strp;
    
    tablep = (CELL*)sukuna_malloc(bytes_to_cells(sizeof(char*)*n));
    q = tablep;
    for(i=0;i<n;i++) {
        strp = (UWORD*)sukuna_malloc(buf_size+1);
        *((CELL*)strp) = (CELL)to_si(cstr_to_sstr((unsigned char*)(*p++),(UWORD*)strp+cell_size,(int)buf_size));
        *q++ = (CELL)cp_to_sp(strp);
    }
    return tablep;
}

int main(int argc, const char* argv[]) {
    char* defaultImage = "sukuna.si";
    char* sourceFile = SNIL;
    char* commandLine = SNIL;
    char** p = (char**)argv;
    UWORD buf[buf_size];
    CELL n;
    CELL size;
    CELL level = SNIL;
    CELL optlv = SNIL;
    int i = 1;
    char* op;
    char c,d;
    
    sukuna_setup_defalt_global_value();
    p++;
    while(i < argc) {
        op = (char*)argv[i];
        c = *op++;
        if(c=='-') {
            c = *op++;
            switch(c) {
                case 'q':
                    /* quiet mode. */
                    gQuietFlag = STRUE;
                    level = to_si(0);
                    break;
                case 'i':
                    /* load memory image. */
                    defaultImage = (char*)argv[++i];
                    break;
                case 'e':
                    /* evaluate arg. */
                    commandLine = (char*)argv[++i];
                    p++;
                    gQuietFlag = STRUE;
                    level = to_si(0);
                    break;
                case 'w':
                    /* warning level. */
                    d = *op;
                    op = (char*)argv[++i];
                    p++;
                    n = cstr_to_sstr((unsigned char*)op,buf,buf_size);
                    if(sukuna_str_to_num(buf,n,&level)==SNIL) {
                        /* error */
                        sukuna_usage();
                        return 0;
                    }
                    break;
                case 'o':
                    /* optimization level. */
                    d = *op;
                    op = (char*)argv[++i];
                    p++;
                    n = cstr_to_sstr((unsigned char*)op,buf,buf_size);
                    if(sukuna_str_to_num(buf,n,&optlv)==SNIL) {
                        /* error */
                        sukuna_usage();
                        return 0;
                    }
                    break;
                case 'm':
                    /* set memory size. */
                    d = *op;
                    op = (char*)argv[++i];
                    p++;
                    n = cstr_to_sstr((unsigned char*)op,buf,buf_size);
                    if(sukuna_str_to_num(buf,n,&size)==SNIL) {
                        /* error */
                        sukuna_usage();
                        return 0;
                    }
                    size = to_ci(size);
                    size = size * 1024 / cellWidth;
                    switch(d) {
                        case 'c': codeSize = size; break;
                        case 'd': dataSize = size; break;
                        case 't': tempSize = size; break;
                        case 's': stackSize = size; break;
                        case 'r': rStackSize = size; break;
                        case 'h': gGcLimitMaxSize = size; break;
                        default:
                            size = size / 41;
                            codeSize = size * 12;
                            dataSize = size * 24;
                            tempSize = size * 2;
                            stackSize = size;
                            rStackSize = size * 2;
                    }
                    break;
                case 'h':
                default:
                    sukuna_usage();
                    return -1;
                    break;
            }
        }
        else {
            /* load source file. */
            sourceFile = (char*)argv[i];
            break;
        }
        i++;
        p++;
    }
     
    sukuna_msg("sukuna "); sukuna_msg(sukuna_version); sukuna_msg("\n");
    if(defaultImage!=SNIL) {
        sukuna_msg("Loading "); sukuna_msg(defaultImage); sukuna_msg("...\n");
        if(sukuna_loadC((unsigned char*)defaultImage)==SNIL) {
            sukuna_init_memory();
            sukuna_init();
        }
        else {
            sukuna_msg(" ok \n");
        }
    }
    else {
        sukuna_init_memory();
        sukuna_init();
    }

    gc_heapmin = 0;
    gc_heapmax = 0;
    gc_limit = to_si(gGcLimitSize);
    gc_limit_max = to_si(gGcLimitMaxSize);
    
    sVariables->argc.Body = (CELL)to_si(argc-i);
    sVariables->argv.Body = (CELL)cp_to_sp(sukuna_copy_args(p,argc-i));
    sVariables->sstdin.Body = (CELL)cp_to_sp(stdin);
    sVariables->sstdout.Body = (CELL)cp_to_sp(stdout);
    sVariables->sstderr.Body = (CELL)cp_to_sp(stderr);
    
    if(level!=SNIL) sVariables->warnlevel.Body = level;
    if(optlv!=SNIL) sVariables->optlevel.Body  = optlv;

    if((type_stack_base = (CELL*)malloc(buf_size*cellWidth))==NULL) {
        sukuna_cputs("sukuna: initialize error.\n");
        return 0;
    }
    
    failure_count_ptr = type_stack_base;
    retval_stack = type_stack_base+1;
    retval_sp = retval_stack;
    argval_stack = type_stack_base + buf_size - 1;
    argval_sp = argval_stack;
    *retval_stack = 0;
    *argval_stack = 0;
    failure_count = 0;
    
    sukuna_evaluateC((unsigned char*)"autoexec");
    if(sourceFile!=SNIL) {
        sukuna_msg("Loading "); sukuna_msg(sourceFile); sukuna_msg("...\n");
        sukuna_includeC((unsigned char*)sourceFile); 
    }
    sukuna_interpret((unsigned char*)commandLine);
    free(memory);
    free(type_stack_base);
    return 0;
}

