/****************************************************************************
 *  File: sukuna.h
 *  Version: 0.4.6c
 *  Sukuna
 *
 *  Created by No22 <gikoforth@s13.xrea.com> on Sun Jul 20 2003.
 *  Copyright (c) 2003,2004,2005,2006,2007 No22. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/
#ifndef SUKUNA_HEADER
#define SUKUNA_HEADER

#define sukuna_version "v0.4.6c"
#define sukuna_ver_hex (0x0046)
#define buf_size (1024)
#define heap_ext (524288)
#define MAXCFUNC (256)

/*
#define __debug__ 
*/

#define __word32__

#define __gcc__

/********************************/
#ifdef __debug__
    #define dbg(x) x
#else
    #define dbg(x) /* x */
#endif
/********************************/
#define SNIL (0)
#define STRUE (0xFFFFFFF6)
#define SFALSE (0x6)
#define SEOF (0x16)
#define SEOS (0x26)
#define EXACTSELF (0xFFFFFFFA)

#define normal_flag (0x0001)
#define immediate_mask (0x0002)
#define ignore_mask (0x0004)
#define hasLocal_mask (0x0008)
#define localv_mask (0x0010)
#define typed_value (0x0020)
#define mixin_mask (0x0040)
#define refivar_mask (0x0080)
#define refmix_mask (0x0100)
#define class_mask (0x0200)
#define variant_value (0x0400)
#define func_value (0x0800)
#define g_object (0x1000)
#define class_variable (0x2000)
#define direct_callable (0x4000)
#define indirect_callable (0x8000)
#define immtype_flag (0x10000)
#define innerfunc_flag (0x20000)
#define compcurr_flag (0x40000)
#define abstract_flag (0x80000)
#define implicitcall_flag (0x100000)

#define offset_mask (0xFFFF)
#define heap_map_mask (0xFFFFFFE0)

#define BLANK (0x20)

#define classTag (0x736b636d)

/********************************/
#define memTop (0)
#define dataTop (gCodeSize)
#define memSize (gMemSize) 
#define stackTop (gStackTop)
#define rStackTop (gRStackTop)
#define tempTop (gTempTop)
#define memMax (gMemMax)
#define codeSize (gCodeSize)
#define dataSize (gDataSize)
#define stackSize (gStackSize)
#define rStackSize (gRStackSize)
#define tempSize (gTempSize)

/********************************/
#define codep (sukunap->cp)
#define datap (sukunap->dp)
#define tempp (sukunap->tp)
#define byeFlag (sukunap->bye)
#define compile_state (sVariables->state.Body)
#define has_local (sVariables->hasLocal.Body)
#define gc_limit (sVariables->gcLimit.Body)
#define gc_limit_max (sVariables->gcLimitMax.Body)
#define gc_debug (sVariables->gcDebug.Body)
#define class_def (sVariables->classDef.Body)
#define method_def (sVariables->methodDef.Body)
#define echo_flag (sVariables->echoflag.Body)
#define line_number (sVariables->linenum.Body)
#define warn_level (sVariables->warnlevel.Body)
#define opt_level (sVariables->optlevel.Body)

/********************************/
#define push(n) {*(--stackp) = (CELL)(n);}
#define pop (*(stackp++))
#define TOS (topOfStack)
#define SOS (*(stackp))
#define drop { TOS = pop; }
#define r_push(n) {*(--rstackp) = (CELL)(n);}
#define r_pop (*(rstackp++))
#define r_drop {rstackp++;}
#define push_data(n) {*(--(currentTask->sp)) = (CELL)(n); }
#define pop_data (*currentTask->sp++)
#define drop_data (currentTask->sp++)
#define push_retn(n) {*(--(currentTask->rp)) = (CELL)(n); }
#define pop_retn (*currentTask->rp++)
#define drop_retn (currentTask->rp++)

/********************************/
#define av_push(n) {*(--argval_sp) = (CELL)(n);}
#define av_pop (*(argval_sp++))
#define rv_push(n) {*(++retval_sp) = (CELL)(n);}
#define rv_pop (*(retval_sp--))
#define failure_count (*failure_count_ptr)

/********************************/
#define cellWidth (sizeof(CELL))
#define wordWidth (sizeof(WORD))
#define cell_size (sizeof(CELL)/sizeof(WORD))

#define check_si(n) (((n)&1)==1)
#define check_sc(n) (((n)&0xF)==2)
#define check_sb(n) (((n)&0xF)==6)
#define check_sv(n) (((n)&0xF)==10)
#define check_sp(n) (((n)&3)==0)

#define ptrd(n) (check_sp(n) ? wordWidth : 16)

#define to_ci(n) ((n)>>1)
#define to_si(n) (((n)<<1)|1)

#define to_sc(n) (((n)<<4)|2)
#define to_cc(n) ((n)>>4)
#define to_sv(n) (((n)<<4)|10)
#define to_cv(n) ((n)>>4)

#define sc_si_to_ci(n) (check_si(n) ? to_ci(n) : to_cc(n))

#define is_false(n) (check_si(n) ? (n == 1) : (check_sp(n) ? (n == SNIL) : (to_cc(n) == SNIL)))

#ifdef __word32__
    #define si_to_ptr(n) (check_si(n) ? ((n)&(~1))<<1 : (n))
    #define ptr_to_si(n) (((n)>>1)|1)
    #define words_to_cells(n) (n)
#else
    #define si_to_ptr(n) ((n)&(~1))
    #define ptr_to_si(n) ((n)|1)
    #define words_to_cells(n) ( ((n)&1) != 0 ? ((CELL)(n)>>1)+1 : (n)>>1 )  
#endif

#define si_to_num(n) ((n)&(~1))
#define num_to_si(n) ((n)|1)

#define cp_to_sp(n) ((CELL*)((CELL)(n)-memOffset))
#define sp_to_cp(n) ((CELL*)((CELL)(n)+memOffset))
#define check_si_calc(n,m) (((n)&(m)&1)==1)
#define isHeapPtr(n) ( ( (CELL)(n) < (CELL)&memory[0] ) || ( (CELL)&memory[memMax-1] < (CELL)(n) ) )
#define isRstackPtr(n) ( ( ( (CELL)&memory[stackTop] < (CELL)(n) ) && (CELL)(n) < (CELL)&memory[rStackTop] ) )
#define isSpace(n) (((n)==BLANK))
#define to_addr_n(n) (n += cell_size, to_ci(*((CELL*)(n-cell_size))))
#define to_upper(c) ( (((UWORD)'a'<=(c)) && ((c)<=(UWORD)'z')) ? ((c) - ((UWORD)'a' - (UWORD)'A')) : (c) )
#define to_lower(c) ( (((UWORD)'A'<=(c)) && ((c)<=(UWORD)'Z')) ? ((c) + ((UWORD)'a' - (UWORD)'A')) : (c) )


#define bytes_to_cells(n) ( ((n)&3) != 0 ? ((CELL)(n)>>2)+1 : (n)>>2 )
#define to_class(p) ((CELL*)sp_to_cp(*(sp_to_cp(p)-1)))
#define to_class_cp(p) ((sukunaClass*)(to_class(p)-1))
#define to_cclass(p) ( check_si(*(sp_to_cp(p)-1)) ? sp_to_cp(*(sp_to_cp(si_to_ptr(*(sp_to_cp(p)-1))))) : sp_to_cp(*(sp_to_cp(p)-1)) )
#define get_selfp ( check_si(*(objectp-1)) ? sp_to_cp(si_to_ptr(*(objectp-1)))+1 : objectp )
#define XtOffset(p_type,field) ((unsigned int)&(((p_type)0)->field))
#define to_heap_header(p) ((CELL*)((CELL)(p)-XtOffset(sukunaHeapHeader*,Body)))
#define is_valid_heapp(p) (isHeapPtr(p) && ((((sukunaHeapHeader*)to_heap_header(p))->magic)==heap_id))

#define is_warnlevel(lv) ( (to_ci(warn_level)>=(lv)) || (to_ci(warn_level)==-(lv)) )
#define is_optlevel(lv) ( (to_ci(opt_level)>=(lv)) || (to_ci(opt_level)==-(lv)) )

/********************************/
#define STYPE(a,b,c,d) ((a<<24)|(b<<16)|(c<<8)|d)
#define creator_id STYPE('s','k','n','a')
#define header_id STYPE('s','k','h','d')
#define memory_id STYPE('s','k','m','e')
#define end_id STYPE('s','k','e','d')
#define heap_id STYPE('s','k','h','m')
#define class_id STYPE('s','k','c','m')

/********************************/
#define save_registers \
    { \
        push(TOS); \
        currentTask->sp = stackp; \
        currentTask->rp = rstackp; \
        currentTask->op = objectp; \
    }

#define load_registers \
    { \
        stackp = currentTask->sp; \
        rstackp = currentTask->rp; \
        objectp = currentTask->op; \
        TOS = pop; \
    }

#define save_pc \
    { \
        currentTask->pc = p; \
        currentTask->op = objectp; \
        currentTask->lp = localsp; \
        currentTask->s_op = cp_to_sp(objectp); \
    }
#define load_pc \
    { \
        p = currentTask->pc; \
        objectp = currentTask->op; \
        localsp = currentTask->lp; \
    }

/********************************/
#define to_cf(n) ((double)(*((double*)sp_to_cp(n))))
#define fbinop(op) \
    { \
        Scratch = SOS; \
        save_registers; \
        Scratch = (CELL)sukuna_new_float_obj(to_cf(Scratch) op to_cf(TOS)); \
        Scratch = (CELL)cp_to_sp(Scratch+cellWidth); \
        load_registers; \
        drop; \
        TOS = Scratch; \
    }

#define ffunc(fun) \
    { \
        save_registers; \
        Scratch = (CELL)sukuna_new_float_obj(fun); \
        Scratch = (CELL)cp_to_sp(Scratch+cellWidth); \
        load_registers; \
        TOS = Scratch; \
    }
    
/********************************/
#define ibop(op,errstr) \
    { \
        Scratch = pop; \
        if(check_si(Scratch) && check_si(TOS)) { \
            TOS = to_si(to_ci(Scratch) op to_ci(TOS)); \
            token = *p++; \
        } \
        else if(check_sc(Scratch) && check_sc(TOS)) { \
            TOS = to_sc(to_cc(Scratch) op to_cc(TOS)); \
            token = *p++; \
        } \
        else if(check_sp(Scratch) && check_sp(TOS)) { \
            TOS = Scratch op TOS; \
            token = *p++; \
        } \
        else if(check_si(Scratch) || check_si(TOS)) { \
            if(check_sc(Scratch) || check_sc(TOS)) { \
                TOS = to_sc(sc_si_to_ci(Scratch) op sc_si_to_ci(TOS)); \
                token = *p++; \
            } \
            else if(check_sp(Scratch) || check_sp(TOS)) { \
                TOS = si_to_ptr(Scratch) op si_to_ptr(TOS); \
                token = *p++; \
            } \
            else { \
                printf("type mismatch. "); sukuna_dot_hex(Scratch); sukuna_dot_hex(TOS); printf(" %s\n",errstr); \
                TOS = to_si(err_type); token = _abort; \
            } \
        } \
        else { \
            printf("type mismatch. "); sukuna_dot_hex(Scratch); sukuna_dot_hex(TOS); printf(" %s\n",errstr); \
            TOS = to_si(err_type); token = _abort; \
        } \
    }
    

#define plusstore(mem) \
    { \
        if(check_si(*(mem)) && check_si(TOS)) { \
            *(mem) = to_si(to_ci(*(mem)) + to_ci(TOS)); \
            drop; token = *p++; \
        } \
        else if(check_sc(*(mem)) && check_sc(TOS)) { \
            *(mem) = to_sc(to_cc(*(mem)) + to_cc(TOS)); \
            token = *p++; \
        } \
        else if(check_sp(*(mem)) && check_sp(TOS)) { \
            *(mem) = *(mem) + TOS; \
            drop; token = *p++; \
        } \
        else if(check_si(*(mem)) || check_si(TOS)) { \
            if(check_sc(*(mem)) || check_sc(TOS)) { \
                *(mem) = to_sc(sc_si_to_ci(*(mem)) + sc_si_to_ci(TOS)); \
                drop; token = *p++; \
            } \
            if(check_sp(Scratch) || check_sp(TOS)) { \
                *(mem) = si_to_ptr(*(mem)) + si_to_ptr(TOS); \
                drop; token = *p++; \
            } \
            else { \
                printf("type mismatch.\n"); TOS = to_si(err_type); token = _abort; \
            } \
        } \
        else { \
             printf("type mismatch.\n"); TOS = to_si(err_type); token = _abort; \
        } \
    }

/********************************/

#define push_if_primitive_obj \
    { \
        if(check_si(TOS)) { push(TOS); TOS = sVariables->systemInt.Body; } \
        else if(TOS==SNIL) { push(TOS); TOS = sVariables->systemNil.Body; } \
        else if(check_sc(TOS)) { push(TOS); TOS = sVariables->systemChar.Body; } \
        else if(check_sb(TOS)) { push(TOS); TOS = sVariables->systemSymb.Body; } \
    }
    
#define push_if_primitive_obj2 \
    { \
        if(check_si(*stackp)) { push(sVariables->systemInt.Body); } \
        else if(*stackp==SNIL) { push(sVariables->systemNil.Body); } \
        else if(check_sc(*stackp)) { push(sVariables->systemChar.Body); } \
        else if(check_sb(*stackp)) { push(sVariables->systemSymb.Body); } \
    }

/********************************/
#ifdef __gcc__
    #define CASE(inst) LABEL##inst :
    #define SWITCH(inst) goto *primitives[inst];
    #define NEXT \
        if(check_sp(token)) { \
            r_push(cp_to_sp(p)); p = sp_to_cp(*(sp_to_cp(token))); \
            token = *p++; \
            continue; \
        } \
        else { \
            SWITCH(token) \
        }
    #define NEXTINST(inst) \
        goto LABEL##inst
    #define DISPATCH NEXT
    #define ENDDISPATCH /* end dispatch */
#else
    #define NEXT break
    #define CASE(inst) case inst : 
    #define SWITCH(inst) switch(inst)     
    #define DISPATCH \
        if(check_sp(token)) { \
            r_push(cp_to_sp(p)); p = sp_to_cp(*(sp_to_cp(token))); \
            token = *p++; \
            continue; \
        } \
        else { \
            SWITCH(token)
    #define NEXTINST(inst) \
        token = inst; \
        NEXT;
    #define ENDDISPATCH }
#endif                        
/********************************/
#define err_abort 	(-1)
#define err_nilpointer	(-2)
#define err_conditional	(-3)
#define err_executing	(-4)
#define err_pairs	(-5)
#define err_defer	(-6)
#define err_underflow	(-7)
#define err_comment	(-8)
#define err_notfound	(-9)
#define err_allocate	(-10)
#define err_class	(-11) 
#define err_object	(-12) 
#define err_methoddef	(-13) 
#define err_notselector (-14) 
#define err_ivar 	(-15) 
#define err_cantbind	(-16) 
#define err_mixin	(-17) 
#define err_methodsfor	(-18) 
#define err_superslist	(-19) 
#define err_notimplemented	(-20) 
#define err_immutable	(-21) 
#define err_alias	(-22) 
#define err_type	(-23) 
#define err_io	(-24) 

/********************************/
enum sukunaMemArea {
    c_area, d_area, s_area, r_area, t_area, h_area
};

enum sukunaValueType {
    var_type  = 1, 
    func_type = 3,
    prim_type = 5,
    int_type  = 7,
    char_type = 9,
    ptr_type  = 11,
    bool_type = 13,
    symb_type = 15,
    float_type = 17
};

/********************************/
typedef signed long CELL;
typedef unsigned long UCELL;
#ifdef __word32__
    typedef signed long WORD;
    typedef unsigned long UWORD;
#else
    typedef signed short WORD;
    typedef unsigned short UWORD;
#endif
/********************************/
typedef CELL (*CFunc)(void);

/********************************/
typedef struct sukunaObject
{
    CELL    Klass;
    CELL    Body;
} sukunaObject;

typedef struct sukunaDicEntry
{
    CELL*   ListLink;
    CELL*   HashLink; 
    CELL    HashKey;
    CELL    Klass;
    CELL    Xt;
    CELL    Data;
    CELL    mixin;
    CELL    Flags;
    CELL    TypeList;
    CELL    Offset;
    CELL    NameSize;
} sukunaDicEntry;

typedef struct sukunaTypeList
{
    CELL    arguments;
    CELL    returnvalues; 
    CELL    imm_arguments;
    CELL    imm_returnvalues; 
} sukunaTypeList;

typedef struct sukunaDictionary
{
    CELL*   ListLink;
    CELL*   LastEntry;
    CELL*   Hash[16];
} sukunaDictionary;

typedef struct sukunaTask
{
    CELL*   ListLink;
    CELL*   sp;
    CELL*   rp;
    CELL*   op;
    CELL*   lp;
    CELL*   pc;
    WORD*   IN_orig;
    WORD*   IN;
    CELL    cnt;
    FILE*   stream;
    CELL    flag;
    CELL    linenum;
    WORD*   filepath;
    WORD*   current_word;
    CELL    current_cnt;
    WORD*   s_IN;
    CELL*   s_op;    
} sukunaTask;

typedef struct sukunaGlobals
{
    CELL*   cp;
    CELL*   dp;
    CELL*   tp;
    CELL    bye;
} sukunaGlobals;

typedef struct sukunaVariables
{
    sukunaObject	state;
    sukunaObject	hasLocal;
    sukunaObject	gcLimit;
    sukunaObject    gcLimitMax;
    sukunaObject    gcDebug;
    sukunaObject	version;
    sukunaObject	classDef;
    sukunaObject	methodDef;
    sukunaObject	echoflag;
    sukunaObject	linenum;
    sukunaObject	dicEntryC;
    sukunaObject	mthEntryC;
    sukunaObject	ivrEntryC;
    sukunaObject	locEntryC;
    sukunaObject	methodClass;
    sukunaObject	blockClass;
    sukunaObject	stringClass;
    sukunaObject	vectorClass;
    sukunaObject	floatClass;    
    sukunaObject	systemPrim;    
    sukunaObject	systemInt;    
    sukunaObject	systemNil;    
    sukunaObject	systemChar;
    sukunaObject	systemSymb;        
    sukunaObject	sizeOfDouble;    
    sukunaObject	localEnv;    
    sukunaObject	quoteValue;    
    sukunaObject	argc;    
    sukunaObject	argv;    
    sukunaObject	sstdin;    
    sukunaObject	sstdout;    
    sukunaObject	sstderr;
    sukunaObject    warnlevel;    
    sukunaObject    optlevel;    
    sukunaObject    typeinference;
    sukunaObject    defaultvartype;
    sukunaObject    exactself;
} sukunaVariables;

typedef struct sukunaImageHeader
{
    CELL	cSize;
    CELL	dSize;
    CELL	sSize;
    CELL	rSize;
    CELL	tSize;
    sukunaDictionary* sdict;
    sukunaVariables* svar;
    sukunaGlobals* sukunap;
    sukunaTask* currentTask;
    CELL*	cp;
    CELL*	dp;
} sukunaImageHeader;

typedef struct sukunaHeapHeader
{
    CELL*	prev;
    CELL*	next;
    CELL	Size;
    CELL	magic;
    CELL	Klass;
    CELL	Body;
} sukunaHeapHeader;

typedef struct sukunaProcBlock
{
    CELL*	prologue;
    CELL*	body;
    CELL	blkin;	
    CELL*	data;
    CELL*	xt;
    CELL	blkout;
    CELL*	savedlp;
    CELL	lvn;
    CELL*	selfp;
    CELL	locals;
} sukunaProcBlock;

typedef struct sukunaClass
{
    CELL	Klass;
    CELL*	vtbl;
    CELL*	ivtbl;
    CELL*	supers;
    CELL	vmlen;
    CELL	ivlen;
    CELL	flag;
    CELL	tag;
    CELL*	ListLink;
    CELL*	LastEntry;
    CELL*	Hash[16];
    CELL*	ivarEntry;
    CELL*	iHash[16];
    CELL*   Name;
} sukunaClass;

typedef struct sukunaMixin
{
    sukunaClass*	Klass;
    UCELL		offset;
} sukunaMixin;

typedef struct sukunaFloat
{
    sukunaClass*	Klass;
    double		Body;
} sukunaFloat;

/***********************************************************************************
* Global Variables
***********************************************************************************/
#ifdef SUKUNA_GLOBAL_DEFINE
    #define GLOBAL
#else
    #define GLOBAL extern
#endif

GLOBAL CELL* memory;
GLOBAL CELL  memOffset;

GLOBAL CELL gCodeSize;
GLOBAL CELL gDataSize;
GLOBAL CELL gStackSize;
GLOBAL CELL gRStackSize;
GLOBAL CELL gTempSize;
GLOBAL CELL gGcLimitSize;
GLOBAL CELL gGcLimitMaxSize;
GLOBAL CELL gQuietFlag;

GLOBAL CELL  sTrueObject;
GLOBAL CELL  sFalseObject;

GLOBAL CELL gMemSize;
GLOBAL CELL gStackTop;
GLOBAL CELL gRStackTop;
GLOBAL CELL gTempTop;
GLOBAL CELL gMemMax;

GLOBAL sukunaDictionary* sdict;
GLOBAL sukunaDictionary* ldict;

GLOBAL sukunaVariables* sVariables;

GLOBAL sukunaTask* currentTask;
GLOBAL sukunaTask* topTask;
GLOBAL sukunaGlobals* sukunap;
GLOBAL sukunaHeapHeader* lastHeap;

GLOBAL UCELL totalHeapSize;

GLOBAL CELL* type_stack_base;

GLOBAL CELL* argval_stack;
GLOBAL CELL* argval_sp;

GLOBAL CELL* retval_stack;
GLOBAL CELL* retval_sp;

GLOBAL CELL* failure_count_ptr;

GLOBAL UCELL gc_heapmin;
GLOBAL UCELL gc_heapmax;

/***********************************************************************************
* Sukuna Primitives
***********************************************************************************/
#define DEFINST(inst,wname,flags)  DUMMY##inst, inst, 
enum sukunaPrimitives {
    #include "primitives.h"
    DEFINST(numPrimitives,SNIL,SNIL)
};
#undef DEFINST

/***********************************************************************************
* prototype
***********************************************************************************/
GLOBAL int cstr_to_sstr(unsigned char* inputs,UWORD* p,int n);
GLOBAL int sstr_to_cstr(UWORD* p,unsigned char* inputs,int n);
GLOBAL WORD* sukuna_fgets(WORD* p, CELL n, FILE* fp);
GLOBAL WORD* sukuna_gets(WORD* p, CELL n);
GLOBAL void sukuna_puts(WORD* p);
GLOBAL void sukuna_type(WORD* p, CELL n);
GLOBAL void sukuna_cputs(char* p);
GLOBAL void sukuna_msg(char* p);
GLOBAL void sukuna_cr();
GLOBAL void sukuna_dot(CELL n);
GLOBAL void sukuna_dot_hex(CELL n);
GLOBAL void sukuna_emit(UWORD n);
GLOBAL void sukuna_memdump(CELL* top, CELL length);
GLOBAL int check_ptr(CELL* n);
GLOBAL void sukuna_debug_nemit(CELL n);
GLOBAL CELL* sukuna_allot(CELL** p, CELL n, int type);
GLOBAL CELL* sukuna_callot(CELL n);
GLOBAL CELL* sukuna_dallot(CELL n);
GLOBAL CELL* sukuna_tallot(CELL n);
GLOBAL void sukuna_forget(CELL* p);
GLOBAL CELL* sukuna_gallot(CELL n, CELL t);
GLOBAL CELL sukuna_heap_blocks ();
GLOBAL CELL sukuna_debug_marked ();
GLOBAL void sukuna_free(CELL* p);
GLOBAL CELL sukuna_check_tohp(sukunaHeapHeader* p, CELL n);
GLOBAL void sukuna_gc_root_check(CELL* st, CELL* ed, UCELL* gc_map);
GLOBAL void sukuna_gc_mark(UCELL* gc_map);
GLOBAL void sukuna_gc_sweep();
GLOBAL void sukuna_gc_finalize();
GLOBAL CELL* sukuna_malloc(CELL size);
GLOBAL void sukuna_memcopy(CELL* src, CELL* dst, CELL len);
GLOBAL void sukuna_memmove(WORD* src, WORD* dst, CELL len);
GLOBAL void sukuna_strcopy(WORD* addr, CELL n, WORD* p);
GLOBAL void sukuna_upper(WORD* p, CELL n, WORD* token);
GLOBAL void sukuna_lower(WORD* p, CELL n, WORD* token);
GLOBAL CELL sukuna_sstr_compare(UWORD* addr, CELL n, UWORD* addr2, CELL n2);
GLOBAL CELL sukuna_sstr_comparei(UWORD* addr, CELL n, UWORD* addr2, CELL n2);
GLOBAL WORD* sukuna_strcompare(WORD* addr, CELL n, WORD* addr2, CELL n2);
GLOBAL CELL sukuna_digit_to_number( UWORD c, CELL base );
GLOBAL CELL sukuna_str_to_num(UWORD* addr, CELL n, CELL* res);
GLOBAL CELL sukuna_num_to_str(UWORD* addr, CELL n, UWORD* fma, CELL flen, CELL x);
GLOBAL CELL sukuna_skip(WORD* p, CELL n, WORD c, WORD** newp);
GLOBAL CELL sukuna_skipw(WORD c);
GLOBAL CELL sukuna_scan(WORD* p, CELL n, WORD c, WORD** newp);
GLOBAL CELL sukuna_scanw(WORD c);
GLOBAL CELL sukuna_scanww(WORD c, WORD d);
GLOBAL CELL sukuna_next_word(WORD** p, CELL* n, WORD** newp);
GLOBAL CELL sukuna_get_next_word(WORD** addr);
GLOBAL CELL sukuna_get_next_wordw(WORD** addr);
GLOBAL UCELL sukuna_hash(WORD* p, CELL n);
GLOBAL UCELL sukuna_mhash(UCELL hash);
GLOBAL sukunaDictionary* sukuna_create_dictionary(CELL dicArea);
GLOBAL void sukuna_create_dic_entry(CELL xt, WORD* addr, CELL n, CELL flags, CELL data, CELL mixin, CELL dicArea, sukunaDictionary* dicp);
GLOBAL void sukuna_create_dic_entryd(CELL xt, WORD* addr, CELL n, CELL flags);
GLOBAL void sukuna_create_dic_entryC(CELL xt, char* addr, CELL Flags);
GLOBAL void sukuna_create_dic_entrydw(CELL xt, WORD* addr, CELL n);
GLOBAL sukunaDicEntry* sukuna_find(WORD* addr, CELL n, sukunaDictionary* dicp);
GLOBAL sukunaDicEntry* sukuna_findC(unsigned char* addr);
GLOBAL CELL sukuna_findw(WORD* addr, CELL n, sukunaDictionary* dicp);
GLOBAL void sukuna_create_new_header(WORD* addr, CELL n);
GLOBAL void sukuna_define_new_word(WORD* addr, CELL n);
GLOBAL sukunaDicEntry* sukuna_local_tickp();
GLOBAL CELL sukuna_compile_lstore(CELL token);
GLOBAL void sukuna_finish_def();
GLOBAL void sukuna_rcb();
GLOBAL void sukuna_rcbw();
GLOBAL void sukuna_create_new_label(WORD* addr, CELL n);
GLOBAL void sukuna_create_new_labelC(char* addr, CELL* p);
GLOBAL void sukuna_createp(WORD* addr, CELL n);
GLOBAL void sukuna_create();
GLOBAL CELL sukuna_to_xt(CELL xt);
GLOBAL CELL* sukuna_generate_block(CELL* xt);
GLOBAL sukunaClass* sukuna_create_class_header(CELL dicArea);
GLOBAL void sukuna_inherit_class(sukunaClass* theClass, sukunaClass* supcl);
GLOBAL void sukuna_create_method_entry(CELL xt, CELL selector, CELL flags, CELL data, CELL mixin, CELL dicArea, sukunaClass* theClass);
GLOBAL void sukuna_create_method_entryw(sukunaClass* theClass, CELL xt, CELL selector,  CELL offset);
GLOBAL void sukuna_create_ivar_entry(CELL xt, WORD* addr, CELL n, CELL flags, CELL data, CELL mixin, CELL dicArea, sukunaClass* theClass);
GLOBAL void sukuna_create_ivar_entryw(sukunaClass* theClass, sukunaClass* ivarClass, WORD* addr, CELL n,  CELL offset);
GLOBAL void sukuna_setup_vtbl(sukunaDicEntry* currentMethod, CELL* newVtbl);
GLOBAL void sukuna_create_vtbl(sukunaClass* theClass, CELL area);
GLOBAL void sukuna_create_vtblw(sukunaClass* theClass);
GLOBAL void sukuna_update_vtbl(sukunaClass* theClass);
GLOBAL sukunaDicEntry* sukuna_method_find(CELL selector, sukunaClass* theClass);
GLOBAL sukunaDicEntry* sukuna_mixin_method_find(CELL selector, sukunaClass* theClass);
GLOBAL sukunaDicEntry* sukuna_find_method(CELL selector, sukunaClass* theClass);
GLOBAL sukunaDicEntry* sukuna_find_methodw(CELL selector, sukunaClass* theClass);
GLOBAL sukunaDicEntry* sukuna_find_self_methodw(CELL selector, sukunaClass* theClass);
GLOBAL sukunaDicEntry* sukuna_find_super_method(CELL selector, sukunaClass* theClass);
GLOBAL sukunaDicEntry* sukuna_find_super_methodw(CELL selector, sukunaClass* theClass);
GLOBAL CELL* sukuna_mixin_offset_find(sukunaClass* theMixin,sukunaClass* theClass);
GLOBAL CELL* sukuna_find_mixin_offset(sukunaClass* theMixin,sukunaClass* theClass);
GLOBAL CELL* sukuna_mixin_super_find(sukunaClass* theMixin,sukunaClass* theClass, CELL obp, CELL selfp, CELL selID, CELL* vofs);
GLOBAL CELL* sukuna_find_mixin_super(sukunaClass* theMixin,sukunaClass* theClass, CELL obp, CELL selfp, CELL selID, CELL* vofs);
GLOBAL CELL sukuna_mixin_offset(sukunaDicEntry* theMethod, sukunaClass* theClass, CELL flag);
GLOBAL sukunaDicEntry* sukuna_ivar_find(WORD* addr, CELL n, sukunaClass* theClass);
GLOBAL sukunaDicEntry* sukuna_mixin_ivar_find(WORD* addr, CELL n, sukunaClass* theClass);
GLOBAL sukunaDicEntry* sukuna_find_ivar(WORD* addr, CELL n, sukunaClass* theClass);
GLOBAL sukunaDicEntry* sukuna_find_ivarw(WORD* addr, CELL n, sukunaClass* theClass);
GLOBAL sukunaDicEntry* sukuna_find_ivari(WORD* addr, CELL n, CELL* count);
GLOBAL void sukuna_compile_ivar(sukunaDicEntry* lfa);
GLOBAL void sukuna_compile_ivar_mixref(sukunaDicEntry* lfa);
GLOBAL void sukuna_compile_class_variable(sukunaDicEntry* lfa);
GLOBAL void sukuna_compile_class_refvar(sukunaDicEntry* lfa);
GLOBAL void sukuna_compile_class_mixref(sukunaDicEntry* lfa);
GLOBAL void sukuna_compile_ivarandref(sukunaDicEntry* lfa);
GLOBAL void sukuna_create_selectors(WORD* addr, CELL n);
GLOBAL void sukuna_create_selectorw();
GLOBAL void sukuna_create_local_dicEntry(CELL xt, WORD* addr, CELL n);
GLOBAL void sukuna_compile_localref(sukunaDicEntry* lfa);
GLOBAL void sukuna_define_locals();
GLOBAL void sukuna_lcb();
GLOBAL CELL sukuna_to_receiver(sukunaDicEntry* mtp, CELL* clp, CELL* obp);
GLOBAL sukunaTask* sukuna_push_current_stream(FILE* newStream);
GLOBAL sukunaTask* sukuna_pop_current_stream();
GLOBAL CELL sukuna_includeC(unsigned char* fileName);
GLOBAL CELL sukuna_include(WORD* addr, CELL n);
GLOBAL CELL sukuna_write_cell(FILE* fp, CELL n);
GLOBAL CELL sukuna_read_cell(FILE* fp, CELL* p);
GLOBAL CELL sukuna_write_chunk(FILE* fp, CELL id, CELL* data, CELL size);
GLOBAL CELL sukuna_saveC(unsigned char* fileName);
GLOBAL CELL sukuna_save(WORD* addr, CELL n);
GLOBAL CELL sukuna_loadC(unsigned char* fileName);
GLOBAL void sukuna_reset_stack();
GLOBAL void sukuna_close_streams();
GLOBAL void sukuna_reset_task();
GLOBAL void sukuna_abort(CELL n);
GLOBAL void sukuna_evaluate(WORD* addr, CELL n);
GLOBAL void sukuna_evaluateC(unsigned char* str);
GLOBAL void sukuna_VM(CELL xt);
GLOBAL void sukuna_init_code_area();
GLOBAL void sukuna_build_dictionary();
GLOBAL void sukuna_init_data_area();
GLOBAL void sukuna_init_globals();
GLOBAL CELL sukuna_init_memory();
GLOBAL void sukuna_init();
GLOBAL void sukuna_interpret(unsigned char* commandLine);
GLOBAL CELL sukuna_find_super_class(CELL class_sp, sukunaClass* theClass);
GLOBAL void sukuna_reset_type_stack();
GLOBAL void sukuna_literal_type_inference(CELL tp);
GLOBAL void sukuna_type_inference_check(sukunaDicEntry* pde);
GLOBAL void sukuna_compile_word(sukunaDicEntry* lfa);
GLOBAL void sukuna_compile_word(sukunaDicEntry* lfa);

/***********************************************************************************
* Sukuna Cfuncs
***********************************************************************************/
#define DEFCFUNC(inst,wname,flags,fname) GLOBAL CELL fname(void); 
#include "cfuncs.h"
#undef DEFCFUNC

#define DEFCFUNC(inst,wname,flags,fname) inst, 
enum sukunaCfuncs {
    #include "cfuncs.h"
    DEFCFUNC(numCfuncs,SNIL,SNIL,SNIL)
};
#undef DEFCFUNC

#ifdef SUKUNA_GLOBAL_DEFINE
    #define DEFCFUNC(inst,wname,flags,fname) fname, 
    GLOBAL void* gCustomFunctions[] = {
        #include "cfuncs.h"
        NULL
    };
    #undef DEFCFUNC
#else
    GLOBAL void* gCustomFunctions[];
#endif

#undef GLOBAL
#endif
