/****************************************************************************
 *  File: sukuna.c
 *  Version: 0.4.6c
 *  Sukuna 
 *
 *  Created by No22 <gikoforth@s13.xrea.com> on Mon Sep 08 2003.
 *  Copyright (c) 2003,2004,2005,2006,2007 No22. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>

#define SUKUNA_GLOBAL_DEFINE
#include "sukuna.h"

/***********************************************************************************
* console IO
***********************************************************************************/
unsigned char* skn_read_utf8c(unsigned char* p, UWORD* d) {
    if ((*p & 0x80) != 0x80) {
        *d =  (UWORD)(*p++);
    } 
    else if ((*p & 0xE0) == 0xC0) {
        *d = (UWORD)(*p++ & 0x1F);
        *d = (*d << 6) | (*p++ & 0x3F);
    }
    else if ((*p & 0xF0) == 0xE0) {
        *d = (UWORD)(*p++ & 0x0F);
        *d = (*d << 6) | (*p++ & 0x3F);
        *d = (*d << 6) | (*p++ & 0x3F);
    }
    else if ((*p & 0xF0) == 0xE0) {
        *d = (UWORD)(*p++ & 0x07);
        *d = (*d << 6) | (*p++ & 0x3F);
        *d = (*d << 6) | (*p++ & 0x3F);
        *d = (*d << 6) | (*p++ & 0x3F);
    }
    *d = to_sc(*d);
    return p;
}

int skn_cstr_to_sstr(unsigned char* inputs, UWORD* p, int n, unsigned char* (*func)(unsigned char* p, UWORD* d)) {
    int i = 0;
    unsigned char c = *inputs;
    
    while(c != '\0' && i < n) {
        inputs = func(inputs, p);
        c = *inputs;
        p++;
        i++;
    }
    return i;
}

int skn_utf8_cstr_to_sstr(unsigned char* inputs, UWORD* p, int n) {
    int length;
    length = skn_cstr_to_sstr(inputs, p, n, skn_read_utf8c);
    *(p+length) = SNIL;
    return length;
}

int cstr_to_sstr(unsigned char* inputs, UWORD* p, int n) {
    return skn_utf8_cstr_to_sstr(inputs, p, n);
}

unsigned char* skn_conv_utf8c(unsigned char* p, UWORD d) {
    UWORD e = to_cc(d);
    if (e >= 0 && e <= 0x7F) {
        *p++ = e;
    } 
    else if (e >= 0x80 && e <= 0x7FF) {
        *p++ = (e >> 6) | 0xC0;
        *p++ = (e & 0x3F) | 0x80;
    }
    else if (e >= 0x800 && e <= 0xFFFF) {
        *p++ = (e >> 12) | 0xE0;
        *p++ = ((e >> 6) & 0x3F) | 0x80;
        *p++ = (e & 0x3F) | 0x80;
    }
    else if (e >= 0x10000 && e <= 0x1FFFFF) {
        *p++ = (e >> 18) | 0xF0;
        *p++ = ((e >> 12)& 0x3F) | 0x80;
        *p++ = ((e >> 6) & 0x3F) | 0x80;
        *p++ = (e & 0x3F) | 0x80;
    }
    return p;
}

unsigned char* skn_write_utf8c(unsigned char* p, UWORD* d) {
    return skn_conv_utf8c(p, *d);
}

int skn_sstr_to_cstr(UWORD* p, unsigned char* outputs, int n, unsigned char* (*func)(unsigned char* p, UWORD* d)) {
    int i = 0;
    unsigned char* top = outputs;
    
    while(*p != SNIL && i < n) {
        outputs = func(outputs, p);
        p++;
        i++;
    }
    return (int)(outputs-top);
}

int skn_sstr_to_utf8_cstr(UWORD* p, unsigned char* outputs, int n) {
    int length;
    length = skn_sstr_to_cstr(p, outputs, n, skn_write_utf8c);
    *(outputs+length) = '\0';
    return length;
}

int sstr_to_cstr(UWORD* p, unsigned char* inputs, int n) {
    return skn_sstr_to_utf8_cstr(p, inputs, n);
}

WORD* sukuna_fgets(WORD* p, CELL n, FILE* fp) {
    char inputs[n*wordWidth+1];
    WORD* result;
    
    if(fp == SNIL) return SNIL;
    if(fgets(inputs,sizeof(inputs),fp)!=NULL) {
        *((CELL*)p) = (CELL)to_si(cstr_to_sstr((unsigned char*)inputs,(UWORD*)p+cell_size,(int)n));
        result = p;
    }
    else {
        result = SNIL;
    }
    return result;
}

WORD* sukuna_gets(WORD* p, CELL n) {    
    return sukuna_fgets(p,n,stdin);
}

CELL sukuna_getsw(void) {
    CELL n = to_ci(pop_data);
    WORD* p = (WORD*)sp_to_cp(pop_data);
    sukuna_gets(p,n);
    return SNIL;
}

void sukuna_puts(WORD* p) {    
    int n = to_addr_n(p);
    unsigned char buf[n*wordWidth+1];
    sstr_to_cstr((UWORD*)p,buf,n);
    printf("%s",buf);
}

CELL sukuna_putsw(void) {
    WORD* p = (WORD*)sp_to_cp(pop_data);
    sukuna_puts(p);
    return SNIL;
}

void sukuna_type(WORD* p, CELL n) {
    unsigned char buf[n*wordWidth+1];
    
    sstr_to_cstr((UWORD*)p,buf,n);
    printf("%s",buf);
}

void sukuna_cputs(char* p) {
    printf("%s",p);
}

void sukuna_msg(char* p) {
    if(gQuietFlag==SNIL) printf("%s",p);
}

void sukuna_cr() {
    printf("\n");
}

void sukuna_dot(CELL n) {
    if(check_si(n)) {
        printf(" %ld",to_ci(n));
    }
    else if(check_sc(n)) {
        printf(" ");
        sukuna_emit(n);
    }
    else if(n == SNIL) {
        printf(" nil");
    }
    else if(n == STRUE) {
        printf(" true");
    }
    else if(n == SFALSE) {
        printf(" false");
    }
    else if(n == SEOF) {
        printf(" eof");
    }
    else if(n == SEOS) {
        printf(" eos");
    }
    else {
        if(n>=0) {
            printf(" $%ld",(long)n);
        }
        else {
            printf("-$%ld",(long)(-n));
        }
    }
}

void sukuna_dot_hex(CELL n) {
    if(check_si(n)) {
        printf(" %lx",to_ci(n));
    }
    else if(check_sc(n)) {
        printf(" &#x%lx;",(long)to_cc(n));
    }
    else if(n == SNIL) {
        printf(" nil");
    }
    else if(n == STRUE) {
        printf(" true");
    }
    else if(n == SFALSE) {
        printf(" false");
    }
    else if(n == SEOF) {
        printf(" eof");
    }
    else if(n == SEOS) {
        printf(" eos");
    }
    else {
        printf(" $0x%lx",(long)n);
    }
}

void sukuna_emit(UWORD n) {
    unsigned char s[5];
    unsigned char* p;
    p = skn_conv_utf8c(s, n);
    *p = '\0';
    printf("%s",s); 
}

/***********************************************************************************
* Debug
***********************************************************************************/
void sukuna_memdump(CELL* top, CELL length) {
    CELL i;
    
    for(i=0; i<length; i++) {
        printf("%.8lx ",*(top+i));
        if(i%8==7) printf("\n");
    }
}

CELL sukuna_memdumpw(void) {
    CELL  length = (CELL)to_ci(pop_data); 
    CELL* top = (CELL*)sp_to_cp(pop_data); 
    sukuna_memdump(top,length);
    return SNIL;
}

int check_ptr(CELL* n) {
    if(check_sp((CELL)n)) {
        if( &memory[0] <= n && n < &memory[dataTop] ) return c_area;
        if( &memory[dataTop] <= n && n < &memory[memSize] ) return d_area;
        if( &memory[memSize] <= n && n < &memory[stackTop] ) return s_area;
        if( &memory[stackTop] <= n && n < &memory[rStackTop] ) return r_area;    
        if( &memory[tempTop] <= n && n < &memory[memMax-1] ) return t_area;
        if( isHeapPtr(n) ) return h_area;
    }
    return 0;
}

CELL sukuna_check_value_type(CELL n) {
    if(check_si(n)) {
        return int_type;
    }
    else if(check_sc(n)) {
        return char_type;
    }
    else if(n == SNIL) {
        return var_type;
    }
    else if(n == STRUE) {
        return bool_type;
    }
    else if(n == SFALSE) {
        return bool_type;
    }
    else if(n == SEOF) {
        return var_type;
    }
    else if(n == SEOS) {
        return var_type;
    }
    else if(check_sb(n)) {
        return symb_type;
    }
    else if(check_sp(n)) {
        return ptr_type;
    }
    else {
        return var_type;
    }
}
CELL sukuna_check_value_typew(void) {
    CELL n = pop_data;
    push_data(sukuna_check_value_type(n));
    return SNIL;
}

CELL sukuna_obj_to_type(void) {
    CELL n = pop_data;
    CELL t = var_type;
    sukunaClass* sc;
    
    if(check_sp(n)) {
        if(check_si(*((CELL*)sp_to_cp(n)-1))) {
            t = *((CELL*)sp_to_cp(si_to_ptr(*(sp_to_cp(n)-1))));
        }
        else {
            t = *((CELL*)sp_to_cp(n)-1);
        }
        sc = (sukunaClass*)sp_to_cp(t-cellWidth);
        if(sc->tag!=classTag) {
            t = ptr_type;
        }
    }
    else {
        t = sukuna_check_value_type(n);
    }
    push_data(t);
    return SNIL;
}

void sukuna_debug_nemit(CELL n) {
    if( check_si(n) ) {
        /* small integer */
        printf("<small integer %ld>",to_ci(n));
    }
    else if( check_sc(n) ) {
        /* unicode character */
        printf("<unicode char &#x%lx;>",(long)to_cc(n));
    }
    else if(n == SNIL) {
        printf("<nil $%ld>", (long)n);
    }
    else if(n == STRUE) {
        printf("<true $0x%x>", (unsigned int)n);
    }
    else if(n == SFALSE) {
        printf("<false $0x%x>", (unsigned int)n);
    }
    else if(n == SEOF) {
        printf(" <eof $0x%x>", (unsigned int)n);
    }
    else if(n == SEOS) {
        printf(" <eos $0x%x>", (unsigned int)n);
    }
    else if( check_sp(n) ) {
        /* pointer or object */
        switch(check_ptr(sp_to_cp(n))) {
            case c_area: printf("<code area pointer $0x%x>",(int)n); break;
            case d_area: printf("<data area pointer $0x%x>",(int)n); break;
            case s_area: printf("<stack area pointer $0x%x>",(int)n); break;
            case r_area: printf("<return stack area pointer $0x%x>",(int)n); break;
            case t_area: printf("<temporary area pointer $0x%x>",(int)n); break;
            case h_area: printf("<heap area pointer $0x%x>",(int)n); break;
            default: printf("<unidentified pointer $0x%x>",(int)n);
        }
    }
    else {
        printf("<unidentified symbol $0x%lx;>",(long)n);
    }
}

CELL sukuna_put_error_info(void) {
    if(currentTask->filepath != SNIL) {
        sukuna_cputs("File: ");
        sukuna_puts(currentTask->filepath);
        sukuna_cr();
    }
    printf("%d: ", (int)currentTask->linenum);
    sukuna_puts(currentTask->IN_orig);
    printf(">>>> ");
    sukuna_type(currentTask->current_word,currentTask->current_cnt);
    sukuna_cr();
    return SNIL;
}

void sukuna_print_class_name(CELL clp) {
    sukunaClass* cp;
    if(clp==SNIL) return;
    if(check_sp(clp)) {
        cp = (sukunaClass*)sp_to_cp(clp-cellWidth);
        if(cp->Name!=SNIL) {
            sukuna_puts((WORD*)sp_to_cp(cp->Name));
             printf(":");
        }
        else {
            sukuna_debug_nemit(clp);
        }
    }
    else if(check_si(clp)) {
        switch(clp) {
            case var_type:  printf("any:"); break;
            case func_type: printf("fn:"); break;
            case prim_type: printf("prim:"); break;
            case int_type:  printf("int:"); break;
            case char_type: printf("char:"); break;
            case symb_type: printf("symb:"); break;
            case bool_type: printf("bool:"); break;
            case ptr_type:  printf("ptr:"); break;
            case float_type: printf("float:"); break;
            default: printf("??:");
        }
    }
    else if(check_sv(clp)) {
        if(clp!=EXACTSELF) {
            printf("'%d:",(int)to_cv(clp));
        }
        else {
            printf("'self:");
        }
    }
}

CELL sukuna_print_class_namew(void) {
    CELL clp = pop_data;
    sukuna_print_class_name(clp);
    return SNIL;
}

CELL sukuna_init_type_stack() {
    sukuna_reset_type_stack();
    failure_count = 0;
    return SNIL;
}

CELL sukuna_failure_type_check(void) {
    sukuna_reset_type_stack();
    failure_count++;
    return SNIL;
}

void sukuna_print_stack_annotation(WORD* addr, CELL n, sukunaDicEntry* dp) {
    sukunaTypeList* tp;
    CELL* p;
    CELL  c;
    int i;
    if(dp->TypeList==SNIL) {
        printf(": ");
        sukuna_type(addr,n);
        printf(" unspecified. \n");
        return;
    }
    tp = (sukunaTypeList*)sp_to_cp(dp->TypeList);
    if(tp->arguments==SNIL) {
        printf(": ");
        sukuna_type(addr,n);
        printf(" unspecified. \n");
        return;
    }
    p = (CELL*)sp_to_cp(tp->arguments);
    printf(": ");
    sukuna_type(addr,n);
    printf(" | ");
    c = to_ci(*p++);
    for(i=0;i<c;++i) {
        sukuna_print_class_name(*p++);
        printf(" ");
    }
    if(tp->returnvalues==SNIL) {
        printf("| ");
        sukuna_cr();
        return;
    }
    printf("-- ");
    p = (CELL*)sp_to_cp(tp->returnvalues);
    c = to_ci(*p++);
    for(i=0;i<c;++i) {
        sukuna_print_class_name(*p++);
        printf(" ");
    }
    printf("| ");
    sukuna_cr();
}
CELL sukuna_print_stack_annotationw(void) {
    CELL tos = pop_data;
    CELL Scratch = pop_data;
    sukuna_print_stack_annotation((WORD*)sp_to_cp(pop_data), (CELL)to_ci(Scratch), (sukunaDicEntry*)sp_to_cp(tos)); 
    return SNIL;
}

void sukuna_print_type_stack(WORD* addr, CELL n) {
    int i;
    CELL* p;
    if((failure_count==0)&&(argval_stack!=NULL)&&(retval_stack!=NULL)) {
        printf(": ");
        sukuna_type(addr,n);
        printf(" | ");
        p = argval_stack - (*argval_stack);
        for(i=0;i<*argval_stack;++i) {
            sukuna_print_class_name(*p++);
            printf(" ");
        }
        printf("-- ");
        p = retval_stack+1;
        for(i=0;i<*retval_stack;++i) {
            sukuna_print_class_name(*p++);
            printf(" ");
        }
        printf("| ");
        sukuna_cr();
    }
}

/***********************************************************************************
* Memory Operation
***********************************************************************************/
CELL* sukuna_allot(CELL** p, CELL n, int type) {
    CELL* result;
    
    if(check_ptr(*p+n)==type) {
        result = *p;
        *p += n;
    }
    else {
        result = SNIL;
    }
    return result;
}

CELL* sukuna_callot(CELL n) {
    return sukuna_allot(&codep, n, c_area);
}

CELL* sukuna_dallot(CELL n) {
    return sukuna_allot(&datap, n, d_area);
}

CELL* sukuna_tallot(CELL n) {
    return sukuna_allot(&tempp, n, t_area);
}

void sukuna_forget(CELL* p) {
    switch(check_ptr(p)) {
        case c_area: codep = p; break;
        case d_area: datap = p; break;
        case t_area: tempp = p; break;
    }
}

CELL* sukuna_gallot(CELL n, CELL t) {
    CELL* result;
    switch(t) {
        case c_area: result = sukuna_callot(n); break;
        case d_area: result = sukuna_dallot(n); break;
        case t_area: result = sukuna_tallot(n); break;
        case h_area: result = sukuna_malloc(n); break;
        default: result = SNIL;
    }
    return result;
}

CELL sukuna_heap_blocks(void) {
    sukunaHeapHeader* p = lastHeap;
    CELL i = 0;
    while(p!=SNIL) {
        i++;
        p = (sukunaHeapHeader*)p->prev;        
    }
    return i;
}

CELL sukuna_heap_blocksw(void) {
    push_data(to_si(sukuna_heap_blocks()));
    return SNIL;
}

CELL sukuna_debug_marked () {
    sukunaHeapHeader* p = lastHeap;
    CELL i = 0;
    while(p!=SNIL) {
        if(p->magic!=0) i++;
        p = (sukunaHeapHeader*)p->prev;        
    }
    return i;
}

void sukuna_free(CELL* p) {
    sukunaHeapHeader* headerp = (sukunaHeapHeader*)to_heap_header(p);
    sukunaHeapHeader* pr;
    sukunaHeapHeader* nx;
    
    if(is_valid_heapp(p)) {
        pr = (sukunaHeapHeader*)headerp->prev;
        nx = (sukunaHeapHeader*)headerp->next;
        if(pr!=SNIL) pr->next = (CELL*)nx;
        if(nx!=SNIL) nx->prev = (CELL*)pr;
        headerp->magic = SNIL;
        if(lastHeap==headerp) lastHeap = (sukunaHeapHeader*)headerp->prev;
        free(headerp);
    }
}

CELL sukuna_check_tohp(sukunaHeapHeader* p, CELL n) {
    UCELL htop = (UCELL)&(p->Klass);
    UCELL hend = htop+(UCELL)((to_ci(p->Size)+1)*cellWidth);
    if(htop<=n && n<hend) return STRUE;
    return SNIL;
}

void sukuna_gc_map_set(UCELL* gc_map, UCELL p) {
    UCELL idx = (p-gc_heapmin)>>5;
    UCELL mask = ((UCELL) 1) << (idx&0x1F);
    idx = idx >> 5;
    *(gc_map+idx) |= mask;
}

void sukuna_gc_root_check(CELL* st, CELL* ed, UCELL* gc_map) {
    CELL* current = st;
    UCELL p;
    
    while(current<ed) {
        p = (UCELL)sp_to_cp(*current++);
        if(check_ptr((CELL*)p)==h_area) {
            if((gc_heapmin <= (p&heap_map_mask)) && ((p&heap_map_mask) <= gc_heapmax)) {
                sukuna_gc_map_set(gc_map, p);
            }
        }
    }
}

UCELL sukuna_check_heap_mark(UCELL* gc_map, sukunaHeapHeader* p) {
    UCELL st_idx = ((UCELL)p-gc_heapmin)>>5;
    UCELL ed_idx = (((UCELL)p-gc_heapmin)+((UCELL)to_ci(p->Size))*cellWidth+(UCELL)sizeof(sukunaHeapHeader))>>5;
    UCELL st_bit = st_idx&0x1F;
    UCELL ed_bit = ed_idx&0x1F;
    UCELL idx;
    UCELL bit;
    UCELL mask;
    
    st_idx = st_idx >> 5;
    ed_idx = ed_idx >> 5;
    
    idx = st_idx;
    bit = st_bit;
    mask = (UCELL)1 << bit;
    
    while(1) {
        if(*(gc_map+idx)&mask) return 1;
        ++bit;
        mask = mask << 1;
        if(idx>=ed_idx && bit > ed_bit) break;
        if(bit > 31) {
            bit = 0;
            mask = 1;
            ++idx;
        }
    }
    return 0;
}

void sukuna_gc_mark(UCELL* gc_map) {
    sukunaHeapHeader* p = lastHeap;
    sukunaHeapHeader* firstp;
    CELL flag;
    UCELL htop,hend;
    
    if(p==SNIL) return;

    sukuna_gc_root_check(&memory[memTop],codep,gc_map);
    sukuna_gc_root_check(&memory[dataTop]+bytes_to_cells(sizeof(sukunaGlobals)),datap,gc_map);
    sukuna_gc_root_check(&memory[tempTop],tempp,gc_map);
    
    sukuna_gc_root_check(currentTask->sp,&memory[stackTop],gc_map);
    sukuna_gc_root_check(currentTask->rp,&memory[rStackTop],gc_map);

    while(p!=SNIL) {
        firstp = p;
        if(sukuna_check_heap_mark(gc_map, p)==0) {
            p->magic = 0;
        }
        else {
            p->magic = 1;
        }
        p = (sukunaHeapHeader*)p->prev;
    }

    flag = SNIL;
    while(flag==SNIL) {
        p = lastHeap;
        flag = STRUE;
        while(p!=SNIL) {
            if(p->magic==1) {
                htop = (UCELL)&(p->Klass);
                hend = htop+(UCELL)((to_ci(p->Size)+1)*cellWidth);
                sukuna_gc_root_check((CELL*)htop,(CELL*)hend,gc_map);
                flag = SNIL;
                p->magic = 2;
            }
            else if(p->magic==0 && sukuna_check_heap_mark(gc_map, p)) {
                htop = (UCELL)&(p->Klass);
                hend = htop+(UCELL)((to_ci(p->Size)+1)*cellWidth);
                sukuna_gc_root_check((CELL*)htop,(CELL*)hend,gc_map);
                flag = SNIL;
                p->magic = 2;
            }
            p = (sukunaHeapHeader*)p->prev;
        }

        if(flag==STRUE) break;
        
        p = firstp;
        flag = STRUE;
        while(p!=SNIL) {
            if(p->magic==1) {
                htop = (UCELL)&(p->Klass);
                hend = htop+(UCELL)((to_ci(p->Size)+1)*cellWidth);
                sukuna_gc_root_check((CELL*)htop,(CELL*)hend,gc_map);
                flag = SNIL;
                p->magic = 2;
            }
            else if(p->magic==0 && sukuna_check_heap_mark(gc_map, p)) {
                htop = (UCELL)&(p->Klass);
                hend = htop+(UCELL)((to_ci(p->Size)+1)*cellWidth);
                sukuna_gc_root_check((CELL*)htop,(CELL*)hend,gc_map);
                flag = SNIL;
                p->magic = 2;
            }
            p = (sukunaHeapHeader*)p->next;
        }
    }
}

void sukuna_check_heap_min_max(CELL* p, CELL size) {
    UCELL p_min = ((UCELL)p)&heap_map_mask;
    UCELL p_max = ((UCELL)((UCELL)p+(UCELL)size*cellWidth+(UCELL)(sizeof(sukunaHeapHeader))+0x20))&heap_map_mask;

    if(p_min < gc_heapmin || gc_heapmin == 0) gc_heapmin = p_min;
    if(p_max > gc_heapmax || gc_heapmax == 0) gc_heapmax = p_max;
}

void sukuna_gc_sweep () {
    sukunaHeapHeader* p = lastHeap;
    sukunaHeapHeader* prevHeap;
    
    totalHeapSize = 0;
    gc_heapmin = 0;
    gc_heapmax = 0;
    while(p!=SNIL) {
        if(p->magic==0) {
            prevHeap = (sukunaHeapHeader*)p->prev;
            p->magic = heap_id;
            sukuna_free(&(p->Body));
            p = prevHeap;
        }
        else {
            p->magic = heap_id;
            totalHeapSize += to_ci(p->Size);
            sukuna_check_heap_min_max((CELL*)p, to_ci(p->Size));
            p = (sukunaHeapHeader*)p->prev;        
        }
    }
}

CELL sukuna_gc(void) {
    UCELL* gc_map;
    if(gc_debug!=SFALSE) { printf("GC start: heap block = %d\n",(int)sukuna_heap_blocks()); }
    if((gc_map = (UCELL*)calloc((((gc_heapmax-gc_heapmin)/32/32)+1)*cellWidth,1))==NULL) return SNIL;
    sukuna_gc_mark(gc_map);
    sukuna_gc_sweep();
    free(gc_map);
    if(gc_debug!=SFALSE) { printf("GC end: heap block = %d\n",(int)sukuna_heap_blocks()); }
    return SNIL;
}

void sukuna_gc_finalize() {
    sukunaHeapHeader* p = lastHeap;
    sukunaHeapHeader* prevHeap;
    while(p!=SNIL) {
        prevHeap = (sukunaHeapHeader*)p->prev;
        sukuna_free(&(p->Body));
        p = prevHeap;
    }
}

CELL* sukuna_malloc(CELL size) {
    sukunaHeapHeader* p;
    if((totalHeapSize+size)>to_ci(gc_limit)) {
        if(gc_debug!=SFALSE) { printf("totalHeapSize+size: %d, gc_limit: %d\n",(unsigned int)(totalHeapSize+size),(int)to_ci(gc_limit)); }
        gc_limit = to_si((totalHeapSize+size)*2);
        if(to_ci(gc_limit)>to_ci(gc_limit_max)) { gc_limit = gc_limit_max; }
        sukuna_gc();
    }
    p = (sukunaHeapHeader*)malloc(sizeof(sukunaHeapHeader)+size*cellWidth);
    if(p==NULL) { return SNIL; }
    if(lastHeap!=SNIL) lastHeap->next = (CELL*)p;
    p->prev = (CELL*)lastHeap;
    p->next = SNIL;
    p->Size = to_si(size);
    p->magic = heap_id;
    p->Klass = SNIL;
    p->Body = to_si(0);
    lastHeap = p;
    totalHeapSize += size;
    sukuna_check_heap_min_max((CELL*)p, size);
    return (CELL*)(&(p->Body));
}

/***********************************************************************************
* String Operation
***********************************************************************************/
CELL sukuna_escchar(UWORD c) {
    CELL res;
    switch(c) {
        case (UWORD)to_sc('0'): res = (CELL)to_sc('\0'); break;
        case (UWORD)to_sc('n'): res = (CELL)to_sc('\n'); break;
        case (UWORD)to_sc('r'): res = (CELL)to_sc('\r'); break;
        case (UWORD)to_sc('f'): res = (CELL)to_sc('\f'); break;
        case (UWORD)to_sc('t'): res = (CELL)to_sc('\t'); break;
        case (UWORD)to_sc('v'): res = (CELL)to_sc('\v'); break;
        case (UWORD)to_sc('b'): res = (CELL)to_sc('\b'); break;
        case (UWORD)to_sc('a'): res = (CELL)to_sc('\a'); break;
        default: res = c;
    }
    return res;
}

void sukuna_memcopy(CELL* src, CELL* dst, CELL len) {
    if(src < dst && dst < src+len) {
        for(dst+=len, src+=len ; len > 0 ; len--) {
            *(--dst) = *(--src);
        }
    }
    else {
        while(len>0) {
            *dst++ = *src++;
            len--;
        }
    }   
} 

void sukuna_memmove(WORD* src, WORD* dst, CELL len) {
    if(src < dst && dst < src+len) {
        for(src+=len, dst+=len; len>0; len--) {
            *(--dst) = *(--src);
        }
    }
    else {
        while(len>0) {
            *dst++ = *src++;
            len--;
        }    
    }   
} 

CELL sukuna_strmove(WORD* src, WORD* dst, CELL len) {
    WORD c;
    CELL cnt = 0;
    
    while(len>0) {
        if((c = *src++)==(WORD)to_sc('\\')) {
            *dst++ = (WORD)sukuna_escchar(*src++);
            len--;
        }
        else {
            *dst++ = c;
        }
        len--;
        cnt++;
    }
    return cnt;    
} 
CELL sukuna_strmovew(void) {
    CELL  len = (CELL)to_ci(pop_data);
    WORD* dst = (WORD*)sp_to_cp(pop_data);
    WORD* src = (WORD*)sp_to_cp(pop_data);
    
    push_data((CELL)to_si(sukuna_strmove(src,dst,len)));
    return SNIL;
}

void sukuna_strcopy(WORD* addr, CELL n, WORD* p) {
    *((CELL*)p) = to_si(n);
    p += cell_size;
    while(n>0) {
        *p++ = *addr++;
        n--;
    }
}

void sukuna_upper(WORD* p, CELL n, WORD* token) {
    WORD c;
    
    while(n>0){
        c = *p++;
        *token++ = (WORD)to_sc(to_upper(to_cc(c)));
        n--;
    }
}

void sukuna_lower(WORD* p, CELL n, WORD* token) {
    WORD c;
    
    while(n>0){
        c = *p++;
        *token++ = (WORD)to_sc(to_lower(to_cc(c)));
        n--;
    }
}

CELL sukuna_sstr_compare(UWORD* addr, CELL n, UWORD* addr2, CELL n2) {
    CELL result,limit;
    int i = 0;

    if(n>n2) { limit = n2; result = 1; }
    else if(n<n2) { limit = n; result = -1; }
    else { limit = n; result = 0; }
    while(i<limit) {
        if(*addr > *addr2) { result =  1; break; }
        if(*addr < *addr2) { result = -1; break; }
        i++; addr++; addr2++;
    }
    return result;
}

CELL sukuna_sstr_comparei(UWORD* addr, CELL n, UWORD* addr2, CELL n2) {
    WORD text1[n+1];
    WORD text2[n2+1];
    
    sukuna_lower((WORD*)addr,n,text1);
    sukuna_lower((WORD*)addr2,n2,text2);
    return sukuna_sstr_compare((UWORD*)text1, n, (UWORD*)text2, n2);
}

WORD* sukuna_strcompare(WORD* addr, CELL n, WORD* addr2, CELL n2) {
    WORD* result = addr;
    int i;
    
    if(n!=n2) return SNIL;
    for(i=0;i<n;i++) {
        if(*addr++ != *addr2++) {
            result = SNIL;
            break;
        }
    }
    return result;
}

WORD* sukuna_cstrcompare(WORD* addr, CELL n, char* cstr) {
    WORD* result = addr;
    int i;
    
    for(i=0;i<n && *cstr ;i++) {
        if(*addr++ != (WORD)to_sc(*cstr++)) {
            result = SNIL;
            break;
        }
    }
    if(n!=i || *cstr) {
        return SNIL;
    }
    else {
        return result;
    }
}


CELL sukuna_digit_to_number( UWORD c, CELL base ) {
    CELL result;
    
    if( ((UWORD)'0' <= c) && (c <= (UWORD)'9') ) {
        result = (CELL)( c - (UWORD)'0' );
    }
    else if( ((UWORD)'a' <= c) && (c <= (UWORD)'z') ) {
        result = (CELL)( c - (UWORD)'a' + 0x0A );
    }
    else {
        result = -1;
    }
    if( result >= base ) result = -1;
    return result;
}

CELL sukuna_str_to_num(UWORD* addr, CELL n, CELL* res) {
    int i;
    CELL sign = 1;
    CELL ptr = SNIL;
    CELL base = 10;
    CELL num = 0;
    CELL d = 0;
    
    if(*addr == (UWORD)to_sc('-')) {
        sign = -1;
        addr++;
        n--;
    }
    if(*addr == (UWORD)to_sc('$')) {
        ptr = STRUE;
        addr++;
        n--;
    }
    if(*addr == (UWORD)to_sc('0')) {
        addr++;
        n--;
        switch(*addr) {
            case (UWORD)to_sc('b'): base = 2; addr++; n--; break;
            case (UWORD)to_sc('o'): base = 8; addr++; n--; break;
            case (UWORD)to_sc('d'): base = 10; addr++; n--; break;
            case (UWORD)to_sc('x'): base = 16; addr++; n--; break;
        }
    }
    for(i=0;i<n;i++) {
        if((d=(CELL)sukuna_digit_to_number(to_cc(*addr++), base)) < 0) {
            return SNIL;
        }
        num = num * base + d;
    }
    if(ptr==SNIL) {
        *res = to_si(num*sign);
    }
    else {
        *res = (num*sign);
    }
    return STRUE;
}

CELL sukuna_num_to_str(UWORD* addr, CELL n, UWORD* fma, CELL flen, CELL x) {
    char fstr[flen*wordWidth+1];
    char rstr[n*wordWidth+1];
    CELL result;
    
    sstr_to_cstr(fma,(unsigned char*)fstr,flen);
    if(check_si(x)) {
        result = sprintf(rstr,fstr,to_ci(x));
    }
    else {
        result = sprintf(rstr,fstr,x);
    }
    if((result==EOF)||(result>n)) return SNIL;
    return cstr_to_sstr((unsigned char*)rstr,addr,result);
}

CELL sukuna_numref_to_char(UWORD* addr, CELL n, CELL* res, UWORD* addr2) {
    int i;
    CELL sign = 1;
    CELL base = 10;
    CELL num = 0;
    CELL d = 0;

    if(n < 3) return SNIL;
    if(n == 3 && *addr == (UWORD)to_sc('\'') && *(addr+2) == (UWORD)to_sc('\'')) {
        *res = (CELL)(*(addr2+1));
        return STRUE;
    }
    if(n == 4 && *addr == (UWORD)to_sc('\'') && *(addr+1) == (UWORD)to_sc('\\') && *(addr+3) == (UWORD)to_sc('\'')) {
        *res = sukuna_escchar(*(addr2+2));
        return STRUE;
    }
    if(*addr == (UWORD)to_sc('-')) {
        sign = -1;
        addr++;
        n--;
    }
    if((*addr == (UWORD)to_sc('&')) && (*(addr+1) == (UWORD)to_sc('#'))) {
        addr += 2;
        n -= 2;
    }
    else {
        return SNIL;
    }
    if(*addr == (UWORD)to_sc('x')) {
        base = 16;
        addr++;
        n--;
    }
    if(*(addr+n-1) == (UWORD)to_sc(';')) {
        n--;
    }
    for(i=0;i<n;i++) {
        if((d=(CELL)sukuna_digit_to_number(to_cc(*addr++), base)) < 0) {
            return SNIL;
        }
        num = num * base + d;
    }
    *res = to_sc(num*sign);
    return STRUE;
}

CELL sukuna_numref_to_charw(void) {
    CELL n = to_ci(pop_data);
    UWORD* addr2 = (UWORD*)sp_to_cp(pop_data);
    WORD addr[n];
    CELL res;
    
    sukuna_lower((WORD*)addr2,n,addr);
    if((sukuna_numref_to_char((UWORD*)addr, n, &res, addr2))==STRUE) {
        push_data(res);
    }
    else {
        push_data(SNIL);
    }
    return SNIL;
}

/***********************************************************************************
* Parse & Find 
***********************************************************************************/
CELL sukuna_skip(WORD* p, CELL n, WORD c, WORD** newp) {
    if(c == (WORD)to_sc(BLANK)) {
        while(n>0&&(*p==(WORD)to_sc(BLANK)||*p==(WORD)to_sc('\t'))){
            p++;
            n--;
        }
    }
    else {
        while(n>0&&*p==c){
            p++;
            n--;
        }
    }
    *newp = p;
    return n;
}

CELL sukuna_skipw(WORD c) {
    CELL result = STRUE;

    while((currentTask->cnt = sukuna_skip(currentTask->IN,currentTask->cnt,c,&(currentTask->IN)))<=0) {
        if(sukuna_fgets(currentTask->IN_orig,buf_size,currentTask->stream)==SNIL) {
            result = SFALSE;
            break;
        }
        currentTask->IN = currentTask->IN_orig;
        currentTask->cnt = to_addr_n(currentTask->IN);
        currentTask->linenum++;
    }
    return result;
}

CELL sukuna_scan(WORD* p, CELL n, WORD c, WORD** newp) {
    if(c == (WORD)to_sc(BLANK)) {
        while(n>0&&(*p!=(WORD)to_sc(BLANK)&&*p!=(WORD)to_sc('\t')&&*p!=(WORD)to_sc('\n')&&*p!=(WORD)to_sc('\r'))){
            p++;
            n--;
        }
    }
    else {
        while(n>0&&*p!=c){
            p++;
            n--;
        }
    }
    *newp = p;
    return n;
}

CELL sukuna_scanw(WORD c) {
    CELL result = STRUE;

    while((currentTask->cnt = sukuna_scan(currentTask->IN,currentTask->cnt,c,&(currentTask->IN)))<=0) {
        if(sukuna_fgets(currentTask->IN_orig,buf_size,currentTask->stream)==SNIL) {
            result = SFALSE;
            break;
        }
        currentTask->IN = currentTask->IN_orig;
        currentTask->cnt = to_addr_n(currentTask->IN);
        currentTask->linenum++;
    }
    return result;
}

CELL sukuna_scanww(WORD c, WORD d) {
    CELL result = STRUE;
    CELL level = 0;
    while((result=sukuna_scanw(c))!=SFALSE) {
        if(*((currentTask->IN)-1)==d) level++;
        currentTask->IN++;
        currentTask->cnt--;
        if(*currentTask->IN==d) {
            if(level==0) { break; }
            else { level--; }
        }
    }
    return result;
}

CELL sukuna_scanpw(WORD c, WORD d) {
    CELL result = STRUE;
    CELL level = 0;
    while((result=sukuna_scanw(c))!=SFALSE) {
        if(*((currentTask->IN)-1)==d) level++;
        currentTask->IN++;
        currentTask->cnt--;
        if(*currentTask->IN==d) {
            if(level==0) { break; }
            else { level--; }
        }
    }
    return result;
}

CELL sukuna_next_word(WORD** p, CELL* n, WORD** newp) {
    WORD* IN = *p;
    WORD* addr;
    CELL cnt = *n;
    CELL newn;
    
    cnt = sukuna_skip(IN,cnt,to_sc(BLANK),&IN);
    addr = IN;
    newn = cnt;
    cnt = sukuna_scan(IN,cnt,to_sc(BLANK),&IN);
    newn -= cnt;
    *newp = addr;
    *p = IN;
    *n = cnt;
    return newn;
}

CELL sukuna_get_next_word(WORD** addr) {
    CELL n;

    while((n = sukuna_next_word(&(currentTask->IN),&(currentTask->cnt),addr))<=0) {
        if(sukuna_fgets(currentTask->IN_orig,buf_size,currentTask->stream)==SNIL) {
            n = SNIL;
            break;
        }
        currentTask->IN = currentTask->IN_orig;
        currentTask->cnt = to_addr_n(currentTask->IN);
        currentTask->linenum++;
    }
    return n;
}

CELL sukuna_get_next_wordw(WORD** addr) {
    CELL n;
    n = sukuna_next_word(&(currentTask->IN),&(currentTask->cnt),addr);
    return n;
}

UCELL sukuna_hash(WORD* p, CELL n) {
    UCELL hash = 0;
    
    while(n>0){
        hash = (hash << 3) | (hash >> 29);
        hash ^= (UCELL)to_cc(*p++);
        n--;
    }
    return to_si(hash);
}

UCELL sukuna_mhash(UCELL hash) {
    UCELL h = to_ci(hash);
    h = h>>16 ^ h;
    h = h>>8 ^ h;
    h = h>>4 ^ h;
    return (h&0xF);
}

sukunaDictionary* sukuna_create_dictionary(CELL dicArea) {
    sukunaDictionary* newDic;
    CELL i;
    
    newDic = (sukunaDictionary*)sukuna_gallot(bytes_to_cells(sizeof(sukunaDictionary)),dicArea);
    newDic->ListLink = SNIL;
    newDic->LastEntry = SNIL;
    for(i=0;i<16;i++) {
        newDic->Hash[i]=SNIL;
    }
    return newDic;
}

void sukuna_create_dic_entry(CELL xt, WORD* addr, CELL n, CELL flags, CELL data, CELL mixin, CELL dicArea, sukunaDictionary* dicp) {
    sukunaDicEntry* newEntry;
    UCELL hkey = sukuna_hash(addr,n);
    UCELL mkey = sukuna_mhash(hkey);
    
    if((addr[0] == (WORD)to_sc('[')) && (addr[n-1] == (WORD)to_sc(']'))) {
        sukuna_create_dic_entry(xt, addr+1, n-2, (flags|implicitcall_flag)&(~ignore_mask), data, mixin, dicArea, dicp);
    }
    newEntry = (sukunaDicEntry*)sukuna_gallot(words_to_cells(n)+bytes_to_cells(sizeof(sukunaDicEntry)),dicArea);
    newEntry->ListLink = dicp->LastEntry;
    newEntry->HashLink = dicp->Hash[mkey];
    newEntry->HashKey = (CELL)hkey;
    newEntry->Klass = sVariables->dicEntryC.Body;
    newEntry->Xt = xt;
    newEntry->Data = data;
    newEntry->mixin = mixin;
    newEntry->Flags = flags;
    newEntry->TypeList = SNIL;
    newEntry->Offset = SNIL;
    sukuna_strcopy(addr,n,(WORD*)&(newEntry->NameSize));
    dicp->LastEntry = cp_to_sp((CELL*)newEntry);
    dicp->Hash[mkey] = cp_to_sp((CELL*)newEntry);
}

void sukuna_create_dic_entryd(CELL xt, WORD* addr, CELL n, CELL flags) {
    sukuna_create_dic_entry(xt,addr,n,flags,SNIL,to_si(0),d_area,sdict);
}

void sukuna_create_dic_entryC(CELL xt, char* addr, CELL Flags) {
    UWORD buf[buf_size+1];
    CELL n;
    n = cstr_to_sstr((unsigned char*)addr,buf,buf_size+1);
    sukuna_create_dic_entryd(xt,(WORD*)buf,n,Flags);
}

void sukuna_create_cfunc_entryC(CELL funcID, char* addr, CELL Flags) {
    UWORD buf[buf_size+1];
    CELL n;
    CELL xt = (CELL)cp_to_sp(codep);
    (*codep++) = _cfunci;
    (*codep++) = to_si(funcID);
    (*codep++) = _exit;
    n = cstr_to_sstr((unsigned char*)addr,buf,buf_size+1);
    sukuna_create_dic_entryd(xt,(WORD*)buf,n,Flags);
}

void sukuna_create_dic_entrydw(CELL xt, WORD* addr, CELL n) {
    WORD word[buf_size+1];
    
    sukuna_lower(addr,n,word);
    sukuna_create_dic_entry(xt,word,n,normal_flag,SNIL,to_si(0),d_area,sdict);
}

CELL sukuna_words(void) {
    sukunaDicEntry* currentEntry;
    WORD* p;
    CELL n;
    
    if(sdict->LastEntry==SNIL) return SNIL;
    currentEntry = (sukunaDicEntry*)sp_to_cp(sdict->LastEntry);
    p = (WORD*)&(currentEntry->NameSize);
    n = to_addr_n(p);
    sukuna_type(p,n);
    printf("  ");

    while(currentEntry->ListLink != SNIL) {
        currentEntry = (sukunaDicEntry*)sp_to_cp(currentEntry->ListLink);
        p = (WORD*)&(currentEntry->NameSize);
        n = to_addr_n(p);
        sukuna_type(p,n);
        printf("  ");
    }

    sukuna_cr();
    return SNIL;
}

sukunaDicEntry* sukuna_find(WORD* addr, CELL n, sukunaDictionary* dicp) {
    sukunaDicEntry* currentEntry;
    sukunaDicEntry* found = SNIL;
    UCELL hkey = sukuna_hash(addr,n);
    UCELL mkey = sukuna_mhash(hkey);
    WORD* addr2;
    CELL n2;
    
    if(dicp==SNIL) return SNIL;
    currentEntry = (sukunaDicEntry*)dicp->Hash[mkey];
    while(currentEntry!=SNIL&&found==SNIL) {
        currentEntry = (sukunaDicEntry*)sp_to_cp(currentEntry);
        if(currentEntry->HashKey==hkey && (currentEntry->Flags&ignore_mask)==0) {
            addr2 = (WORD*)&(currentEntry->NameSize);
            n2 = to_addr_n(addr2);
            if(sukuna_strcompare(addr,n,addr2,n2)!=SNIL) {
                found = currentEntry;
            }
        }
        currentEntry = (sukunaDicEntry*)currentEntry->HashLink;
    }
    
    return found;
}

sukunaDicEntry* sukuna_findQ(WORD* addr, CELL n, sukunaDictionary* dicp, CELL* count) {
    WORD buf[n+2];
    sukunaDicEntry* lfa;
    
    if((lfa=sukuna_find(addr,n,dicp))!=SNIL && (lfa->Flags&implicitcall_flag)!=0) {
        sukuna_memmove(addr, buf+1, n);
        buf[0] = (WORD)to_sc('[');
        buf[n+1] = (WORD)to_sc(']');
        *count += 1;
        lfa = sukuna_findQ(buf, n+2, dicp, count);
    }
    if(lfa==SNIL) { *count = 0; }
    return lfa;
}

sukunaDicEntry* sukuna_findC(unsigned char* addr) {
    UWORD buf[buf_size+1];
    CELL n;
    n = cstr_to_sstr(addr,buf,buf_size+1);
    return sukuna_find((WORD*)buf,n,sdict);
}

CELL sukuna_findw(WORD* addr, CELL n, sukunaDictionary* dicp) {
    sukunaDicEntry* foundEntry;
    CELL result = SNIL;
    WORD word[buf_size+1];
    
    sukuna_lower(addr,n,word);
    foundEntry = sukuna_find(word,n,dicp);
    if(foundEntry != SNIL) {
        result = (CELL)cp_to_sp(&(foundEntry->Xt));
    }
    return result;
}

/***********************************************************************************
* word definition
***********************************************************************************/
void sukuna_create_new_header(WORD* addr, CELL n) {
    sukuna_create_dic_entryd((CELL)cp_to_sp(codep), addr, n, normal_flag|ignore_mask);
}

void sukuna_define_new_word(WORD* addr, CELL n) {
    sukuna_create_new_header(addr,n);
    compile_state = STRUE;
}

sukunaDicEntry* sukuna_local_tickp() {
    WORD buf[buf_size+1];
    WORD* addr;
    CELL n;

    n = sukuna_get_next_word(&addr);
    if(n<=0) return SNIL;
    sukuna_lower(addr,n,buf);
    return sukuna_find(buf,n,ldict);
}

CELL sukuna_compile_lstore(CELL token) {
    sukunaDicEntry* found;
    CELL result = SNIL;
    
    found = sukuna_local_tickp();
    if(found==SNIL) return SNIL;
    if(compile_state==STRUE) {
        result = found->Xt;
        *(codep++) = token;
        *(codep++) = result;
    }
    return result;
}

CELL sukuna_check_argument_type(CELL tscl, CELL agcl) {
    if((tscl==var_type)||(agcl==var_type)) return 1;
    if((tscl==prim_type)&&(agcl>=int_type)&&(agcl<=symb_type)) return 1;
    if((agcl==prim_type)&&(tscl>=int_type)&&(tscl<=symb_type)) return 1;
    if((agcl==func_type)&&(tscl==(CELL)sVariables->blockClass.Body)) return 1;
    if((tscl==func_type)&&(agcl==(CELL)sVariables->blockClass.Body)) return 1;
    if((agcl==ptr_type)&&check_sp(tscl)) return 1;
    if((tscl==ptr_type)&&check_sp(agcl)) return 1;
    if((agcl==float_type)&&(tscl==(CELL)sVariables->floatClass.Body)) return 1;
    if((tscl==float_type)&&(agcl==(CELL)sVariables->floatClass.Body)) return 1;
    if(agcl==tscl) { return 1; }
    else if(check_sp(tscl)&&check_sp(agcl)) {
        if(sukuna_find_super_class(agcl, (sukunaClass*)sp_to_cp(tscl-cellWidth))==SNIL) { return 0; }
        else { return 1; }
    }
    else if((tscl==prim_type)&&check_sp(agcl)&&((CELL)sVariables->systemPrim.Body!=SNIL)) {
        if(sukuna_find_super_class(agcl, to_class_cp((CELL)sVariables->systemPrim.Body))==SNIL) { return 0; }
        else { return 1; }
    }
    else if((tscl==int_type)&&check_sp(agcl)&&((CELL)sVariables->systemInt.Body!=SNIL)) {
        if(sukuna_find_super_class(agcl, to_class_cp((CELL)sVariables->systemInt.Body))==SNIL) { return 0; }
        else { return 1; }
    }
    else if((tscl==char_type)&&check_sp(agcl)&&((CELL)sVariables->systemChar.Body!=SNIL)) {
        if(sukuna_find_super_class(agcl, to_class_cp((CELL)sVariables->systemChar.Body))==SNIL) { return 0; }
        else { return 1; }
    }
    else if((tscl==bool_type)&&check_sp(agcl)&&((CELL)sVariables->systemSymb.Body!=SNIL)) {
        if(sukuna_find_super_class(agcl, to_class_cp((CELL)sVariables->systemSymb.Body))==SNIL) { return 0; }
        else { return 1; }
    }
    else if((tscl==symb_type)&&check_sp(agcl)&&((CELL)sVariables->systemSymb.Body!=SNIL)) {
        if(sukuna_find_super_class(agcl, to_class_cp((CELL)sVariables->systemSymb.Body))==SNIL) { return 0; }
        else { return 1; }
    }
    else if((tscl==float_type)&&check_sp(agcl)&&((CELL)sVariables->floatClass.Body!=SNIL)) {
        if(sukuna_find_super_class(agcl, (sukunaClass*)sp_to_cp((CELL)sVariables->floatClass.Body-cellWidth))==SNIL) { return 0; }
        else { return 1; }
    }
    else { return 0; }
}

CELL sukuna_check_retvalue_type(CELL tscl, CELL agcl) {
    if((tscl==var_type)||(agcl==var_type)) return 1;
    if((agcl==func_type)||(agcl==(CELL)sVariables->blockClass.Body)) return 1;    
    if((tscl==prim_type)&&(agcl>=int_type)&&(agcl<=symb_type)) return 1;
    if((agcl==prim_type)&&(tscl>=int_type)&&(tscl<=symb_type)) return 1;
    if((agcl==ptr_type)&&check_sp(tscl)) return 1;
    if((tscl==ptr_type)&&check_sp(agcl)) return 1;
    if((agcl==float_type)&&(tscl==(CELL)sVariables->floatClass.Body)) return 1;
    if((tscl==float_type)&&(agcl==(CELL)sVariables->floatClass.Body)) return 1;
    if(agcl==tscl) { return 1; }
    else if(check_sp(tscl)&&check_sp(agcl)) {
        if(sukuna_find_super_class(agcl, (sukunaClass*)sp_to_cp(tscl-cellWidth))==SNIL) { return 0; }
        else { return 1; }
    }
    else if((tscl==prim_type)&&check_sp(agcl)&&((CELL)sVariables->systemPrim.Body!=SNIL)) {
        if(sukuna_find_super_class(agcl, to_class_cp((CELL)sVariables->systemPrim.Body))==SNIL) { return 0; }
        else { return 1; }
    }
    else if((tscl==int_type)&&check_sp(agcl)&&((CELL)sVariables->systemInt.Body!=SNIL)) {
        if(sukuna_find_super_class(agcl, to_class_cp((CELL)sVariables->systemInt.Body))==SNIL) { return 0; }
        else { return 1; }
    }
    else if((tscl==char_type)&&check_sp(agcl)&&((CELL)sVariables->systemChar.Body!=SNIL)) {
        if(sukuna_find_super_class(agcl, to_class_cp((CELL)sVariables->systemChar.Body))==SNIL) { return 0; }
        else { return 1; }
    }
    else if((tscl==bool_type)&&check_sp(agcl)&&((CELL)sVariables->systemSymb.Body!=SNIL)) {
        if(sukuna_find_super_class(agcl, to_class_cp((CELL)sVariables->systemSymb.Body))==SNIL) { return 0; }
        else { return 1; }
    }
    else if((tscl==symb_type)&&check_sp(agcl)&&((CELL)sVariables->systemSymb.Body!=SNIL)) {
        if(sukuna_find_super_class(agcl, to_class_cp((CELL)sVariables->systemSymb.Body))==SNIL) { return 0; }
        else { return 1; }
    }
    else if((tscl==float_type)&&check_sp(agcl)&&((CELL)sVariables->floatClass.Body!=SNIL)) {
        if(sukuna_find_super_class(agcl, (sukunaClass*)sp_to_cp((CELL)sVariables->floatClass.Body-cellWidth))==SNIL) { return 0; }
        else { return 1; }
    }
    else { return 0; }
}

void sukuna_type_inference_finalize(CELL* p) {
    CELL* ap;
    CELL* rp;
    CELL* pp;
    CELL pcl;
    CELL ncl;
    int i;
    int c;
    int d;
    int w = 0;
    
    if(p==SNIL) return;
    p = (CELL*)sp_to_cp(p);
    ap = (CELL*)sp_to_cp(*p);
    d = to_ci(*ap);
    c = *argval_stack;
    for(i=0;i<c;++i) {
        rv_push(av_pop);
    }
    *argval_stack = 0;
    ap = ap + d;
    for(i=0;i<d;++i) {
        av_push(*ap--);
        (*argval_stack)++;
    }
    for(i=0;i<c;++i) {
        av_push(rv_pop);
        (*argval_stack)++;
    }
        
    ap = (CELL*)sp_to_cp(*p);
    c = *argval_stack;
    d = to_ci(*ap++);
    pp = argval_stack - c;
    if(c!=d) {
        sukuna_put_error_info();
        printf("Warning: argument number mismatch. \n");
    }
    else {
        for(i=0;i<c;++i) {
            pcl = *pp++;
            ncl = *ap++;
            if(sukuna_check_argument_type(pcl,ncl)==0) w = 1;
        }
        if(w==1) {
            sukuna_put_error_info();
            printf("Warning: argument type mismatch. \n");
        }
    }
    
    w = 0;
    
    rp = (CELL*)(*(p+1));
    if(rp==SNIL) return;
    rp = (CELL*)sp_to_cp(rp);
    ap = (CELL*)sp_to_cp(*p);
    
    c = *retval_stack;
    d = to_ci(*rp++);
    pp = retval_stack+1;
    if(c!=d) {
        sukuna_put_error_info();
        printf("Warning: return value number mismatch. \n");
    }
    else {
        for(i=0;i<c;++i) {
            pcl = *pp++;
            ncl = *rp++;
            if(check_sv(ncl)) {
                if(ncl!=EXACTSELF) {
                    ncl = *(ap+to_cv(ncl));
                }
                else {
                    ncl = method_def;
                }
            }
            if(sukuna_check_argument_type(pcl,ncl)==0) w = 1;
        }
        if(w==1) {
            sukuna_put_error_info();
            printf("Warning: return value type mismatch. \n");
        }
    }
}

CELL sukuna_make_default_vtype() {
    sukunaTypeList* typep;
    typep = (sukunaTypeList*)sukuna_dallot(bytes_to_cells(sizeof(sukunaTypeList)));
    typep->arguments = (CELL)cp_to_sp(datap);
    (*datap++) = to_si(0);
    typep->returnvalues = (CELL)cp_to_sp(datap);
    (*datap++) = to_si(1);
    (*datap++) = ptr_type;
    typep->imm_arguments = SNIL;
    typep->imm_returnvalues = SNIL;
    return (CELL)cp_to_sp(typep);
}

void sukuna_set_newtypelist(sukunaDicEntry* newEntry) {
    sukunaTypeList* typep;
    sukunaTypeList* prep = SNIL;
    CELL* ap;
    CELL argset = 0;
    CELL retset = 0;
    int i;
    int c;
    
    if(newEntry->TypeList==SNIL) {
        argset = 1;
        retset = 1;
    }
    else {
        prep = (sukunaTypeList*)sp_to_cp(newEntry->TypeList);
        if((prep->arguments!=SNIL)&&(prep->returnvalues!=SNIL)) return;
        if(prep->arguments==SNIL)  {
            argset = 1;
        }
        else {
            ap = (CELL*)sp_to_cp(prep->arguments);
            if(to_ci(*ap)==0&&prep->returnvalues==SNIL) argset = 1;
        }
        if(prep->returnvalues==SNIL) retset = 1;
    }
    typep = (sukunaTypeList*)sukuna_dallot(bytes_to_cells(sizeof(sukunaTypeList)));
    if(argset==1) {
        typep->arguments = (CELL)cp_to_sp(datap);
        c = *argval_stack;
        (*datap++) = to_si(c);
        ap = argval_stack - c;
        for(i=0;i<c;++i) {
            (*datap++) = *ap++;
        }
    }
    else {
        if(prep!=SNIL) {
            typep->arguments = prep->arguments;
        }
        else {
            typep->arguments = SNIL;
        }
    }
    if(retset==1) {
        typep->returnvalues = (CELL)cp_to_sp(datap);
        c = *retval_stack;
        (*datap++) = to_si(c);
        ap = retval_stack+1;
        for(i=0;i<c;++i) {
            (*datap++) = *ap++;
        }
    }
    else {
        if(prep!=SNIL) {
            typep->returnvalues = prep->returnvalues;
        }
        else {
            typep->returnvalues = SNIL;
        }
    }
    if(prep!=SNIL) {
        typep->imm_arguments = prep->imm_arguments;
        typep->imm_returnvalues = prep->imm_returnvalues;
    }
    else {
        typep->imm_arguments = SNIL;
        typep->imm_returnvalues = SNIL;
    }
    newEntry->TypeList = (CELL)cp_to_sp(typep);
}

void sukuna_clean_type_stack(sukunaDicEntry* newEntry,sukunaDicEntry* selEntry) {
    WORD* addr;
    CELL n;
    
    if((failure_count==0)&&(argval_stack!=NULL)&&(retval_stack!=NULL)) {
        sukuna_type_inference_finalize((CELL*)newEntry->TypeList);
        if(selEntry==SNIL) {
            addr = (WORD*)(&(newEntry->NameSize));
            n = to_addr_n(addr);
        }
        else {
            addr = (WORD*)(&(selEntry->NameSize));
            n = to_addr_n(addr);
        }
        if(is_optlevel(1)&&is_warnlevel(1)) sukuna_print_type_stack(addr,n);
        if(is_optlevel(1)&&(sVariables->typeinference.Body==STRUE)) sukuna_set_newtypelist(newEntry);
        
        sukuna_reset_type_stack();
        failure_count = 0;
    }
}
CELL sukuna_clean_type_stackw(void) {
    CELL tos = pop_data;
    CELL Scratch = pop_data;
    if(Scratch!=SNIL) {
        sukuna_clean_type_stack((sukunaDicEntry*)sp_to_cp(tos),(sukunaDicEntry*)sp_to_cp(Scratch));
    }
    else {
        sukuna_clean_type_stack((sukunaDicEntry*)sp_to_cp(tos),(sukunaDicEntry*)SNIL);
    }
    return SNIL;
}

void sukuna_finish_type_stack() {
    sukunaDicEntry* newEntry;

    newEntry = (sukunaDicEntry*)sp_to_cp(sdict->LastEntry);
    sukuna_clean_type_stack(newEntry,(sukunaDicEntry*)SNIL);
}

void sukuna_finish_def() {
    sukunaDicEntry* newEntry;

    newEntry = (sukunaDicEntry*)sp_to_cp(sdict->LastEntry);
    newEntry->Flags &= ~ignore_mask;
}

void sukuna_rcb() {
    if((has_local)!=SFALSE) { 
        *(codep++) = _lc_exit;
        sukuna_forget((CELL*)ldict);
        ldict = SNIL;
        has_local = SFALSE;
    }
    else {
        *(codep++) = _exit;
    }
    *(codep++) = _halt;
    compile_state = SFALSE;
}

void sukuna_rcbw() {
    sukuna_rcb();
    sukuna_finish_def();
    sukuna_finish_type_stack();
}

void sukuna_create_new_label(WORD* addr, CELL n) {
    sukunaDicEntry* newEntry;

    sukuna_create_new_header(addr,n);
    newEntry = (sukunaDicEntry*)sp_to_cp(sdict->LastEntry);
    newEntry->Data = (CELL)cp_to_sp(datap);
    *(codep++) = _lit_at;
    *(codep++) = (CELL)cp_to_sp(&(newEntry->Data));
    *(codep++) = _exit;
    *(codep++) = _exit;
    *(codep++) = _halt;
    sukuna_finish_def();    
}

void sukuna_create_new_labelC(char* addr, CELL* p) {
    sukunaDicEntry* newEntry;
    UWORD buf[buf_size+1];
    CELL n;
    n = cstr_to_sstr((unsigned char*)addr,buf,buf_size+1);
    sukuna_create_new_label((WORD*)buf,n);
    newEntry = (sukunaDicEntry*)sp_to_cp(sdict->LastEntry);
    newEntry->Data = (CELL)cp_to_sp(p);
    newEntry->TypeList = sVariables->defaultvartype.Body;
}

void sukuna_createp(WORD* addr, CELL n) {
    WORD buf[buf_size+1];
    sukuna_lower(addr,n,buf);
    sukuna_create_new_label(buf,n);
}

void sukuna_create() {
    WORD* addr;
    CELL n;
    
    if((n = sukuna_get_next_word(&addr)) > 0) {
        sukuna_createp(addr,n);
    }
}

CELL sukuna_to_xt(CELL xt) {
    CELL result;
    result = *((CELL*)sp_to_cp(xt));
    return check_si(result) ? result : xt ;
}

CELL* sukuna_generate_block(CELL* xt) {
    UCELL lv_total,lv_num;
    UCELL* lv_p,op;
    CELL* lp;
    sukunaProcBlock* newxt;
    sukunaProcBlock* orig;
    
    if(*(xt+1)==SNIL) return xt;
    lv_p = (UCELL*)sp_to_cp((*(xt+3)));
    lp = (CELL*)sp_to_cp(*lv_p);
    lv_num = *(lv_p+1);
    lv_total = to_ci(lv_num);
    op = *(lv_p+2);
    newxt = (sukunaProcBlock*)sukuna_malloc(bytes_to_cells(sizeof(sukunaProcBlock))+lv_total);
    orig = (sukunaProcBlock*)xt;
    *(((CELL*)newxt)-1) = (CELL)sVariables->blockClass.Body;
    newxt->prologue = (CELL*)cp_to_sp(&(newxt->blkin));
    newxt->body = orig->body;
    newxt->blkin = _bl_entry;
    newxt->data = (CELL*)cp_to_sp(&(newxt->savedlp));
    newxt->xt = (CELL*)cp_to_sp(&(newxt->body));
    newxt->blkout = _bl_exit;
    newxt->savedlp = (CELL*)cp_to_sp((CELL*)(&(newxt->locals))+lv_total);
    newxt->lvn = lv_num;
    newxt->selfp = (CELL*)op;
    sukuna_memcopy((lp-lv_total),(CELL*)(&(newxt->locals)),lv_total);
    
    return (CELL*)newxt;
}

CELL* sukuna_block_to_closure(CELL* xt) {
    if(check_ptr(xt)==h_area) {
        return xt;
    }
    else {
        return sukuna_generate_block(xt);
    }
}

/***********************************************************************************
* object
***********************************************************************************/
sukunaClass* sukuna_create_class_header(CELL dicArea) {
    sukunaClass* newClass;
    CELL i;
    
    newClass = (sukunaClass*)sukuna_gallot(bytes_to_cells(sizeof(sukunaClass)),dicArea);
    newClass->Klass = SNIL;
    newClass->vtbl = SNIL;
    newClass->ivtbl = SNIL;
    newClass->supers = SNIL;
    newClass->vmlen = to_si(0);
    newClass->ivlen = to_si(0);
    newClass->flag = to_si(0);
    newClass->tag = class_id;
    newClass->ListLink = SNIL;
    newClass->LastEntry = SNIL;
    for(i=0;i<16;i++) {
        newClass->Hash[i] = SNIL;
    }
    newClass->ivarEntry = SNIL;
    for(i=0;i<16;i++) {
        newClass->iHash[i] = SNIL;
    }
    newClass->Name = SNIL;
    return newClass;
}

void sukuna_inherit_class(sukunaClass* theClass, sukunaClass* supcl) { 
    theClass->Klass = supcl->Klass;
    theClass->vmlen = supcl->vmlen;
    theClass->ivlen = supcl->ivlen;
    theClass->flag = supcl->flag;
    theClass->ListLink = (CELL*)cp_to_sp(&(supcl->Klass));
    theClass->LastEntry = SNIL;
    theClass->ivarEntry = SNIL;
}

void sukuna_create_method_entry(CELL xt, CELL selector, CELL flags, CELL data, CELL mixin, CELL dicArea, sukunaClass* theClass) {
    sukunaDicEntry* newEntry;
    UCELL hkey = selector;
    UCELL mkey = (selector >> 2)&15 ;
    
    newEntry = (sukunaDicEntry*)sukuna_gallot(bytes_to_cells(sizeof(sukunaDicEntry)),dicArea);
    newEntry->ListLink = theClass->LastEntry;
    newEntry->HashLink = theClass->Hash[mkey];
    newEntry->HashKey = (CELL)hkey;
    newEntry->Klass = sVariables->mthEntryC.Body;
    newEntry->Xt = xt;
    newEntry->Data = data;
    newEntry->mixin = mixin;
    newEntry->Flags = flags;
    newEntry->TypeList = SNIL;
    newEntry->Offset = SNIL;
    theClass->LastEntry = cp_to_sp((CELL*)newEntry);
    theClass->Hash[mkey] = cp_to_sp((CELL*)newEntry);
}

void sukuna_create_method_entryw(sukunaClass* theClass, CELL xt, CELL selector,  CELL offset) {
    sukuna_create_method_entry(xt, selector, normal_flag|ignore_mask, offset, to_si(0), d_area, theClass);
}

void sukuna_create_ivar_entry(CELL xt, WORD* addr, CELL n, CELL flags, CELL data, CELL mixin, CELL dicArea, sukunaClass* theClass) {
    sukunaDicEntry* newEntry;
    UCELL hkey = sukuna_hash(addr,n);
    UCELL mkey = sukuna_mhash(hkey);
    
    if((addr[0] == (WORD)to_sc('[')) && (addr[n-1] == (WORD)to_sc(']'))) {
        sukuna_create_ivar_entry(xt, addr+1, n-2, (flags|implicitcall_flag)&(~ignore_mask), data, mixin, dicArea, theClass);
    }
    newEntry = (sukunaDicEntry*)sukuna_gallot(words_to_cells(n)+bytes_to_cells(sizeof(sukunaDicEntry)),dicArea);
    newEntry->ListLink = theClass->ivarEntry;
    newEntry->HashLink = theClass->iHash[mkey];
    newEntry->HashKey = (CELL)hkey;
    newEntry->Klass = sVariables->ivrEntryC.Body;
    newEntry->Xt = xt;
    newEntry->Data = data;
    newEntry->mixin = mixin;
    newEntry->Flags = flags;
    newEntry->TypeList = SNIL;
    newEntry->Offset = to_si(1);
    sukuna_strcopy(addr,n,(WORD*)&(newEntry->NameSize));
    theClass->ivarEntry = cp_to_sp((CELL*)newEntry);
    theClass->iHash[mkey] = cp_to_sp((CELL*)newEntry);
}

void sukuna_create_ivar_entryw(sukunaClass* theClass, sukunaClass* ivarClass, WORD* addr, CELL n,  CELL offset) {
    WORD word[buf_size+1];
    
    sukuna_lower(addr,n,word);    
    sukuna_create_ivar_entry((CELL)ivarClass, word, n, normal_flag, offset, to_si(0), d_area, theClass);
}

void sukuna_setup_vtbl(sukunaDicEntry* currentMethod, CELL* newVtbl) {
    if(currentMethod!=SNIL) {
        currentMethod = (sukunaDicEntry*)sp_to_cp(currentMethod);
        sukuna_setup_vtbl((sukunaDicEntry*)(currentMethod->ListLink),newVtbl);
        *((CELL*)((CELL)newVtbl+(currentMethod->Data))) = currentMethod->Xt;
    }
}

void sukuna_create_vtbl(sukunaClass* theClass, CELL area) {
    sukunaClass* supcl;
    CELL* newVtbl = (CELL*)sukuna_gallot(to_ci(theClass->vmlen), area);
    sukunaDicEntry* currentMethod;
    sukunaMixin* supers;
    
    if(theClass->supers!=SNIL) {
        supers = (sukunaMixin*)sp_to_cp(theClass->supers);
        while((supers->Klass)!=SNIL) {
            supcl = (sukunaClass*)(sp_to_cp(supers->Klass)-1);
            sukuna_memcopy((CELL*)sp_to_cp(supcl->vtbl),(CELL*)((CELL)newVtbl+((supers->offset)&offset_mask)),to_ci(supcl->vmlen));
            supers++;    
        }
    }
    
    currentMethod = (sukunaDicEntry*)theClass->LastEntry;
    sukuna_setup_vtbl(currentMethod, newVtbl);

    if(to_ci(theClass->vmlen)!=0) theClass->vtbl = cp_to_sp(newVtbl);
}

void sukuna_create_vtblw(sukunaClass* theClass) {
    sukuna_create_vtbl(theClass, d_area);
}

void sukuna_update_vtbl(sukunaClass* theClass) {
    CELL* newVtbl = (CELL*)sp_to_cp(theClass->vtbl);
    sukunaDicEntry* currentMethod;

    currentMethod = (sukunaDicEntry*)theClass->LastEntry;
    sukuna_setup_vtbl(currentMethod, newVtbl);
}

sukunaDicEntry* sukuna_method_find(CELL selector, sukunaClass* theClass) {
    sukunaDicEntry* currentEntry;
    sukunaDicEntry* found = SNIL;
    UCELL hkey = selector;
    UCELL mkey = (selector >> 2)&15;
    
    if(theClass==SNIL) return SNIL;
    currentEntry = (sukunaDicEntry*)theClass->Hash[mkey];
    while(currentEntry!=SNIL&&found==SNIL) {
        currentEntry = (sukunaDicEntry*)sp_to_cp(currentEntry);
        if(currentEntry->HashKey==hkey && (currentEntry->Flags&ignore_mask)==0) {
            found = currentEntry;
        }
        currentEntry = (sukunaDicEntry*)currentEntry->HashLink;
    }
    
    return found;
}

sukunaDicEntry* sukuna_mixin_method_find(CELL selector, sukunaClass* theClass) {
    sukunaDicEntry* found = SNIL;
    sukunaMixin* supers;
    
    if(theClass==SNIL) return SNIL;
    supers = (sukunaMixin*)theClass->supers;
    if(supers==SNIL) return SNIL;
    supers = (sukunaMixin*)sp_to_cp(supers);
    while(((supers+1)->Klass)!=SNIL&&found==SNIL) {
        found = sukuna_method_find(selector,(sukunaClass*)((CELL)sp_to_cp(supers->Klass)-cellWidth));
        supers++;
    }
    
    return found;
}

sukunaDicEntry* sukuna_find_method(CELL selector, sukunaClass* theClass) {
    sukunaClass* currentClass = theClass;
    sukunaDicEntry* found = SNIL;
    
    found = sukuna_method_find(selector, currentClass);
    if(found==SNIL) found = sukuna_mixin_method_find(selector, currentClass);
    while(currentClass->ListLink!=SNIL && found==SNIL) {
        currentClass = (sukunaClass*)sp_to_cp(currentClass->ListLink);
        found = sukuna_method_find(selector, currentClass);
        if(found==SNIL) found = sukuna_mixin_method_find(selector, currentClass);
    }
    
    return found;
}

sukunaDicEntry* sukuna_find_reqmixin_method(CELL selector, sukunaClass* theMixin, sukunaClass** reqClass) {
    sukunaClass* currentMixin = theMixin;
    sukunaDicEntry* found = SNIL;
    sukunaMixin* superMixin;
    
    if((found = sukuna_method_find(selector, theMixin))!=SNIL) {
        return found;
    }
    else {
        superMixin = (sukunaMixin*)theMixin->supers;
        if(superMixin==SNIL) return SNIL;
        superMixin = (sukunaMixin*)sp_to_cp(superMixin);
        superMixin += 2;
        while((superMixin->Klass)!=SNIL&&found==SNIL) {
            currentMixin = (sukunaClass*)((CELL)sp_to_cp(superMixin->Klass)-cellWidth);
            if(currentMixin->flag&mixin_mask) {
                found = sukuna_find_reqmixin_method(selector,currentMixin,reqClass);
            }
            else {
                found = sukuna_find_method(selector,currentMixin);
                if(found!=SNIL) { *reqClass = currentMixin; }
            }
            superMixin++;
        }
    }
    return found;
}
CELL sukuna_find_reqmixin_methodw() {
    sukunaClass* reqClass = SNIL;
    sukunaClass* theMixin = (sukunaClass*)((CELL)sp_to_cp(pop_data)-cellWidth);
    CELL selector = pop_data;
    sukunaDicEntry* found;

    found = sukuna_find_reqmixin_method(selector,theMixin,&reqClass);
    if(reqClass!=SNIL) {
        push_data((CELL)cp_to_sp(&(reqClass->vtbl)));
    }
    else {
        push_data(SNIL);
    }
    if(found!=SNIL) {
        push_data((CELL)cp_to_sp(&(found->Xt)));
    }
    else {
        push_data(SNIL);
    }
    return SNIL;
}

sukunaDicEntry* sukuna_find_methodw(CELL selector, sukunaClass* theClass) {
    sukunaDicEntry* found = SNIL;
    
    found = sukuna_find_method(selector, theClass);
    if(found!=SNIL) found = (sukunaDicEntry*)cp_to_sp(&(found->Xt));    
    return found;
}

sukunaDicEntry* sukuna_find_self_methodw(CELL selector, sukunaClass* theClass) {
    sukunaDicEntry* found;
    found = sukuna_method_find(selector, theClass);
    if(found!=SNIL) found = (sukunaDicEntry*)cp_to_sp(&(found->Xt));
    return found;
}

sukunaDicEntry* sukuna_find_super_method(CELL selector, sukunaClass* theClass) {
    sukunaClass* currentClass = theClass;
    sukunaDicEntry* found = SNIL;
    
    if(found==SNIL) found = sukuna_mixin_method_find(selector, currentClass);
    while(currentClass->ListLink!=SNIL && found==SNIL) {
        currentClass = (sukunaClass*)sp_to_cp(currentClass->ListLink);
        found = sukuna_method_find(selector, currentClass);
        if(found==SNIL) found = sukuna_mixin_method_find(selector, currentClass);
    }
    
    return found;
}

sukunaDicEntry* sukuna_find_super_methodw(CELL selector, sukunaClass* theClass) {
    sukunaDicEntry* found = SNIL;

    found = sukuna_find_super_method(selector, theClass);
    if(found!=SNIL) found = (sukunaDicEntry*)cp_to_sp(&(found->Xt));
    return found;
}

CELL* sukuna_mixin_offset_find(sukunaClass* theMixin,sukunaClass* theClass) {
    CELL* found = SNIL;
    sukunaMixin* supers;
    
    if(theClass==SNIL) return SNIL;
    supers = (sukunaMixin*)theClass->supers;
    if(supers==SNIL) return SNIL;
    supers = (sukunaMixin*)sp_to_cp(supers);
    while(((supers+1)->Klass)!=SNIL&&found==SNIL) {
        if(supers->Klass==theMixin) found = (CELL*)(&(supers->offset));
        supers++;
    }    
    return found;
}

CELL* sukuna_find_mixin_offset(sukunaClass* theMixin,sukunaClass* theClass) {
    sukunaClass* currentClass = theClass;
    CELL* found = SNIL;
    
    found = sukuna_mixin_offset_find(theMixin, currentClass);
    while(currentClass->ListLink!=SNIL && found==SNIL) {
        currentClass = (sukunaClass*)sp_to_cp(currentClass->ListLink);
        found = sukuna_mixin_offset_find(theMixin, currentClass);
    }
    
    return found;
}

CELL sukuna_is_stateless_mixin(CELL mixin) {
    sukunaClass* theMixin;
    
    if(mixin!=SNIL && check_sp(mixin)) {
        theMixin = (sukunaClass*)sp_to_cp(mixin-cellWidth);
        if( theMixin->ivlen==to_si(0) ) {
            return STRUE;
        }
        else {
            return SFALSE;
        }
    }
    else {
        return SFALSE;
    }
}

CELL sukuna_is_stateless_mixinw(void) {
    CELL mixin = pop_data;
    push_data(sukuna_is_stateless_mixin(mixin));
    return SNIL;
}

CELL* sukuna_mixin_super_find(sukunaClass* theMixin,sukunaClass* theClass, CELL obp, CELL selfp, CELL selID, CELL* vofs) {
    CELL* found = SNIL;
    CELL* mtp = SNIL;
    CELL* ofsp =SNIL;
    sukunaMixin* supers;
    
    if(theClass==SNIL) return SNIL;
    supers = (sukunaMixin*)theClass->supers;
    if(supers==SNIL) return SNIL;
    supers = (sukunaMixin*)sp_to_cp(supers);
    
    while(((supers+1)->Klass)!=SNIL && found==SNIL) {
        if(supers->Klass==theMixin) {
            if(sukuna_is_stateless_mixin((CELL)theMixin)==SFALSE) {
                if(*((CELL*)(selfp+(CELL)((supers->offset)>>16)))==obp) found = (CELL*)(supers+1);
            }
            else {
                found = (CELL*)(supers+1);
            }
        }
        supers++;
    }
    if(found!=SNIL) {
        while((supers->Klass)!=SNIL && mtp==SNIL) {
            mtp = (CELL*)sukuna_find_method(selID, (sukunaClass*)sp_to_cp(((CELL)(supers->Klass)-cellWidth)));
            if(mtp!=SNIL) {
                if(((supers+1)->Klass)!=SNIL) {
                    if(sukuna_is_stateless_mixin((CELL)supers->Klass)==SFALSE) {
                        *vofs = (CELL)((supers->offset)>>16);
                    }
                    else {
                        *vofs = to_si(0);
                    }
                }
                else {
                    if(check_sp(((sukunaDicEntry*)mtp)->mixin)) {
                        if(sukuna_is_stateless_mixin(((sukunaDicEntry*)mtp)->mixin)==SFALSE) {
                            ofsp = sukuna_find_mixin_offset((sukunaClass*)(((sukunaDicEntry*)mtp)->mixin), (sukunaClass*)sp_to_cp(((CELL)(supers->Klass)-cellWidth)));
                            if(ofsp!=SNIL) {
                                *vofs = *ofsp >> 16;
                            }
                            else {
                                mtp = SNIL;
                            }
                        }
                        else {
                            *vofs = to_si(0);
                        }
                    }
                    else {
                        *vofs = to_si(0);
                    }
                }
                found = (CELL*)cp_to_sp(&(((sukunaDicEntry*)mtp)->Xt));
            }
            supers++;
        }
        if(mtp==SNIL) found = SNIL;
    }
    return found;
}

CELL* sukuna_find_mixin_super(sukunaClass* theMixin,sukunaClass* theClass, CELL obp, CELL selfp, CELL selID, CELL* vofs) {
    sukunaClass* currentClass = theClass;
    CELL* found = SNIL;
    
    found = sukuna_mixin_super_find(theMixin, currentClass, obp, selfp, selID, vofs);
    while(currentClass->ListLink!=SNIL && found==SNIL) {
        currentClass = (sukunaClass*)sp_to_cp(currentClass->ListLink);
        found = sukuna_mixin_super_find(theMixin, currentClass, obp, selfp, selID, vofs);
    }
    return found;
}

sukunaDicEntry* sukuna_ivar_find(WORD* addr, CELL n, sukunaClass* theClass) {
    sukunaDicEntry* currentEntry;
    sukunaDicEntry* found = SNIL;
    UCELL hkey = sukuna_hash(addr,n);
    UCELL mkey = sukuna_mhash(hkey);
    WORD* addr2;
    CELL n2;
    
    if(theClass==SNIL) return SNIL;
    currentEntry = (sukunaDicEntry*)theClass->iHash[mkey];
    while(currentEntry!=SNIL&&found==SNIL) {
        currentEntry = (sukunaDicEntry*)sp_to_cp(currentEntry);
        if(currentEntry->HashKey==hkey && (currentEntry->Flags&ignore_mask)==0) {
            addr2 = (WORD*)&(currentEntry->NameSize);
            n2 = to_addr_n(addr2);
            if(sukuna_strcompare(addr,n,addr2,n2)!=SNIL) {
                found = currentEntry;
            }
        }
        currentEntry = (sukunaDicEntry*)currentEntry->HashLink;
    }
    
    return found;
}

sukunaDicEntry* sukuna_mixin_ivar_find(WORD* addr, CELL n, sukunaClass* theClass) {
    sukunaDicEntry* found = SNIL;
    sukunaMixin* supers;
    
    if(theClass==SNIL) return SNIL;
    supers = (sukunaMixin*)theClass->supers;
    if(supers==SNIL) return SNIL;
    supers = (sukunaMixin*)sp_to_cp(supers);
    while(((supers+1)->Klass)!=SNIL&&found==SNIL) {
        found = sukuna_ivar_find(addr,n,(sukunaClass*)((CELL)sp_to_cp(supers->Klass)-cellWidth));
        supers++;
    }
    
    return found;
}

sukunaDicEntry* sukuna_find_ivar(WORD* addr, CELL n, sukunaClass* theClass) {
    sukunaClass* currentClass = theClass;
    sukunaDicEntry* found = SNIL;
    
    found = sukuna_ivar_find(addr,n,currentClass);
    if(found==SNIL) found = sukuna_mixin_ivar_find(addr,n, currentClass);
    while(currentClass->ListLink!=SNIL && found==SNIL) {
        currentClass = (sukunaClass*)sp_to_cp(currentClass->ListLink);
        found = sukuna_ivar_find(addr,n,currentClass);
        if(found==SNIL) found = sukuna_mixin_ivar_find(addr,n, currentClass);
    }
    
    return found;
}

sukunaDicEntry* sukuna_find_ivarQ(WORD* addr, CELL n, sukunaClass* theClass, CELL* count) {
    WORD buf[n+2];
    sukunaDicEntry* lfa;
    
    if((lfa=sukuna_find_ivar(addr,n,theClass))!=SNIL && (lfa->Flags&implicitcall_flag)!=0) {
        sukuna_memmove(addr, buf+1, n);
        buf[0] = (WORD)to_sc('[');
        buf[n+1] = (WORD)to_sc(']');
        *count += 1;
        lfa = sukuna_find_ivarQ(buf, n+2, theClass, count);
    }
    if(lfa==SNIL) { *count = 0; }
    return lfa;
}

sukunaDicEntry* sukuna_find_ivarw(WORD* addr, CELL n, sukunaClass* theClass) {
    sukunaDicEntry* found = SNIL;
    WORD word[buf_size+1];
    
    sukuna_lower(addr,n,word);        
    found = sukuna_find_ivar(word,n,theClass);
    if(found!=SNIL) found = (sukunaDicEntry*)cp_to_sp(&(found->Xt));    
    return found;
}

sukunaDicEntry* sukuna_find_ivari(WORD* addr, CELL n, CELL* count) {
    sukunaClass* theClass;
    sukunaDicEntry* found = SNIL;
    if(class_def==SNIL) return SNIL;
    theClass = (sukunaClass*)sp_to_cp(class_def-cellWidth);
    found = sukuna_find_ivarQ(addr,n,theClass,count);
    if(compile_state==SFALSE) {
        if(found!=SNIL&&found->Flags&innerfunc_flag) {
            return found;
        }
        else if(found!=SNIL&&found->Flags&class_variable) {
            return found;
        }
        else {
            return SNIL;
        }
    }
    return found;
}

void sukuna_check_abstract(sukunaClass* theClass, sukunaDicEntry* theMethod) {
    sukunaDicEntry* selentry;
    WORD* clsname;
    
    if(theMethod->Flags&abstract_flag) {
        if((theClass->flag&mixin_mask)==0) {
            sukuna_put_error_info();
            printf("Warning: method not implemented ( ");
            selentry = (sukunaDicEntry*)(((CELL*)sp_to_cp(theMethod->HashKey-XtOffset(sukunaDicEntry*,Xt))));
            sukuna_puts((WORD*)&(selentry->NameSize));
            printf(" ");
            clsname = (WORD*)sp_to_cp(theClass->Name);
            sukuna_puts(clsname);
            printf(" )\n");
        }
    }
}

void sukuna_compile_localfetch(sukunaDicEntry* lfa) {
    if(to_ci(lfa->Xt) <= 8) {
        *(codep++) = _lc_fetch1 + (to_ci(lfa->Xt)-1)*2;
    }
    else {
        *(codep++) = _lc_fetch;
        *(codep++) = lfa->Xt;
    }
}
CELL sukuna_compile_localfetchw(void) {
    sukunaDicEntry* lfa = (sukunaDicEntry*)sp_to_cp(pop_data-XtOffset(sukunaDicEntry*,Xt));
    sukuna_compile_localfetch(lfa);
    return SNIL;
}

CELL sukuna_get_postfix_selector(void) {
    sukunaDicEntry* selector;
    WORD    buf[buf_size+1];
    WORD*   addr;
    CELL    n;
    
    if((n = sukuna_get_next_word(&addr)) > 0) {
        sukuna_lower(addr,n,buf);
        if(n>1 && *buf==to_sc(':')) {
            selector = sukuna_find(buf,n,sdict);
            if(selector!=SNIL) {
                return (CELL)selector->Data;
            }
        }
    }
    currentTask->IN = addr;
    currentTask->cnt = currentTask->cnt + n;
    return SNIL;
}

void sukuna_warning_method(sukunaClass* cl, CELL selID, char* msg) {
    sukunaDicEntry* selector = (sukunaDicEntry*)(sp_to_cp(selID-XtOffset(sukunaDicEntry*,Xt)));
    if(is_warnlevel(1)) {
        sukuna_put_error_info();
        printf("Warning: %s ( ", msg);
        sukuna_puts((WORD*)&(selector->NameSize));
        printf(" ");
        sukuna_puts((WORD*)sp_to_cp(cl->Name));
        printf(" )\n");
    }
}

void sukuna_warning_method_not_found(sukunaClass* cl, CELL selID) {
    sukuna_warning_method(cl,selID,"method not found");
}
void sukuna_warning_method_not_implemented(sukunaClass* cl, CELL selID) {
    sukuna_warning_method(cl,selID,"method not implemented");
}

void sukuna_error_msg(char* msg, WORD* msg2, WORD* msg3) {
    sukuna_put_error_info();
    printf("Error: %s ( ", msg);
    sukuna_puts(msg2);
    printf(" ");
    sukuna_puts(msg3);
    printf(" )\n");
}

void sukuna_error_method_not_found(sukunaClass* cl, CELL selID) {
    sukunaDicEntry* selector = (sukunaDicEntry*)(sp_to_cp(selID-XtOffset(sukunaDicEntry*,Xt)));
    sukuna_error_msg("method not found",(WORD*)&(selector->NameSize),(WORD*)sp_to_cp(cl->Name));
}

void sukuna_error_mixin_not_found(sukunaClass* cl, CELL mx) {
    sukunaClass*    mix = (sukunaClass*)((CELL)sp_to_cp(mx)-cellWidth);
    sukuna_error_msg("mixin not found",(WORD*)sp_to_cp(mix->Name),(WORD*)sp_to_cp(cl->Name));
}

void sukuna_compile_method_call_inst(sukunaDicEntry* mt, sukunaClass* cl, CELL flags, CELL selfType) {
    CELL*           ofsp;
    CELL            offset;
    if(flags) {
        if(sukuna_is_stateless_mixin(mt->mixin)==SFALSE) {
            *(codep++) = _invoke_mixini;
            *(codep++) = mt->Xt;
            *(codep++) = mt->Data;
            *(codep++) = mt->mixin;
        }
        else {
            *(codep++) = _invoke_stless_mixini;
            *(codep++) = mt->Data;
            *(codep++) = mt->mixin;
        }
    }
    else {
        if(check_sp(mt->mixin)) {
            ofsp = sukuna_find_mixin_offset((sukunaClass*)mt->mixin,cl);
            if(ofsp!=SNIL) {
                offset = *ofsp;
                if(sukuna_is_stateless_mixin(mt->mixin)==SFALSE) {
                    *(codep++) = _invoke_mixed_virtuali;
                    *(codep++) = mt->Xt;
                    *(codep++) = (offset & offset_mask) + (CELL)(mt->Data);
                    *(codep++) = offset >> 16;
                }
                else {
                    *(codep++) = _invoke_virtuali;
                    *(codep++) = (offset & offset_mask) + (CELL)(mt->Data);
                }
            }
            else {
                if(is_warnlevel(1)) {
                    sukuna_put_error_info();
                    printf("Warning: illegal method call \n");
                }
                *(codep++) = _lit;
                *(codep++) = err_cantbind;
                *(codep++) = _abort;
            }
        }
        else {
            *(codep++) = _invoke_virtuali;
            *(codep++) = (CELL)mt->Data;
        }
    }
    sVariables->exactself.Body = selfType;
    sukuna_type_inference_check(mt);
    sVariables->exactself.Body = SNIL;
}
CELL sukuna_compile_method_call_instw() {
    CELL selfType   = pop_data;
    CELL flags      = to_ci(pop_data);
    sukunaClass*    cl = (sukunaClass*)((CELL)sp_to_cp(pop_data)-cellWidth);
    sukunaDicEntry* mt = (sukunaDicEntry*)(sp_to_cp(pop_data-XtOffset(sukunaDicEntry*,Xt)));

    sukuna_compile_method_call_inst(mt, cl, flags, selfType);
    return SNIL;
}

void sukuna_compile_method_call(CELL selID, CELL scl, CELL flags, CELL selfType) {
    sukunaClass*    cl;
    sukunaDicEntry* mt;
    sukunaClass*    reqClass = SNIL;
    
    cl = (sukunaClass*)((CELL*)sp_to_cp(scl)-1);
    mt = sukuna_find_method(selID, cl);
    if(mt!=SNIL) {
        sukuna_check_abstract(cl,mt);
        sukuna_compile_method_call_inst(mt, cl, flags, selfType);
    }
    else {
        if(flags!=0&&((mt = sukuna_find_reqmixin_method(selID,cl,&reqClass))!=SNIL)) {
            if(reqClass!=SNIL) {
                sukuna_check_abstract(reqClass,mt);
            }
            sukuna_compile_method_call_inst(mt, reqClass, (reqClass==SNIL), selfType);
        }
        else {
            sukuna_warning_method_not_found(cl,selID);
            *(codep++) = _invoke_latei;
            *(codep++) = selID;
            sukuna_failure_type_check();
        }
    }
}
CELL sukuna_compile_method_callw(void) {
    CELL selfType  = pop_data;
    CELL flags     = to_ci(pop_data);
    CELL scl       = pop_data;
    CELL selID     = pop_data;
    sukuna_compile_method_call(selID, scl, flags, selfType);
    return SNIL;
}

void sukuna_compile_localref(sukunaDicEntry* lfa) {
    CELL    selID;
    
    if((selID = sukuna_get_postfix_selector()) != SNIL) {
        sukuna_compile_method_call(selID,lfa->Data,lfa->Flags&refmix_mask,lfa->Data);
    }
    else {
        sukuna_literal_type_inference(lfa->Data);
    }
    return;
}

void sukuna_compile_ivarfetch(sukunaDicEntry* lfa) {
    CELL* ofsp;
    CELL offset;
    
    if((!check_sp(lfa->mixin))||(lfa->mixin == method_def)) {
        *(codep++) = _lit_ivar;
        *(codep++) = lfa->Data;
    }
    else {
        ofsp = sukuna_find_mixin_offset((sukunaClass*)lfa->mixin,(sukunaClass*)sp_to_cp(method_def-cellWidth));
        if(ofsp!=SNIL) {
            offset = *ofsp;
            *(codep++) = _lit_mixed_ivar;
            *(codep++) = offset >> 16;                                
            *(codep++) = lfa->Data;
        }
        else {
            if(is_warnlevel(1)) {
                sukuna_put_error_info();
                printf("Warning: illegal instance variable access \n");
            }
            *(codep++) = _lit;
            *(codep++) = err_cantbind;
            *(codep++) = _abort;
        }
    }
}
CELL sukuna_compile_ivarfetchw(void) {
    sukunaDicEntry* lfa = (sukunaDicEntry*)(sp_to_cp(pop_data-XtOffset(sukunaDicEntry*,Xt)));
    sukuna_compile_ivarfetch(lfa);
    return SNIL;
}

void sukuna_compile_ivar(sukunaDicEntry* lfa) {
    CELL    selID;
    
    sukuna_compile_ivarfetch(lfa);
    
    if(sVariables->quoteValue.Body == STRUE) goto normal;
    if(check_si(lfa->Xt)) goto normal;
    
    if((selID = sukuna_get_postfix_selector()) != SNIL) {
        sukuna_compile_method_call(selID,lfa->Xt,lfa->Flags&refmix_mask,lfa->Xt);
    }
    else {
        sukuna_literal_type_inference(lfa->Xt);
    }
    return;
normal:
    sukuna_literal_type_inference(lfa->Xt);
    if(sVariables->quoteValue.Body == STRUE) {
        *(codep++) = _call;
        sukuna_type_inference_check(lfa);
    }
    return;
}

CELL sukuna_to_receiver(sukunaDicEntry* mtp, CELL* clp, CELL* obp) {
    sukunaClass* mxp = (sukunaClass*)(mtp->mixin);
    CELL* ofsp;
    CELL offset;
    
    if(check_sp((CELL)mxp) && (sukuna_is_stateless_mixin((CELL)mxp)==SFALSE)) {
        ofsp = sukuna_find_mixin_offset(mxp, (sukunaClass*)clp);
        if(ofsp!=SNIL) {
            offset = *ofsp;
            obp = (CELL*)sp_to_cp(*((CELL*)((CELL)obp + (offset>>16))));
        }
    }
    return (CELL)obp;
}

void sukuna_compile_class_variable(sukunaDicEntry* lfa) {
    sukunaClass* theClass;
    sukunaDicEntry* theMethod;
    CELL objp;
    CELL selID;

    if(sVariables->quoteValue.Body == STRUE) goto normal;
    if(check_si(lfa->Xt)) goto normal;
    if(compile_state==SFALSE) goto normal;
    
    theClass = (sukunaClass*)((CELL*)sp_to_cp(lfa->Xt)-1);
    if((selID=sukuna_get_postfix_selector())!=SNIL) {
        theMethod = sukuna_find_method(selID,theClass);
        if(theMethod!=SNIL) {
            sukuna_check_abstract(theClass,theMethod);
            objp = sukuna_to_receiver(theMethod,(CELL*)theClass,(CELL*)sp_to_cp(lfa->Data));
            *(codep++) = _lit;
            *(codep++) = (CELL)cp_to_sp(objp);
            *(codep++) = theMethod->Xt;
            sVariables->exactself.Body = lfa->Xt;
            sukuna_type_inference_check(theMethod);
            sVariables->exactself.Body = SNIL;
            return;
        }
    }    
normal:
    if(compile_state==SFALSE) {
        push_data(lfa->Data);
        if(sVariables->quoteValue.Body == STRUE) sukuna_VM(_call);
    }
    else {
        *(codep++) = _lit;
        *(codep++) = lfa->Data;
        sukuna_literal_type_inference(lfa->Xt);
        if(sVariables->quoteValue.Body == STRUE) {
            *(codep++) = _call;
            sukuna_type_inference_check(lfa);
        }
    }        
    return;
}

void sukuna_compile_class_ivarfetch(sukunaDicEntry* lfa) {
    *(codep++) = _lit_at;
    *(codep++) = lfa->Data;
}

void sukuna_compile_class_refvar(sukunaDicEntry* lfa) {
    CELL    selID;
    
    if(compile_state!=SFALSE) {
        sukuna_compile_class_ivarfetch(lfa);
        if(sVariables->quoteValue.Body == STRUE) goto normal;
        if(check_si(lfa->Xt)) goto normal;
        if((selID = sukuna_get_postfix_selector()) != SNIL) {
            sukuna_compile_method_call(selID,lfa->Xt,lfa->Flags&refmix_mask,lfa->Xt);
        }
        else {
            sukuna_literal_type_inference(lfa->Xt);
        }
        return;
      normal:
        sukuna_literal_type_inference(lfa->Xt);
        if(sVariables->quoteValue.Body == STRUE) {
            *(codep++) = _call;
            sukuna_type_inference_check(lfa);
        }
        return;
    }
    else {
        push_data(*(sp_to_cp(lfa->Data)));
        if(sVariables->quoteValue.Body == STRUE) sukuna_VM(_call);
        return;
    }
}

void sukuna_compile_inner_function(sukunaDicEntry* lfa) {
    CELL xt;
    sukunaDicEntry* latest;
    sukunaClass* theClass;
    CELL offset;
    CELL* ofsp;

    xt = lfa->Xt;
    if(!check_si(xt)) xt = (CELL)cp_to_sp(&(lfa->Xt));
    if((lfa->Flags&immediate_mask)!=0) {
        /* it's IMMEDIATE ! */
        if(compile_state==STRUE) {
            sukuna_VM(xt);
            if(sVariables->quoteValue.Body == STRUE) sukuna_VM(_call);
            sukuna_type_inference_check(lfa);
        }
        else {
            sukuna_VM(xt);
            if(sVariables->quoteValue.Body == STRUE) sukuna_VM(_call);
        }
    }
    else {
        if(compile_state==SFALSE) {
        /* interpret */
            sukuna_VM(xt);
            if(sVariables->quoteValue.Body == STRUE) sukuna_VM(_call);
        }
        else {
        /* compile */
            if(lfa->mixin == method_def || lfa->mixin == to_si(0) || sukuna_is_stateless_mixin(lfa->mixin)!=SFALSE) {
                if(sVariables->quoteValue.Body == STRUE) {
                    *(codep++) = xt;
                    *(codep++) = _call;
                    sukuna_failure_type_check();
                }
                else {
                    *(codep++) = xt;
                    theClass = (sukunaClass*)sp_to_cp(method_def-cellWidth);
                    latest = (sukunaDicEntry*)sp_to_cp(theClass->ivarEntry);
                    if((((CELL)cp_to_sp(&(latest->Xt))) == xt) && (((latest->Flags)&ignore_mask) == 0) && (((latest->Flags)&compcurr_flag) != 0)) {
                    /* recursive call */
                        *(codep++) = _noop;
                        *(codep++) = _noop;
                    }
                    sVariables->exactself.Body = method_def;
                    sukuna_type_inference_check(lfa);
                    sVariables->exactself.Body = SNIL;
                }
            }
            else {
                ofsp = sukuna_find_mixin_offset((sukunaClass*)lfa->mixin,(sukunaClass*)sp_to_cp(method_def-cellWidth));
                if(ofsp!=SNIL) {
                    offset = *ofsp;
                    xt = *((CELL*)sp_to_cp(xt));
                    xt = xt-(cellWidth*5);
                    if(sVariables->quoteValue.Body == STRUE) {
                        *(codep++) = _lit_ivar;
                        *(codep++) = offset >> 16;
                        *(codep++) = xt;
                        *(codep++) = _call;
                        sukuna_failure_type_check();
                    }
                    else {
                        *(codep++) = _lit_ivar;
                        *(codep++) = offset >> 16;
                        *(codep++) = xt;
                        theClass = (sukunaClass*)sp_to_cp(method_def-cellWidth);
                        latest = (sukunaDicEntry*)sp_to_cp(theClass->ivarEntry);
                        sVariables->exactself.Body = method_def;
                        sukuna_type_inference_check(lfa);
                        sVariables->exactself.Body = SNIL;
                    }
                }
                else {
                    if(is_warnlevel(1)) {
                        sukuna_put_error_info();
                        printf("Warning: illegal function call \n");
                    }
                    *(codep++) = _lit;
                    *(codep++) = err_cantbind;
                    *(codep++) = _abort;                    
                }
            }
        }
    }
}

void sukuna_compile_ivarandref(sukunaDicEntry* lfa) {
    if(lfa->Flags&refivar_mask) {
        if(lfa->Flags&class_variable) {
            sukuna_compile_class_refvar(lfa);
        }
        else {
            sukuna_compile_ivar(lfa);
        }
    }
    else if(lfa->Flags&innerfunc_flag) {
        sukuna_compile_inner_function(lfa);
    }
    else if(lfa->Flags&class_variable) {
        sukuna_compile_class_variable(lfa);
    }
    else {
        sukuna_compile_ivar(lfa);
    }
}

void sukuna_create_selectors(WORD* addr, CELL n) {
    sukunaDicEntry* newEntry;
    CELL selID;
    WORD buf[buf_size+1];

    sukuna_lower(addr,n,buf+1);
    buf[0] = (WORD)to_sc(':');
    buf[n+1] = (WORD)to_sc(':');
    sukuna_create_new_label(buf+1, n+1);
    newEntry = (sukunaDicEntry*)sp_to_cp(sdict->LastEntry);
    selID = (CELL)cp_to_sp(&(newEntry->Xt));
    newEntry->Data = selID;
    newEntry->Flags |= immediate_mask;        
    sukuna_create_new_label(buf, n+1);
    newEntry = (sukunaDicEntry*)sp_to_cp(sdict->LastEntry);
    newEntry->Data = selID;
    newEntry->Flags |= immediate_mask;        
}

void sukuna_create_selectorw() {
    WORD* addr;
    CELL n;

    if((n = sukuna_get_next_word(&addr)) > 0) {
        sukuna_create_selectors(addr, n);
    }
}

CELL sukuna_str_to_value_type(WORD* addr, CELL n) {
    CELL res;
    
    if(sukuna_cstrcompare(addr,n,"any")!=SNIL) {
        return var_type;
    }
    if(sukuna_cstrcompare(addr,n,"fn")!=SNIL) {
        return func_type;
    }
    if(sukuna_cstrcompare(addr,n,"prim")!=SNIL) {
        return prim_type;
    }
    if(sukuna_cstrcompare(addr,n,"int")!=SNIL) {
        return int_type;
    }
    if(sukuna_cstrcompare(addr,n,"char")!=SNIL) {
        return char_type;
    }
    if(sukuna_cstrcompare(addr,n,"symb")!=SNIL) {
        return symb_type;
    }
    if(sukuna_cstrcompare(addr,n,"bool")!=SNIL) {
        return bool_type;
    }
    if(sukuna_cstrcompare(addr,n,"ptr")!=SNIL) {
        return ptr_type;
    }
    if(sukuna_cstrcompare(addr,n,"float")!=SNIL) {
        if(sVariables->floatClass.Body!=SNIL) {
            return sVariables->floatClass.Body;
        }
        else {
            return float_type;
        }
    }
    if(addr[0]==to_sc('\'')&&(sukuna_str_to_num((UWORD*)addr+1,n-1,&res)!=SNIL)) {
         return to_sv(to_ci(res));
    }
    if(addr[0]==to_sc('\'')&&(sukuna_cstrcompare(addr+1,n-1,"self")!=SNIL)) {
         return EXACTSELF;
    }
    return SNIL;
}

void sukuna_set_immediate_types(sukunaDicEntry* newEntry) {
    sukunaTypeList* tp;
    
    if(newEntry->TypeList==SNIL) return;
    tp = (sukunaTypeList*)sp_to_cp(newEntry->TypeList);
    tp->imm_arguments = tp->arguments;
    tp->imm_returnvalues = tp->returnvalues;
    tp->arguments = SNIL;
    tp->returnvalues = SNIL;
}

CELL sukuna_find_class_sp(WORD* addr, CELL n) {
    sukunaDicEntry* thisClassEntry;
    sukunaClass* thisClass;
    
    thisClassEntry = sukuna_find(addr,n,sdict);
    if(thisClassEntry!=SNIL) {
        thisClass = (sukunaClass*)sp_to_cp(thisClassEntry->Data);
        if(thisClass!=SNIL) {
            if(thisClass->tag==class_id) {
                return (CELL)cp_to_sp(&(thisClass->vtbl));;
            }
        }
    }
    return SNIL;
}

void sukuna_create_local_dicEntry(CELL xt, WORD* addr, CELL n) {
    WORD* clName = addr;
    WORD* ivName;
    CELL clen;
    CELL vlen;
    sukunaClass* thisClass;
    CELL clptr;
    WORD buf[buf_size+1];
    CELL valtype;
    
    vlen = sukuna_scan(addr,n,to_sc(':'), &ivName);
        
    if(vlen == 0 || vlen == n) {
        
        if((addr[0] == (WORD)to_sc('[')) && (addr[n-1] == (WORD)to_sc(']'))) {
            sukuna_create_dic_entry(xt,addr,n,normal_flag|localv_mask,sVariables->blockClass.Body,to_si(0),t_area,ldict);
        }
        else {
            sukuna_create_dic_entry(xt,addr,n,normal_flag|localv_mask,var_type,to_si(0),t_area,ldict);
        }
    }
    else {
        clen = n - vlen;
        if(vlen == 1) {
            vlen = sukuna_get_next_word(&ivName);
            sukuna_lower(ivName,vlen,buf);
            ivName = buf;
        }
        else {
            vlen = sukuna_skip(ivName,vlen,to_sc(':'), &ivName);
        }
        
        if((valtype=sukuna_str_to_value_type(clName,clen))!=SNIL) {
            sukuna_create_dic_entry(xt,ivName,vlen,normal_flag|localv_mask,valtype,to_si(0),t_area,ldict);
            goto exit;
        }
        
        if((clptr=sukuna_find_class_sp(clName,clen))!=SNIL) {
            thisClass = (sukunaClass*)sp_to_cp(clptr-cellWidth);
            if(thisClass->flag&mixin_mask) {
                sukuna_create_dic_entry(xt,ivName,vlen,normal_flag|localv_mask|refmix_mask,clptr,to_si(0),t_area,ldict);
            }
            else {
                sukuna_create_dic_entry(xt,ivName,vlen,normal_flag|localv_mask,clptr,to_si(0),t_area,ldict);
            }
            goto exit;
        }
        sukuna_create_dic_entry(xt,ivName,vlen,normal_flag|localv_mask,var_type,to_si(0),t_area,ldict);
    }
exit:
    return;
}

void sukuna_reset_type_stack() {
    if((argval_stack!=NULL)&&(retval_stack!=NULL)) {
        argval_sp = argval_stack;
        retval_sp = retval_stack;
        *argval_stack = 0;
        *retval_stack = 0;
    }
}

CELL sukuna_new_type_stack_frame(void) {
    av_push(failure_count);
    av_push(argval_stack);
    av_push(0);
    argval_stack = argval_sp;
    
    rv_push(retval_stack);
    rv_push(0);
    retval_stack = retval_sp;
    return SNIL;
}

CELL sukuna_dispose_type_stack_frame(void) {
    argval_sp = argval_stack;
    av_pop;
    argval_stack = (CELL*)av_pop;
    failure_count = av_pop;
    
    retval_sp = retval_stack;
    rv_pop;
    retval_stack = (CELL*)rv_pop;
    return SNIL;
}

CELL sukuna_swap_type_stack_frame(void) {
    CELL* ap;
    CELL* rp;
    CELL c;
    CELL d;
    CELL i;
    
    c = *argval_stack;
    push_data(failure_count);
    for(i = 0; i < c; ++i) {
        push_data(av_pop);
    } 
    push_data(av_pop);
    
    ap = (CELL*)av_pop;
    
    d = *ap;
    push_retn(av_pop);
    for(i = 0; i < d; ++i) {
        push_retn(av_pop);
    } 
    push_retn(av_pop);
    
    av_push(pop_data);
    for(i = 0; i < c; ++i) {
        av_push(pop_data);
    } 
    av_push(pop_data);
    av_push(ap);
    av_push(pop_retn);
    argval_stack = argval_sp;
    for(i = 0; i < d; ++i) {
        av_push(pop_retn);
    } 
    failure_count = pop_retn;
    
    c = *retval_stack;
    for(i = 0; i < c; ++i) {
        push_data(rv_pop);
    } 
    push_data(rv_pop);
    
    rp = (CELL*)rv_pop;
    
    d = *rp;
    for(i = 0; i < d; ++i) {
        push_retn(rv_pop);
    } 
    push_retn(rv_pop);
    
    rv_push(pop_data);
    for(i = 0; i < c; ++i) {
        rv_push(pop_data);
    } 
    rv_push(rp);
    rv_push(pop_retn);
    retval_stack = retval_sp;
    for(i = 0; i < d; ++i) {
        rv_push(pop_retn);
    }
    return SNIL; 
}

CELL sukuna_dup_type_stack_frame(void) {
    CELL* ap;
    CELL* rp;
    CELL c;
    CELL d;
    CELL i;

    c = *argval_stack;
    ap = argval_stack;
    av_push(failure_count);
    av_push(argval_stack);
    av_push(*argval_stack);
    argval_stack = argval_sp;
    for(i = 0; i < c; ++i) {
        av_push(*(--ap));
    } 

    d = *retval_stack;
    rp = retval_stack;
    rv_push(retval_stack);
    rv_push(*retval_stack);
    retval_stack = retval_sp;
    for(i = 0; i < d; ++i) {
        rv_push(*(++rp));
    }
    return SNIL;
}

void sukuna_literal_type_inference(CELL tp) {
    rv_push(tp);
    (*retval_stack)++;
}
CELL sukuna_literal_type_inferencew(void) {
    CELL tp = pop_data;
    sukuna_literal_type_inference(tp);
    return SNIL;
}

void sukuna_type_inference_check(sukunaDicEntry* pde) {
    sukunaTypeList* ptl;
    CELL* avp;
    CELL* rvp;
    CELL* np;
    CELL pcl;
    CELL ncl;
    CELL current_arg[64];
    int alen;
    int rlen;
    int plen;
    int c;
    int i,j;
    int w = 0;
    
    if(pde->Flags&immtype_flag) {
        return;
    }
    if(pde->TypeList==SNIL) {
        sukuna_failure_type_check();
        return;
    }
    ptl = (sukunaTypeList*)sp_to_cp(pde->TypeList);
    if(ptl->arguments==SNIL) {
        sukuna_failure_type_check();
        return;
    }
    if(ptl->returnvalues==SNIL) {
        sukuna_failure_type_check();
        return;
    }
    avp = (CELL*)sp_to_cp(ptl->arguments);
    rvp = (CELL*)sp_to_cp(ptl->returnvalues);
    alen = to_ci(*avp);
    rlen = to_ci(*rvp);
    np = avp + alen;
    plen = *retval_stack;
    if(plen>alen) {
        c = alen;
    } 
    else {
        c = *retval_stack;
    }
    j = alen;
    for(i=0;i<c;++i) {
        pcl = rv_pop;
        current_arg[--j] = pcl;
        (*retval_stack)--;
        ncl = *np--;
        if(sukuna_check_retvalue_type(pcl,ncl)==0) w = 1;
    }
    if(w==1) {
        sukuna_put_error_info();
        printf("Warning: argument type mismatch. \n");
    }
    if(alen>plen) {
        for(i=0;i<(alen-plen);++i) {
            ncl = *np--;
            current_arg[--j] = ncl;
            av_push(ncl);
            (*argval_stack)++;
        }
    }

    for(i=0;i<rlen;++i) {
        ncl = *(++rvp);
        if(check_sv(ncl)) {
            if(ncl==EXACTSELF) {
                rv_push(sVariables->exactself.Body);
                (*retval_stack)++;
            }
            else {
                j = to_cv(ncl)-1;
                rv_push(current_arg[j]);
                (*retval_stack)++;
            }
        }
        else {
            rv_push(ncl);
            (*retval_stack)++;
        }
    }
}
CELL sukuna_type_inference_checkw(void) {
    sukunaDicEntry* pde = (sukunaDicEntry*)sp_to_cp(pop_data);
    sukuna_type_inference_check(pde);
    return SNIL;
}

CELL sukuna_return_value_class(WORD* addr, CELL n) {
    WORD* clName = addr;
    WORD* ivName;
    CELL clen;
    CELL vlen;
    CELL clptr;
    WORD buf[buf_size+1];
    CELL result;
    CELL exactTypeFlag = 0;
    sukunaDicEntry* lfa;
    
    vlen = sukuna_scan(addr,n,to_sc(':'), &ivName);
        
    if(vlen == 0 || vlen == n) {
        if((addr[0] == (WORD)to_sc('[')) && (addr[n-1] == (WORD)to_sc(']'))) {
            return sVariables->blockClass.Body;
        }
        else {
            for( ; addr[n-1] == (WORD)to_sc('\'') ; -- n ) ;
            if(addr[0] == (WORD)to_sc('\'')) {
                exactTypeFlag = 1;
                addr++;
                --n;
            }
            lfa = sukuna_find(addr,n,ldict);
            if(lfa!=SNIL) {
                if(exactTypeFlag == 1) {
                    return to_sv(to_ci(lfa->Xt));
                }
                else {
                    return lfa->Data;
                }
            }
            else if (sukuna_cstrcompare(addr,n,"self")!=SNIL) {
                if(exactTypeFlag == 1) {
                    return EXACTSELF;
                }
                else if(method_def!=SNIL) {
                    return method_def;
                }
                else {
                    return var_type;
                }
            }
            else {
                return var_type;
            }
        }
    }
    else {
        clen = n - vlen;
        if(vlen == 1) {
            vlen = sukuna_get_next_word(&ivName);
            sukuna_lower(ivName,vlen,buf);
            ivName = buf;
        }
        else {
            vlen = sukuna_skip(ivName,vlen,to_sc(':'), &ivName);
        }
        if((result=sukuna_str_to_value_type(clName,clen))!=SNIL) return result;
        if((clptr=sukuna_find_class_sp(clName,clen))!=SNIL) return clptr;
    }
    return var_type;
}

sukunaTypeList* sukuna_arg_type_list() {
    WORD buf[buf_size+1];
    WORD* addr;
    CELL n;
    UCELL i = 0;
    UCELL j = 0;
    CELL state = 0;
    CELL result;
    sukunaTypeList* typep = SNIL;
    
    if(is_optlevel(1)) {
        typep = (sukunaTypeList*)sukuna_dallot(bytes_to_cells(sizeof(sukunaTypeList)));
        typep->arguments = (CELL)cp_to_sp(datap);
        typep->returnvalues = SNIL;
        *(datap++) = to_si(0);
    }
    n = sukuna_get_next_word(&addr);
    while(n>0) {
        if(n==1 && *addr==to_sc('|')) break;
        if(state == 0 && n==2 && addr[0]==to_sc('-') && addr[1]==to_sc('-')) { 
            state = 1;
            if(is_optlevel(1)) {
                typep->returnvalues = (CELL)cp_to_sp(datap);
                *(datap++) = to_si(0);
            }
        }
        else if(addr[n-1]==to_sc(':')) {
            if(state==0) i++;
            if(state==1) j++;
            if(is_optlevel(1)) {
                sukuna_lower(addr,n-1,buf);
                if((result=sukuna_str_to_value_type(buf,n-1))!=SNIL) {
                    *(datap++) = result;
                }
                else {
                    *(datap++) = sukuna_find_class_sp(buf,n-1);
                }
            }
        }
        n = sukuna_get_next_word(&addr);
    }
    if(is_optlevel(1)) {
        *((CELL*)sp_to_cp(typep->arguments)) = to_si(i);
        if(typep->returnvalues!=SNIL) *((CELL*)sp_to_cp(typep->returnvalues)) = to_si(j);
    }
    return typep;
}

sukunaTypeList* sukuna_setup_type_list(int flag) {
    WORD buf[buf_size+1];
    WORD* addr;
    CELL n;
    UCELL i = 0;
    UCELL j = 0;
    UCELL k = 0;
    CELL state = 0;
    sukunaTypeList* typep = SNIL;
    sukunaDicEntry* latestEntry;
    
    has_local = STRUE;
    ldict = sukuna_create_dictionary(t_area);
    if(is_optlevel(1)) {
        typep = (sukunaTypeList*)sukuna_dallot(bytes_to_cells(sizeof(sukunaTypeList)));
        typep->arguments = (CELL)cp_to_sp(datap);
        typep->returnvalues = SNIL;
        *(datap++) = to_si(0);
    }
    n = sukuna_get_next_word(&addr);
    while(n>0) {
        if(n==1 && *addr==to_sc('|')) break;
        if(state < 1 && n==1 && *addr==to_sc(':')) {
            state = 1;
        }
        
        else if(state < 2 && n==2 && *addr==to_sc('-') && *(addr+1)==to_sc('-')) { 
            state = 2;
            if(is_optlevel(1)&&(typep!=SNIL)) {
                typep->returnvalues = (CELL)cp_to_sp(datap);
                *(datap++) = to_si(0);
            }
        }
        
        else if(state == 0 && n==2 && *addr==to_sc('.') && *(addr+1)==to_sc('.')) { 
            ldict->LastEntry = SNIL;
            typep = SNIL;
       }
        
        else if(state < 2) {
            i++;
            sukuna_lower(addr,n,buf);
            sukuna_create_local_dicEntry(to_si(i),buf,n);
            if(state == 0 && is_optlevel(1)) {
                latestEntry = (sukunaDicEntry*)sp_to_cp(ldict->LastEntry);
                *(datap++) = (CELL)(latestEntry->Data);
            }
            if(state == 1) j++;
        }
        else if(state == 2 && is_optlevel(1) && (typep!=SNIL)) {
            if(n==1 && *addr==to_sc('/')) {
                typep->returnvalues = SNIL;
            }
            if(n==2 && *addr==to_sc('.') && *(addr+1)==to_sc('.')) {
                typep->returnvalues = SNIL;
            }
            else {
                k++;
                sukuna_lower(addr,n,buf);
                *(datap++) = sukuna_return_value_class(buf,n);
            }
        }
        n = sukuna_get_next_word(&addr);
    }
    i -= j;
    if(flag == 0) {
        *(codep++) = _lc_entry;
        *(codep++) = to_si((j<<16)|i);
    }
    sVariables->localEnv.Body = to_si((j<<16)|i);
    if(i > 0 && is_optlevel(1) && (typep!=SNIL)) {
        if(typep->arguments!=SNIL) {
            *((CELL*)sp_to_cp(typep->arguments)) = to_si(i);
        }
    }
    if(k > 0 && is_optlevel(1) && (typep!=SNIL)) {
        if(typep->returnvalues!=SNIL) {
            *((CELL*)sp_to_cp(typep->returnvalues)) = to_si(k);
        }
    }
    return typep;
}

void sukuna_read_type_list(sukunaDicEntry* lfa) {
    sukunaTypeList* result;
    if((result=sukuna_setup_type_list(0))!=SNIL) {
        if(lfa!=SNIL) {
            lfa->TypeList = (CELL)cp_to_sp(result);
        }
    }
}

void sukuna_define_locals() {
    sukunaDicEntry* latestEntry;
    latestEntry = (sukunaDicEntry*)sp_to_cp(sdict->LastEntry);
    sukuna_read_type_list(latestEntry);
}

void sukuna_define_method_locals() {
    sukunaClass* theClass = (sukunaClass*)sp_to_cp(method_def-cellWidth);
    sukunaDicEntry* latestEntry;
    latestEntry = (sukunaDicEntry*)sp_to_cp(theClass->LastEntry);
    sukuna_read_type_list(latestEntry);
}

void sukuna_define_innerfunc_locals() {
    sukunaClass* theClass = (sukunaClass*)sp_to_cp(method_def-cellWidth);
    sukunaDicEntry* latestEntry;
    latestEntry = (sukunaDicEntry*)sp_to_cp(theClass->ivarEntry);
    sukuna_read_type_list(latestEntry);
}

void sukuna_set_argtypes(sukunaDicEntry* xta) {
    sukunaTypeList* prep;
    sukunaTypeList* result;
    if((result=sukuna_arg_type_list())!=SNIL) {
        if(xta->TypeList!=SNIL) {
            prep = (sukunaTypeList*)sp_to_cp(xta->TypeList);
            result->imm_arguments = prep->imm_arguments;
            result->imm_returnvalues = prep->imm_returnvalues;
        }
        xta->TypeList = (CELL)cp_to_sp(result);
    }
}

void sukuna_push_argtype(sukunaDicEntry* latestEntry) {
    CELL* vfa = (CELL*)latestEntry->TypeList;
    CELL* ap;
    int arg_count;
    int i;
    
    if(vfa==SNIL) return;
    vfa = (CELL*)sp_to_cp(vfa);
    ap = (CELL*)sp_to_cp(*vfa);
    arg_count = to_ci(*ap);
    ap++;
    for(i = 0; i < arg_count; ++i) {
        rv_push(*ap++);
        (*retval_stack)++;
    }
}

CELL sukuna_set_argtypes_w(void) {
    sukunaDicEntry* xta = (sukunaDicEntry*)sp_to_cp(pop_data);
    sukuna_set_argtypes(xta);
    if(compile_state==STRUE) sukuna_push_argtype(xta);
    return SNIL;
}

CELL sukuna_set_latest_argtypes() {
    sukuna_set_argtypes((sukunaDicEntry*)sp_to_cp(sdict->LastEntry));
    if(compile_state==STRUE) sukuna_push_argtype((sukunaDicEntry*)sp_to_cp(sdict->LastEntry));
    return SNIL;
}

CELL sukuna_set_word_argtypes() {
    sukunaDicEntry* dicEntry;
    WORD buf[buf_size+1];
    WORD* addr;
    CELL n;

    if((n = sukuna_get_next_word(&addr)) > 0) {
        sukuna_lower(addr,n,buf);        
        if((dicEntry=sukuna_find(buf,n,sdict))==SNIL) return SNIL;
        if((n = sukuna_get_next_word(&addr)) > 0) {
            if(n==1 && *addr==to_sc('|')) {
                sukuna_set_argtypes(dicEntry);
                return (CELL)dicEntry;
            }
            else {
                return SNIL;
            }
        }
    }
    return SNIL;
}
CELL sukuna_set_word_argtypesw(void) {
    if(sukuna_set_word_argtypes()==SNIL) {
        push_data(to_si(err_notfound));
        return _abort;
    }
    return SNIL;
}

void sukuna_lcb() {
    sukunaDicEntry* newEntry;
    WORD buf[buf_size+1];
    WORD* addr;
    CELL n;

    if((n = sukuna_get_next_word(&addr)) > 0) {
        sukuna_lower(addr,n,buf);
        if((to_ci(warn_level)>=2)||(to_ci(warn_level)==-2)) {
            if(sukuna_find(buf,n,sdict)!=SNIL) {
                sukuna_put_error_info();
                printf("Warning: ");
                sukuna_type(buf,n);
                printf(" was redefined.\n");
            }
        }
        sukuna_define_new_word(buf,n);
        if((n = sukuna_get_next_word(&addr)) > 0) {
            if(n==1 && *addr==to_sc('|')) {
                newEntry = (sukunaDicEntry*)sp_to_cp(sdict->LastEntry);
                newEntry->Flags |= hasLocal_mask;
                sukuna_define_locals();
            }
            else {
                currentTask->IN = addr;
                currentTask->cnt = currentTask->cnt + n;
            }
        }
        sukuna_init_type_stack();
    }
}

CELL sukuna_search_super_class(CELL class_sp,sukunaClass* theClass) {
    CELL found = SNIL;
    sukunaMixin* supers;
    
    if(theClass==SNIL) return SNIL;
    supers = (sukunaMixin*)theClass->supers;
    if(supers==SNIL) return SNIL;
    supers = (sukunaMixin*)sp_to_cp(supers);
    if(theClass->flag&mixin_mask) supers += 2;
    while((supers->Klass)!=SNIL && found==SNIL) {
        if((CELL)(supers->Klass)==class_sp) found = class_sp;
        supers++;
    }
    return found;
}

CELL sukuna_find_super_class(CELL class_sp, sukunaClass* theClass) {
    sukunaClass* currentClass = theClass;
    CELL found = SNIL;
    
    found = sukuna_search_super_class(class_sp, currentClass);
    while(currentClass->ListLink!=SNIL && found==SNIL) {
        currentClass = (sukunaClass*)sp_to_cp(currentClass->ListLink);
        found = sukuna_search_super_class(class_sp, currentClass);
    }
    return found;
}

/***********************************************************************************
* file IO
***********************************************************************************/
CELL sukuna_fopen(WORD* fnp, CELL n, WORD* mop, CELL m) {
    unsigned char path[buf_size+1];
    unsigned char mode[buf_size+1];
    FILE* fp;

    sstr_to_cstr((UWORD*)fnp,path,n);
    sstr_to_cstr((UWORD*)mop,mode,m);
    if(fp = fopen((char*)path,(char*)mode)) {
        return (CELL)cp_to_sp(fp);
    }
    else {
        return SNIL;
    }
}
CELL sukuna_fopenw(void) {
    CELL tos = pop_data;
    CELL mad = pop_data;
    CELL n = pop_data;
    CELL Scratch = pop_data;
    push_data(sukuna_fopen((WORD*)sp_to_cp(Scratch), to_ci(n), (WORD*)sp_to_cp(mad), to_ci(tos)));
    return SNIL;
}

CELL sukuna_fclose(CELL fp) {
    if(fclose((FILE*)sp_to_cp(fp))) {
        return SEOF;
    }
    else {
        return SNIL;
    }
}
CELL sukuna_fclosew(void) {
    CELL tos = pop_data;
    push_data(sukuna_fclose(tos));
    return SNIL;
}

CELL sukuna_fgetc(CELL fp) {
    CELL c ;
    if((c = (CELL)fgetc((FILE*)sp_to_cp(fp)))!=EOF) {
        return to_si(c);
    }
    else {
        return SEOF;
    }
}
CELL sukuna_fgetcw(void) {
    CELL tos = pop_data;
    push_data(sukuna_fgetc(tos));
    return SNIL;
}

CELL sukuna_fgetwc(CELL fp) {
    CELL c ;
    UWORD d = 0;
    if((c = (CELL)fgetc((FILE*)sp_to_cp(fp)))==EOF) return SEOF;

    if ((c & 0x80) != 0x80) {
        d =  (UWORD)c;
    } 
    else if ((c & 0xE0) == 0xC0) {
        d = (UWORD)(c & 0x1F);
        if((c = (CELL)fgetc((FILE*)sp_to_cp(fp)))==EOF) return SEOF;
        d = (d << 6) | (c & 0x3F);
    }
    else if ((c & 0xF0) == 0xE0) {
        d = (UWORD)(c & 0x0F);
        if((c = (CELL)fgetc((FILE*)sp_to_cp(fp)))==EOF) return SEOF;
        d = (d << 6) | (c & 0x3F);
        if((c = (CELL)fgetc((FILE*)sp_to_cp(fp)))==EOF) return SEOF;
        d = (d << 6) | (c & 0x3F);
    }
    else if ((c & 0xF0) == 0xE0) {
        d = (UWORD)(c & 0x07);
        if((c = (CELL)fgetc((FILE*)sp_to_cp(fp)))==EOF) return SEOF;
        d = (d << 6) | (c & 0x3F);
        if((c = (CELL)fgetc((FILE*)sp_to_cp(fp)))==EOF) return SEOF;
        d = (d << 6) | (c & 0x3F);
        if((c = (CELL)fgetc((FILE*)sp_to_cp(fp)))==EOF) return SEOF;
        d = (d << 6) | (c & 0x3F);
    }
    return to_sc(d);
}
CELL sukuna_fgetwcw(void) {
    CELL tos = pop_data;
    push_data(sukuna_fgetwc(tos));
    return SNIL;
}

CELL sukuna_fgetsw(void) {
    CELL tos = pop_data;
    CELL blen = pop_data;
    CELL Scratch = pop_data;
    Scratch = (CELL)sukuna_fgets((WORD*)sp_to_cp(Scratch), to_ci(blen), (FILE*)sp_to_cp(tos));
    if(Scratch!=SNIL) { 
        push_data((CELL)cp_to_sp(Scratch)); 
    }
    else { 
        push_data(SNIL); 
    }
    return SNIL;
}

CELL sukuna_fputc(CELL c, CELL fp) {
    if((CELL)fputc((int)to_ci(c), (FILE*)sp_to_cp(fp))!=EOF) {
        return c;
    }
    else {
        return SEOF;
    }
}
CELL sukuna_fputcw(void) {
    CELL fp = pop_data;
    CELL c = pop_data;
    push_data(sukuna_fputc(c,fp));
    return SNIL;
}

CELL sukuna_fputwc(CELL c, CELL fp) {
    unsigned char s[5];
    unsigned char* p;
    p = skn_conv_utf8c(s, (UWORD)c);
    *p = '\0';
    p = s;
    while(*p!='\0') {
        if((CELL)fputc((int)*p++, (FILE*)sp_to_cp(fp))==EOF) return SEOF;
    }
    return c;
}
CELL sukuna_fputwcw(void) {
    CELL fp = pop_data;
    CELL c = pop_data;
    push_data(sukuna_fputwc(c,fp));
    return SNIL;
}

CELL sukuna_fputs(WORD* p, CELL n, FILE* fp) {
    unsigned char outputs[buf_size+1];
    CELL length;
    
    length = sstr_to_cstr((UWORD*)p,outputs,n);
    if(fputs((char*)outputs,fp)!=EOF) {
        return to_si(length);
    }
    else {
        return SEOF;
    }
}
CELL sukuna_fputsw(void) {
    CELL tos = pop_data;
    CELL n = pop_data;
    CELL Scratch = pop_data;
    push_data(sukuna_fputs((WORD*)sp_to_cp(Scratch), to_ci(n), (FILE*)sp_to_cp(tos)));
    return SNIL;
}

CELL sukuna_fflushw(void) {
    push_data((fflush((FILE*)sp_to_cp(pop_data)) ? SEOF : SNIL));
    return SNIL;
}

sukunaTask* sukuna_push_current_stream(FILE* newStream) {
    sukunaTask* newTask;
    
    newTask = (sukunaTask*)sukuna_tallot(bytes_to_cells(sizeof(sukunaTask)));
    newTask->ListLink = (CELL*)currentTask;
    newTask->sp = currentTask->sp;
    newTask->rp = currentTask->rp;
    newTask->op = currentTask->op;
    newTask->pc = &memory[0];
    newTask->IN_orig = (WORD*)sukuna_tallot(words_to_cells(buf_size));
    newTask->IN = newTask->IN_orig;
    newTask->cnt = 0;
    newTask->stream = newStream;
    newTask->flag = 0;
    newTask->linenum = 0;
    newTask->filepath = SNIL;
    currentTask = newTask;
    return newTask;
}

sukunaTask* sukuna_pop_current_stream() {
    sukunaTask* oldTask;
    
    if(currentTask->ListLink==SNIL) return currentTask;
    oldTask = (sukunaTask*)currentTask->ListLink;
    oldTask->sp = currentTask->sp;
    oldTask->rp = currentTask->rp;
    oldTask->op = currentTask->op;
    tempp = (CELL*)currentTask;
    currentTask = oldTask;
    return oldTask;
}

CELL sukuna_includeC(unsigned char* fileName) {
    FILE* newStream;
    CELL result;
    
    if((newStream=fopen((char*)fileName,"r"))==NULL) {
        sukuna_put_error_info();
        sukuna_cputs("File could not open.\n");
        result = SNIL;
    }
    else {
        sukuna_push_current_stream(newStream);
        result = STRUE;
    }
    return result;
}

CELL sukuna_include(WORD* addr, CELL n) {
    unsigned char buf[buf_size+1];
    CELL result;
    WORD* filepath;
    
    sstr_to_cstr((UWORD*)addr,buf,n);
    result = sukuna_includeC(buf);
    filepath = (WORD*)sukuna_tallot(n+2);
    sukuna_strcopy(addr, n, filepath);
    currentTask->filepath = filepath;
    return result;
}

CELL sukuna_includepw(void) {
    CELL n = to_ci(pop_data);
    WORD* addr = (WORD*)sp_to_cp(pop_data);
    sukuna_include(addr, n);
    return SNIL;
}

CELL sukuna_include_file(void) {
    WORD* addr;
    CELL n;

    if((n = sukuna_get_next_word(&addr)) > 0) {
        sukuna_include(addr,n);
    }
    return SNIL;
}

CELL sukuna_write_cell(FILE* fp, CELL n) {
	return (((fwrite((char*)&n,1,sizeof(CELL),fp))!=sizeof(CELL)) ? SNIL : STRUE);
}

CELL sukuna_read_cell(FILE* fp, CELL* p) {
	return (((fread((char*)p,1,sizeof(CELL),fp))!=sizeof(CELL)) ? SNIL : STRUE);
}

CELL sukuna_write_chunk(FILE* fp, CELL id, CELL* data, CELL size) {
	if(sukuna_write_cell(fp,id)==SNIL) return SNIL;
        return (((fwrite(data,1,size,fp))!=size) ? SNIL : STRUE);
}

CELL sukuna_saveC(unsigned char* fileName) {
    FILE* fp;
    CELL n;
    sukunaImageHeader header;
    
    if((fp=fopen((char*)fileName,"wb"))==NULL) {
        sukuna_cputs("File could not open.\n");
        return SNIL;
    }
    if(sukuna_write_cell(fp,creator_id)==SNIL) goto error;
    if(sukuna_write_cell(fp,header_id)==SNIL) goto error;
    header.cSize = codeSize;
    header.dSize = dataSize;
    header.sSize = stackSize;
    header.rSize = rStackSize;
    header.tSize = tempSize;
    header.sdict = (sukunaDictionary*)cp_to_sp(sdict);
    header.svar = (sukunaVariables*)cp_to_sp(sVariables);
    header.currentTask = (sukunaTask*)cp_to_sp(topTask);
    header.sukunap = (sukunaGlobals*)cp_to_sp(sukunap);
    header.cp = cp_to_sp(codep);
    header.dp = cp_to_sp(datap);
    if((n = fwrite((char*)&header,1,sizeof(sukunaImageHeader),fp))!=sizeof(sukunaImageHeader)) goto error;
    if(sukuna_write_cell(fp,memory_id)==SNIL) goto error;
    if((n = fwrite((char*)memory,1,memMax*cellWidth,fp))!=(memMax*cellWidth)) goto error;
    if(sukuna_write_cell(fp,end_id)==SNIL) goto error;
    
    fclose(fp);
    return STRUE;
error:
    fclose(fp);
    return SNIL;
}

CELL sukuna_save(WORD* addr, CELL n) {
    unsigned char buf[buf_size+1];
    sstr_to_cstr((UWORD*)addr,buf,n);
    return sukuna_saveC(buf);
}

CELL sukuna_savei(void) {
    WORD* addr;
    CELL n;
    
    if((n = sukuna_get_next_word(&addr)) > 0) {
        sukuna_save(addr,n);
    }
    return SNIL;
}

CELL sukuna_loadC(unsigned char* fileName) {
    FILE* fp;
    CELL id,n,res;
    sukunaImageHeader header;
    
    if((fp=fopen((char*)fileName,"rb"))==NULL) {
        sukuna_cputs("File could not open.\n");
        return SNIL;
    }
    if(sukuna_read_cell(fp,&id)==SNIL) goto error;
    if(id!=creator_id) goto error;

    res = 0;
    if(sukuna_read_cell(fp,&id)==SNIL) goto error;
    while(id!=end_id) {
        switch(id) {
            case header_id:
                n = fread((char*)&header,1,sizeof(sukunaImageHeader),fp);
                if(n!=sizeof(sukunaImageHeader)) goto error;
                codeSize = header.cSize;
                dataSize = header.dSize;
                stackSize = header.sSize;
                rStackSize = header.rSize;
                tempSize = header.tSize;
                memSize = codeSize+dataSize; 
                stackTop = memSize+stackSize-1;
                rStackTop = memSize+stackSize+rStackSize-1;
                tempTop = memSize+stackSize+rStackSize;
                memMax = memSize+stackSize+rStackSize+tempSize;
                if(memory!=NULL) free(memory);
                if((memory = (CELL*)malloc(memMax*cellWidth))==NULL) {
                    sukuna_cputs("initialize error.\n");
                    goto error;
                }
                memOffset = (CELL)&memory[0];
                res += 1;
                break;
            case memory_id:
                n = fread((char*)memory,1,memMax*cellWidth,fp);
                if(n!=(memMax*cellWidth)) goto error;
                sdict = (sukunaDictionary*)sp_to_cp(header.sdict);
                sVariables = (sukunaVariables*)sp_to_cp(header.svar);
                currentTask = (sukunaTask*)sp_to_cp(header.currentTask);
                sukunap = (sukunaGlobals*)sp_to_cp(header.sukunap);
                sukunap->cp = sp_to_cp(header.cp);
                sukunap->dp = sp_to_cp(header.dp);
                sukunap->tp = &memory[tempTop];
                currentTask->ListLink = SNIL;
                currentTask->sp = &memory[stackTop];
                currentTask->rp = &memory[rStackTop];
                currentTask->op = SNIL;
                currentTask->IN_orig = (WORD*)sukuna_tallot(words_to_cells(buf_size));
                currentTask->IN = currentTask->IN_orig;
                currentTask->cnt = 0;
                currentTask->stream = stdin;
                currentTask->flag = 0;
                push_data(0); push_data(0); push_data(0); push_data(0);
                push_retn(0); push_retn(0); push_retn(0); push_retn(0);
                res += 2;
                break;
        }
        if(sukuna_read_cell(fp,&id)==SNIL) goto error;
    }
    if(res!=3) goto error;

    sukuna_gc_finalize();
    lastHeap = SNIL;
    totalHeapSize = 0;
    topTask = currentTask;
    fclose(fp);
    return STRUE;
    
error:
    fclose(fp);
    return SNIL;
}

void sukuna_reset_stack(){
    currentTask->sp = &memory[stackTop];
    currentTask->rp = &memory[rStackTop];
    
    push_data(0); push_data(0); push_data(0); push_data(0);
    push_retn(0); push_retn(0); push_retn(0); push_retn(0);
}

void sukuna_close_streams() {
    while(currentTask->ListLink!=SNIL) {
        if(currentTask->flag == 0) fclose(currentTask->stream);
        currentTask = (sukunaTask*)currentTask->ListLink;
    }
}

void sukuna_reset_task() {
    sukuna_close_streams();
    currentTask = topTask;
    sukunap->tp = &memory[tempTop];
    currentTask->IN_orig = (WORD*)sukuna_tallot(words_to_cells(buf_size));
    currentTask->IN = currentTask->IN_orig;
    currentTask->cnt = 0;
    currentTask->stream = stdin;
    currentTask->pc = &memory[0];
    sukuna_reset_stack();
    compile_state = SFALSE;
    has_local = SFALSE;
    sukuna_reset_type_stack();
    failure_count = 0;
}

void sukuna_abort(CELL n) {
    switch(-n) {
        case 1:
            printf("aborted.\n");
            break;
        case 2:
            printf("nil pointer access.\n");
            break;
        case 3:
            printf("conditional error.\n");
            break;
        case 4:
            printf("execute compile time word while not compiling.\n");
            break;
        case 5:
            printf("mismatch in conditional.\n");
            break;
        case 6:
            printf("not a deferred word.\n");
            break;
        case 7:
            printf("stack underflow.\n");
            break;
        case 8:
            printf("comment mismatch.\n");
            break;
        case 9:
            printf("not found error.\n");
            break;
        case 10:
            printf("can't allocate memory.\n");
            break;
        case 11:
            printf("not a class object.\n");
            break;
        case 12:
            printf("not an object.\n");
            break;
        case 13:
            printf("not a method definition.\n");
            break;
        case 14:
            printf("not a selector.\n");
            break;
        case 15:
            printf("instance variable error.\n");
            break;
        case 16:
            printf("can't bind method.\n");
            break;
        case 17:
            printf("not a mixin definition.\n");
            break;
        case 18:
            printf("not a methodsFor definition.\n");
            break;
        case 19:
            printf("super class error.\n");
            break;
        case 20:
            printf("not implemented.\n");
            break;
        case 21:
            printf("immutable error.\n");
            break;
        case 22:
            printf("alias error.\n");
            break;
        case 24:
            printf("io error.\n");
            break;
        default:
            printf("Error: %ld\n",n);
            break;
    }
}

void sukuna_evaluate(WORD* addr, CELL n) {
    sukuna_push_current_stream(SNIL);
    currentTask->IN_orig = addr;
    currentTask->IN = addr;
    currentTask->cnt = n;    
    currentTask->flag = STRUE;
    currentTask->s_IN = cp_to_sp(addr);
}

void sukuna_evaluateC(unsigned char* str) {
    CELL old_byeflag;
    CELL old_taskflag;
    CELL old_quietflag;
    
    old_quietflag = gQuietFlag;
    old_taskflag = currentTask->flag;
    old_byeflag = byeFlag;
    byeFlag = SNIL;
    gQuietFlag = STRUE;
    sukuna_interpret(str);
    byeFlag = old_byeflag;
    currentTask->flag = old_taskflag;
    gQuietFlag = old_quietflag;
}

/***********************************************************************************
* float
***********************************************************************************/
sukunaFloat* sukuna_new_float_obj(double x) {
    sukunaFloat* newFloat;
    newFloat = (sukunaFloat*)sukuna_malloc(bytes_to_cells(sizeof(sukunaFloat)));
    newFloat->Klass = (sukunaClass*)sVariables->floatClass.Body;
    newFloat->Body = x;
    return newFloat;
}

sukunaFloat* sukuna_static_float_obj(double x) {
    sukunaFloat* newFloat;
    newFloat = (sukunaFloat*)sukuna_dallot(bytes_to_cells(sizeof(sukunaFloat)));
    newFloat->Klass = (sukunaClass*)sVariables->floatClass.Body;
    newFloat->Body = x;
    return newFloat;
}

CELL sukuna_str_to_float(UWORD* addr, CELL n, double* res) {
    char buf[buf_size+1];
    char* endp;
    sstr_to_cstr(addr,(unsigned char*)buf,n);
    *res = strtod(buf, &endp);
    if( buf == endp ) return SNIL;
    return STRUE;
}

CELL sukuna_float_to_str(UWORD* addr, CELL n, UWORD* fma, CELL flen, double x) {
    char fstr[buf_size+1];
    char rstr[buf_size+1];
    CELL result;
    
    sstr_to_cstr(fma,(unsigned char*)fstr,flen);
    result = sprintf(rstr,fstr,x);
    if((result==EOF)||(result>n)) return SNIL;
    return cstr_to_sstr((unsigned char*)rstr,addr,result);
}

void sukuna_float_dot(sukunaFloat* fobj) {
    printf("%f",fobj->Body);
}

/***********************************************************************************
* Etc.
***********************************************************************************/
CELL sukuna_get_env(void) {
    int tos = (int)to_ci(pop_data);
    UWORD* result = (UWORD*)sp_to_cp(pop_data);
    int n = (int)to_ci(pop_data);
    UWORD* keystr = (UWORD*)sp_to_cp(pop_data);
    unsigned char tmp[buf_size];
    char* env;
    sstr_to_cstr(keystr, tmp, n);
    if(env=getenv((char*)tmp)) {
        push_data(cp_to_sp(result));
        push_data((CELL)to_si(cstr_to_sstr((unsigned char*)env,result,(int)to_ci(tos))));
    }
    else {
        push_data(SNIL);
    }
    return SNIL;
}

CELL sukuna_curr_time(void) {
    push_data((CELL)to_si((CELL)time(NULL)));
    return SNIL;
}
CELL sukuna_sys_clock(void) {
    struct timeval tv;
    gettimeofday(&tv, NULL);
    push_data(to_si(tv.tv_sec));
    push_data(to_si(tv.tv_usec));
    return SNIL;
}

CELL sukuna_rand(void) {
    push_data(to_si((UCELL)rand()));
    return SNIL;
}

CELL sukuna_srand(void) {
    srand(to_ci(pop_data));
    return SNIL;
}

/***********************************************************************************
* Sukuna VM
***********************************************************************************/
void sukuna_VM(CELL xt) {
    CELL* stackp;
    CELL* rstackp;
    CELL* localsp = SNIL;
    CELL* objectp;
    CELL* p = &memory[0];
    CELL  topOfStack;  
    CELL  Scratch = 0;
    CELL  token = xt;
    CELL  result;

#ifdef __gcc__
    static void* primitives[numPrimitives] = {
        #define DEFINST(inst,wname,flags)  &&LABEL_halt, && LABEL##inst, 
        #include "primitives.h"
        &&LABEL_halt
        #undef DEFINST
    };
#endif                        
    
    load_registers;
    while(1) {
        DISPATCH 
            {
                CASE(_halt)     goto EXIT_VM;
                CASE(_exit)     p = sp_to_cp(r_pop); token = *p++; NEXT;
                CASE(_noop)     token = *p++; NEXT;
                CASE(_call)
                    if(check_sp(TOS)&&(TOS != SNIL)) {
                        if(*(sp_to_cp(TOS-cellWidth))==SNIL) {
                            r_push(cp_to_sp(p)); 
                            p = sp_to_cp(*(sp_to_cp(TOS))); 
                            drop; 
                            token = *p++;
                            NEXT;
                        }
                        else if((to_class_cp(TOS))->flag & direct_callable) {
                            r_push(cp_to_sp(p)); 
                            p = sp_to_cp(*(sp_to_cp(TOS))); 
                            drop; 
                            token = *p++;
                            NEXT;
                        }
                        else { token = *p++; NEXT; }
                    }
                    else { token = *p++; NEXT; }

                CASE(_execute)  
                    if(check_si(TOS)) { token = TOS; drop; NEXT; }
                    else { 
                        r_push(cp_to_sp(p)); 
                        p = sp_to_cp(*(sp_to_cp(TOS))); 
                        drop; 
                        token = *p++; 
                        NEXT; 
                    }

                CASE(_branch)   token = *p; p = (CELL*)((token)+((CELL)p)); token = *p++; NEXT;
                CASE(_fbranch) 
                    token = *p; p = is_false(TOS) ? (CELL*)((token)+((CELL)p)) : (p+1) ; drop;
                    token = *p++; NEXT;  
                CASE(_jmp1c)    p++; token = *p++; NEXT;
                CASE(_jump)  
                    if(check_sp(TOS)&&(TOS != SNIL)) {
                        if(*(sp_to_cp(TOS-cellWidth))==SNIL) {
                            p = sp_to_cp(*(sp_to_cp(TOS))); 
                            drop; 
                            token = *p++;
                            NEXT;
                        }
                        else if((to_class_cp(TOS))->flag & direct_callable) {
                            p = sp_to_cp(*(sp_to_cp(TOS))); 
                            drop; 
                            token = *p++;
                            NEXT;
                        }
                        else { token = *p++; NEXT; }
                    }
                    else { token = *p++; NEXT; }
                CASE(_abort)
                    sukuna_put_error_info();
                    Scratch = TOS; drop;
                    sukuna_reset_task();
                    p = &memory[0];
                    load_registers;
                    sukuna_abort(to_ci(Scratch));
                    token = *p++;
                    NEXT;
                CASE(_errorq)
                    Scratch = TOS; drop;
                    if(!(is_false(TOS))) { TOS = Scratch; NEXTINST(_abort); }
                    else{ drop; token = *p++; NEXT; }
                CASE(_throw)
                    drop;
                    sukuna_reset_task();
                    p = &memory[0];
                    load_registers;
                    NEXTINST(_halt);
                CASE(_dup)      push(TOS); token = *p++; NEXT;
                CASE(_swap)     Scratch = *(stackp); *(stackp) = TOS; TOS = Scratch; token = *p++; NEXT;
                CASE(_drop)     drop; token = *p++; NEXT;
                CASE(_over)     Scratch = *(stackp); push(TOS); TOS = Scratch; token = *p++; NEXT;
                CASE(_rot)  
                    Scratch = *(stackp+1); *(stackp+1) = *(stackp); *(stackp) = TOS; TOS = Scratch;
                    token = *p++; NEXT;
                CASE(_to_r)     r_push(TOS); drop; token = *p++; NEXT;
                CASE(_from_r)   push(TOS); TOS = r_pop ; token = *p++; NEXT;
                CASE(_r_drop)   r_drop ; token = *p++; NEXT;
                CASE(_r_fetch)  push(TOS); TOS = *(rstackp); token = *p++; NEXT;
                    
                CASE(_1minus)   TOS = check_si(TOS) ? TOS-2 : TOS-ptrd(TOS) ; token = *p++; NEXT;
                CASE(_1plus)    TOS = check_si(TOS) ? TOS+2 : TOS+ptrd(TOS) ; token = *p++; NEXT; 
                CASE(_plus)     ibop(+,"+"); NEXT;                    
                CASE(_minus)    ibop(-,"-"); NEXT;
                CASE(_times)    ibop(*,"*"); NEXT;
                CASE(_divide)   ibop(/,"/"); NEXT;
                CASE(_mod)      ibop(%,"mod"); NEXT;
                CASE(_lshift) 
                    TOS = check_si(*(stackp)) ? to_si(((UCELL)to_ci(pop)) << to_ci(TOS)) : ((UCELL)pop) << to_ci(TOS);
                    token = *p++; NEXT;
                CASE(_rshift) 
                    TOS = check_si(*(stackp)) ? to_si(((UCELL)to_ci(pop)) >> to_ci(TOS)) : ((UCELL)pop) >> to_ci(TOS);
                    token = *p++; NEXT;
                
                CASE(_lteqgt)  
                    Scratch = pop; TOS = (( Scratch < TOS ) ? to_si(-1) : (( Scratch == TOS ) ? to_si(0) : to_si(1)));
                    token = *p++; NEXT;
                CASE(_eq)   TOS = ( pop == TOS ) ? STRUE : SFALSE ; token = *p++; NEXT;
                CASE(_gt)   TOS = ( pop >  TOS ) ? STRUE : SFALSE ; token = *p++; NEXT;
                CASE(_lt)   TOS = ( pop <  TOS ) ? STRUE : SFALSE ; token = *p++; NEXT;
                CASE(_gteq) TOS = ( pop >= TOS ) ? STRUE : SFALSE ; token = *p++; NEXT;
                CASE(_lteq) TOS = ( pop <= TOS ) ? STRUE : SFALSE ; token = *p++; NEXT;
                CASE(_neq)  TOS = ( pop != TOS ) ? STRUE : SFALSE ; token = *p++; NEXT;
                CASE(_niltest) TOS = ( TOS == SNIL ) ? STRUE : SFALSE ; token = *p++; NEXT;
                CASE(_zeq)  TOS = ( TOS == to_si(0) ) ? STRUE : SFALSE ; token = *p++; NEXT;
                CASE(_b_and)    ibop(&,"and"); NEXT;
                CASE(_b_or)     ibop(|,"or"); NEXT;
                CASE(_b_xor)    ibop(^,"xor"); NEXT;
                CASE(_b_not)                
                    TOS = check_si(TOS) ? to_si(~to_ci(TOS)) : (check_sp(TOS) ? si_to_ptr(~TOS) : ((~to_cc(TOS))<<4) | (TOS&3));
                    token = *p++; NEXT;
                CASE(_l_and)    TOS = ((pop == STRUE) && (TOS == STRUE)) ? STRUE : SFALSE ; token = *p++; NEXT;
                CASE(_l_or)     TOS = ((pop == STRUE) || (TOS == STRUE)) ? STRUE : SFALSE ; token = *p++; NEXT;
                CASE(_l_not)    TOS = (TOS == STRUE) ? SFALSE : STRUE ; token = *p++; NEXT;
                CASE(_strue)    push(TOS); TOS = STRUE ; token = *p++; NEXT;
                CASE(_sfalse)   push(TOS); TOS = SFALSE; token = *p++; NEXT;
                CASE(_snil)     push(TOS); TOS = SNIL  ; token = *p++; NEXT;
                CASE(_lit)      push(TOS); TOS = *p++; token = *p++; NEXT;
                CASE(_lit_at)   push(TOS); TOS = (CELL)(*(sp_to_cp(*p++))); token = *p++; NEXT;                

                CASE(_fetch)    TOS = *(sp_to_cp(TOS)); token = *p++; NEXT;
                CASE(_store) 
                    if(TOS!=SNIL) { *(sp_to_cp(TOS)) = pop; drop; token = *p++; NEXT; }
                    else { TOS = to_si(err_nilpointer); NEXTINST(_abort); }
                CASE(_w_fetch)  TOS = *((WORD*)sp_to_cp(TOS)); token = *p++; NEXT;
                CASE(_w_store)  *((WORD*)sp_to_cp(TOS)) = (WORD)pop; drop; token = *p++; NEXT;

                CASE(_pstore) 
                    if(TOS!=SNIL) {
                        Scratch = (CELL)sp_to_cp(TOS); drop;
                        plusstore((CELL*)Scratch);
                        NEXT;
                    }
                    else { TOS = to_si(err_nilpointer); NEXTINST(_abort); }
                CASE(_gv_inc) 
                    if(TOS!=SNIL) { 
                        *(sp_to_cp(TOS)) = check_si(*(sp_to_cp(TOS))) ? *(sp_to_cp(TOS))+2 : *(sp_to_cp(TOS))+ptrd(*(sp_to_cp(TOS)));
                        drop; 
                        token = *p++;
                        NEXT;
                    } 
                    else { TOS = to_si(err_nilpointer); NEXTINST(_abort); }
                CASE(_gv_dec) 
                    if(TOS!=SNIL) { 
                        *(sp_to_cp(TOS)) = check_si(*(sp_to_cp(TOS))) ? *(sp_to_cp(TOS))-2 : *(sp_to_cp(TOS))-ptrd(*(sp_to_cp(TOS)));
                        drop;
                        token = *p++;
                        NEXT;
                    } 
                    else { TOS = to_si(err_nilpointer); NEXTINST(_abort); }

                CASE(_cell_plus)    TOS += cellWidth; token = *p++; NEXT;
                CASE(_cell_minus)   TOS -= cellWidth; token = *p++; NEXT;
                CASE(_cells_plus)   TOS = pop + to_ci(TOS)*cellWidth; token = *p++; NEXT;
                    
                CASE(_sp_fetch)     push(TOS); TOS = (CELL)cp_to_sp(stackp); token = *p++; NEXT;
                CASE(_sp_store)     stackp = (CELL*)sp_to_cp(TOS); drop; token = *p++; NEXT;
                CASE(_rp_fetch)     push(TOS); TOS = (CELL)cp_to_sp(rstackp); token = *p++; NEXT;
                CASE(_rp_store)     rstackp = (CELL*)sp_to_cp(TOS); drop; token = *p++; NEXT;
                CASE(_cp_fetch)     push(TOS); TOS = (CELL)cp_to_sp(codep); token = *p++; NEXT;
                CASE(_cp_store)     codep = (CELL*)sp_to_cp(TOS); drop; token = *p++; NEXT;
                CASE(_dp_fetch)     push(TOS); TOS = (CELL)cp_to_sp(datap); token = *p++; NEXT;
                CASE(_dp_store)     datap = (CELL*)sp_to_cp(TOS); drop; token = *p++; NEXT;
                CASE(_tp_fetch)     push(TOS); TOS = (CELL)cp_to_sp(tempp); token = *p++; NEXT;
                CASE(_tp_store)     tempp = (CELL*)sp_to_cp(TOS); drop; token = *p++; NEXT;
                CASE(_lp_fetch)     push(TOS); TOS = (CELL)cp_to_sp(localsp); token = *p++; NEXT;
                CASE(_lp_store)     localsp = (CELL*)sp_to_cp(TOS); drop; token = *p++; NEXT;
                CASE(_op_fetch)     push(TOS); TOS = (CELL)cp_to_sp(objectp); token = *p++; NEXT;
                CASE(_op_store)     objectp = (CELL*)sp_to_cp(TOS); drop; token = *p++; NEXT;
                CASE(_in_fetch) 
                    push(TOS); TOS = (CELL)cp_to_sp(currentTask->IN);
                    push(TOS); TOS = (CELL)to_si(currentTask->cnt); token = *p++; NEXT;
                CASE(_in_store) 
                    currentTask->cnt = to_ci(TOS); drop; 
                    currentTask->IN = (WORD*)sp_to_cp(TOS); drop; token = *p++; NEXT;
                CASE(_failcountat)  push(TOS); TOS = (CELL)to_si(failure_count); token = *p++; NEXT;
                CASE(_failcountst)  failure_count = (CELL)to_ci(TOS); drop; token = *p++; NEXT;
                
                CASE(_c_base)   push(TOS); TOS = (CELL)cp_to_sp(&memory[0]); token = *p++; NEXT;
                CASE(_d_base)   push(TOS); TOS = (CELL)cp_to_sp(&memory[dataTop]); token = *p++; NEXT;
                CASE(_t_base)   push(TOS); TOS = (CELL)cp_to_sp(&memory[tempTop]); token = *p++; NEXT;
                CASE(_s_base)   push(TOS); TOS = (CELL)cp_to_sp(&memory[stackTop]); token = *p++; NEXT;
                CASE(_r_base)   push(TOS); TOS = (CELL)cp_to_sp(&memory[rStackTop]); token = *p++; NEXT;
                    
                CASE(_p_to_i)   TOS = ptr_to_si(TOS); token = *p++; NEXT;
                CASE(_i_to_p)   TOS = si_to_ptr(TOS); token = *p++; NEXT;
                CASE(_n_to_i)   TOS = num_to_si(TOS); token = *p++; NEXT;
                CASE(_i_to_n)   TOS = si_to_num(TOS); token = *p++; NEXT;
                CASE(_c_to_i)   TOS = to_si(to_cc(TOS)); token = *p++; NEXT;
                CASE(_i_to_c)   TOS = to_sc(to_ci(TOS)); token = *p++; NEXT;

                CASE(_intq)     TOS = check_si(TOS) ? STRUE : SFALSE ; token = *p++; NEXT;
                CASE(_ptrq)     TOS = check_sp(TOS) ? STRUE : SFALSE ; token = *p++; NEXT;
                CASE(_charq)    TOS = check_sc(TOS) ? STRUE : SFALSE ; token = *p++; NEXT;
                CASE(_symbq)    TOS = check_sb(TOS) ? STRUE : SFALSE ; token = *p++; NEXT;
                CASE(_vartq)    TOS = check_sv(TOS) ? STRUE : SFALSE ; token = *p++; NEXT;
                
                CASE(_check_ptr) TOS = (CELL)to_si(check_ptr(sp_to_cp(TOS))); token = *p++; NEXT;
                    
                CASE(_sstr_to_si) 
                    TOS = (sukuna_str_to_num((UWORD*)sp_to_cp(pop),to_ci(TOS),&result)!=SNIL ? result : SNIL);
                    token = *p++;
                    NEXT;
                CASE(_int_to_sstr) {
                        UWORD* fma = (UWORD*)sp_to_cp(pop);
                        CELL n = to_ci(pop);
                        Scratch = pop;
                        TOS = to_si(sukuna_num_to_str((UWORD*)sp_to_cp(Scratch),n,fma,to_ci(TOS),pop));
                        push(Scratch);
                        token = *p++;
                        NEXT;
                    }
                CASE(_cstoss)  
                    Scratch = pop;
                    TOS = to_si(cstr_to_sstr((unsigned char*)sp_to_cp(Scratch), (UWORD*)sp_to_cp(pop), (CELL)to_ci(TOS)));
                    token = *p++; NEXT;
                CASE(_sstr_cmp) {
                        CELL n2 = to_ci(TOS); 
                        UWORD* addr2 = (UWORD*)sp_to_cp(pop);
                        CELL n  = to_ci(pop);
                        UWORD* addr  = (UWORD*)sp_to_cp(pop);
                        TOS = to_si(sukuna_sstr_compare(addr,n,addr2,n2));
                        token = *p++;
                        NEXT;
                    }
                CASE(_sstr_cmpi) {
                        CELL n2 = to_ci(TOS); 
                        UWORD* addr2 = (UWORD*)sp_to_cp(pop);
                        CELL n  = to_ci(pop);
                        UWORD* addr  = (UWORD*)sp_to_cp(pop);
                        TOS = to_si(sukuna_sstr_comparei(addr,n,addr2,n2));
                        token = *p++;
                        NEXT;
                    }

                CASE(_mem_copy) 
                    Scratch = to_ci(TOS); drop;
                    sukuna_memcopy(sp_to_cp(pop),sp_to_cp(TOS),Scratch); drop; 
                    token = *p++; NEXT;
                CASE(_mem_move) 
                    Scratch = to_ci(TOS); drop;
                    sukuna_memmove((WORD*)sp_to_cp(pop),(WORD*)sp_to_cp(TOS),Scratch); drop; 
                    token = *p++; NEXT;
                
                CASE(_cr)       sukuna_cr(); token = *p++; NEXT;
                CASE(_emit)     sukuna_emit((UWORD)TOS); drop; token = *p++; NEXT;
                CASE(_dot)      sukuna_dot(TOS); drop; token = *p++; NEXT;
                CASE(_dot_hex)  sukuna_dot_hex(TOS); drop; token = *p++; NEXT;
                CASE(_nemit)    sukuna_debug_nemit(TOS); drop; token = *p++; NEXT;
                CASE(_type)     sukuna_type((WORD*)sp_to_cp(pop),to_ci(TOS)); drop; token = *p++; NEXT;
                    
                CASE(_context)      push(TOS); TOS = (CELL)cp_to_sp(&(sdict->LastEntry)); token = *p++; NEXT;
                CASE(_lc_context)   push(TOS); TOS = (CELL)cp_to_sp(&(ldict->LastEntry)); token = *p++; NEXT;
                
                CASE(_nextline) 
                        if(sukuna_fgets(currentTask->IN_orig,buf_size,currentTask->stream)==SNIL) {
                            push(TOS);
                            TOS = SFALSE;
                            token = *p++; 
                            NEXT;
                        }
                        else {
                            currentTask->IN = currentTask->IN_orig;
                            currentTask->cnt = to_addr_n(currentTask->IN);
                            currentTask->linenum++;
                            push(TOS);
                            TOS = STRUE;
                            token = *p++; 
                            NEXT;
                        }
                CASE(_parse_w) 
                    push(TOS); Scratch = sukuna_get_next_word((WORD**)&result); 
                    push(cp_to_sp(result)); TOS = to_si(Scratch); token = *p++; NEXT;
                CASE(_parse_ww)  
                    push(TOS); Scratch = sukuna_get_next_wordw((WORD**)&result); 
                    push(cp_to_sp(result)); TOS = to_si(Scratch); token = *p++; NEXT;
                CASE(_find)     TOS = sukuna_findw((WORD*)sp_to_cp(pop),to_ci(TOS),sdict); token = *p++; NEXT;
                CASE(_skip)     sukuna_skipw(TOS); drop; token = *p++; NEXT;
                CASE(_scan)     sukuna_scanw(TOS); drop; token = *p++; NEXT;
                CASE(_scanw)    sukuna_scanww(pop,TOS); drop; token = *p++; NEXT;
                CASE(_to_xt)    TOS = sukuna_to_xt(TOS); token = *p++; NEXT;
                        
                CASE(_create) sukuna_create(); token = *p++; NEXT;
                CASE(_createp) 
                    sukuna_createp((WORD*)sp_to_cp(pop),to_ci(TOS)); drop;
                    token = *p++; NEXT;
                CASE(_create_entry)  
                    Scratch = pop;
                    sukuna_create_dic_entrydw(TOS,(WORD*)sp_to_cp(pop),to_ci(Scratch)); drop;
                    token = *p++; NEXT;

                CASE(_lcb)      sukuna_lcb() ; token = *p++; NEXT;
                CASE(_rcb)      sukuna_rcbw(); token = *p++; NEXT;
                CASE(_rcbp)     sukuna_rcb() ; token = *p++; NEXT;
                CASE(_immedip)  sukuna_set_immediate_types((sukunaDicEntry*)sp_to_cp(TOS)); drop; token = *p++; NEXT;

                CASE(_save_dstack) {
                        CELL* p;
                        Scratch = to_ci(pop);
                        p = (CELL*)sp_to_cp(TOS);
                        p += Scratch;
                        while(Scratch-- > 0) { *(--p) = pop; }
                        drop;
                    }
                    token = *p++;
                    NEXT;
                CASE(_load_dstack) {
                        CELL* p;
                        Scratch = to_ci(TOS);
                        p = (CELL*)sp_to_cp(pop);
                        while(Scratch-- > 0) { push(*p++); }
                    }
                    token = *p++;
                    NEXT;
                CASE(_lc_entry) {
                        UCELL i = to_ci(*p++);
                        UCELL j = i >> 16;
                        CELL* lp;
                        i &= 0xFFFF;
                        r_push(cp_to_sp(localsp)); localsp = rstackp;
                        rstackp -= (i+j); lp = rstackp;
                        push(TOS);
                        while(j-- > 0) { *lp++ = to_si(0); }                        
                        while(i-- > 0) { *lp++ = pop; }
                        drop;
                    }
                    token = *p++;
                    NEXT;
                CASE(_lc_exit) 
                    if(isRstackPtr(localsp)) {
                        rstackp = localsp; localsp = (CELL*)sp_to_cp(r_pop);
                    }
                    p = sp_to_cp(r_pop);
                    token = *p++; 
                    NEXT;
                CASE(_dispose_locf) 
                    if(isRstackPtr(localsp)) {
                        rstackp = localsp; localsp = (CELL*)sp_to_cp(r_pop);
                    }
                    token = *p++; 
                    NEXT;
                CASE(_lc_fetch)  push(TOS); TOS = *(localsp-(CELL)to_ci(*p++)); token = *p++; NEXT;
                CASE(_lc_store)  *(localsp-(CELL)to_ci(*p++)) = TOS; drop; token = *p++; NEXT;
                CASE(_lc_pstore)  {
                        Scratch = (CELL)(localsp-(CELL)to_ci(*p++));
                        plusstore((CELL*)Scratch);
                        NEXT;
                    }
                    NEXT;
                CASE(_lc_inc) 
                    Scratch = (CELL)(localsp-(CELL)to_ci(*p++));
                    *((CELL*)Scratch) = check_si(*((CELL*)Scratch)) ? *((CELL*)Scratch)+2 : *((CELL*)Scratch)+ptrd(*((CELL*)Scratch));
                    token = *p++; NEXT;
                CASE(_lc_dec) 
                    Scratch = (CELL)(localsp-(CELL)to_ci(*p++));
                    *((CELL*)Scratch) = check_si(*((CELL*)Scratch)) ? *((CELL*)Scratch)-2 : *((CELL*)Scratch)-ptrd(*((CELL*)Scratch));
                    token = *p++; NEXT;
                CASE(_cp_lstore) 
                    if(sukuna_compile_lstore(_lc_store)!=SNIL) { token = *p++; NEXT; }
                    else { TOS = to_si(err_pairs); NEXTINST(_abort); }
                CASE(_cp_lpstore) 
                    if(sukuna_compile_lstore(_lc_pstore)!=SNIL) { token = *p++; NEXT; }
                    else { TOS = to_si(err_pairs); NEXTINST(_abort); }
                CASE(_cp_lentry) 
                    if(TOS!=SNIL) { 
                        sukuna_read_type_list((sukunaDicEntry*)sp_to_cp(TOS)); drop; token = *p++;
                        NEXT;
                    }
                    else {
                        sukuna_read_type_list((sukunaDicEntry*)SNIL); drop; token = *p++;
                        NEXT; 
                    }
                CASE(_lc_find)  TOS = sukuna_findw((WORD*)sp_to_cp(pop),to_ci(TOS),ldict); token = *p++; NEXT;
                CASE(_add_locv) {
                        WORD buf[buf_size+1];
                        Scratch = TOS; drop;
                        sukuna_lower((WORD*)sp_to_cp(pop),to_ci(TOS),buf);
                        sukuna_create_local_dicEntry(Scratch,buf,to_ci(TOS)); drop;
                        token = *p++; 
                        NEXT;
                    }

                CASE(_lc_fetch1)  push(TOS); TOS = *(localsp-1); token = *p++; NEXT;
                CASE(_lc_fetch2)  push(TOS); TOS = *(localsp-2); token = *p++; NEXT;
                CASE(_lc_fetch3)  push(TOS); TOS = *(localsp-3); token = *p++; NEXT;
                CASE(_lc_fetch4)  push(TOS); TOS = *(localsp-4); token = *p++; NEXT;
                CASE(_lc_fetch5)  push(TOS); TOS = *(localsp-5); token = *p++; NEXT;
                CASE(_lc_fetch6)  push(TOS); TOS = *(localsp-6); token = *p++; NEXT;
                CASE(_lc_fetch7)  push(TOS); TOS = *(localsp-7); token = *p++; NEXT;
                CASE(_lc_fetch8)  push(TOS); TOS = *(localsp-8); token = *p++; NEXT;

                CASE(_lbp) 
                    token = *p; Scratch = ((token)+((CELL)p));
                    push(TOS); ++p; TOS = (CELL)cp_to_sp(++p); p += 3;
                    *((CELL*)sp_to_cp(*p)) = (CELL)cp_to_sp(localsp);
                    *((CELL*)sp_to_cp(*p)+2) = (CELL)cp_to_sp(objectp);
                    p = (CELL*)Scratch; token = *p++;
                    NEXT;
                CASE(_bl_entry) 
                    r_push(cp_to_sp(objectp)); r_push(cp_to_sp(localsp));
                    localsp = (CELL*)sp_to_cp(*((CELL*)sp_to_cp(*p)));
                    objectp = (CELL*)sp_to_cp(*((CELL*)sp_to_cp(*p++)+2));
                    token = *p++;
                    NEXT;
                CASE(_bl_exit)  localsp = (CELL*)sp_to_cp(r_pop); objectp = (CELL*)sp_to_cp(r_pop); p = sp_to_cp(r_pop); token = *p++; NEXT;
                CASE(_gen_blk) 
                    save_registers;                
                    Scratch = (CELL)cp_to_sp(sukuna_generate_block((CELL*)sp_to_cp(TOS)));
                    load_registers;
                    TOS = Scratch;
                    token = *p++; NEXT;
                CASE(_to_closure) 
                    save_registers;                
                    Scratch = (CELL)cp_to_sp(sukuna_block_to_closure((CELL*)sp_to_cp(TOS)));
                    load_registers;
                    TOS = Scratch;
                    token = *p++; NEXT;
                CASE(_lp_lc_store)  *((CELL*)sp_to_cp(TOS)-(CELL)to_ci(*p++)) = pop; drop; token = *p++; NEXT;

                CASE(_create_clh)
                    save_registers;
                    Scratch = (CELL)cp_to_sp((CELL)sukuna_create_class_header(to_ci(TOS))+cellWidth);
                    load_registers;
                    TOS = Scratch;
                    token = *p++; NEXT;
                CASE(_inherit_sup)  
                    sukuna_inherit_class((sukunaClass*)sp_to_cp(pop-cellWidth),(sukunaClass*)sp_to_cp(TOS-cellWidth));
                    drop;
                    token = *p++; NEXT;
                CASE(_create_vtbl)  {
                        sukunaClass* theClass = (sukunaClass*)sp_to_cp(SOS-cellWidth);
                        save_registers;
                        sukuna_create_vtbl(theClass,to_ci(TOS));
                        load_registers;
                        drop; 
                        drop;
                        token = *p++;
                        NEXT;
                    }
                CASE(_update_vtbl)  {
                        sukunaClass* theClass = (sukunaClass*)sp_to_cp(TOS-cellWidth);
                        sukuna_update_vtbl(theClass);
                        drop;
                        token = *p++;
                        NEXT;
                    }
                CASE(_create_selr)  sukuna_create_selectorw(); token = *p++; NEXT;
                CASE(_create_sels)  sukuna_create_selectors((WORD*)sp_to_cp(pop),to_ci(TOS)); drop; token = *p++; NEXT;
                CASE(_create_ivar)  {
                        sukunaClass* theClass = (sukunaClass*)sp_to_cp(pop-cellWidth);
                        CELL n = to_ci(pop);
                        WORD* addr = (WORD*)sp_to_cp(pop);
                        sukunaClass* ivClass = (sukunaClass*)pop;
                        sukuna_create_ivar_entryw(theClass,ivClass,addr,n,to_ci(TOS)*cellWidth);
                        drop;
                        token = *p++;
                        NEXT;
                    }
                CASE(_create_mth)  {
                        sukunaClass* theClass = (sukunaClass*)sp_to_cp(pop-cellWidth);
                        CELL xta = pop;
                        CELL selid = pop;
                        sukuna_create_method_entryw(theClass,xta,selid,to_ci(TOS)*cellWidth);
                        drop;
                        token = *p++;
                        NEXT;
                    }
                CASE(_mth_entry)    r_push(cp_to_sp(objectp)); objectp = sp_to_cp(TOS); drop; token = *p++; NEXT;
                CASE(_mth_exit)     objectp = sp_to_cp(r_pop); NEXTINST(_exit);
                    
                CASE(_find_method)  
                    TOS = (CELL)sukuna_find_methodw(pop,(sukunaClass*)sp_to_cp(TOS-cellWidth)); token = *p++; NEXT;
                CASE(_find_self_m)  
                    TOS = (CELL)sukuna_find_self_methodw(pop,(sukunaClass*)sp_to_cp(TOS-cellWidth)); token = *p++; NEXT;
                CASE(_find_ivar) 
                    Scratch = to_ci(pop);
                    if(TOS!=SNIL) { 
                        TOS = (CELL)sukuna_find_ivarw((WORD*)sp_to_cp(pop),Scratch,(sukunaClass*)sp_to_cp(TOS-cellWidth));
                        token = *p++; NEXT;
                    }
                    else { drop; TOS = SNIL; token = *p++; NEXT; }
                CASE(_find_mixin_super)  {
                        CELL* clp;
                        CELL vofs;
                        clp = to_cclass(cp_to_sp(objectp));
                        TOS = (CELL)sukuna_find_mixin_super((sukunaClass*)TOS,(sukunaClass*)(clp-1),(CELL)cp_to_sp(objectp),(CELL)get_selfp, pop, &vofs);
                        if (TOS == SNIL) {
                            TOS = to_si(err_notfound);
                            NEXTINST(_abort);
                        }
                        else {
                            push(vofs);
                            token = *p++;
                            NEXT;
                        }
                    }
                CASE(_find_mixin_offs) 
                    Scratch = (CELL)sukuna_find_mixin_offset((sukunaClass*)pop,(sukunaClass*)sp_to_cp(TOS-cellWidth));
                    if(Scratch!=SNIL) {
                        TOS = *((CELL*)Scratch);
                        token = *p++; NEXT;
                    }
                    else {
                        TOS = SFALSE;
                        token = *p++; NEXT;
                    }
                CASE(_find_super_mth)  
                    TOS = (CELL)sukuna_find_super_methodw(pop,(sukunaClass*)sp_to_cp(TOS-cellWidth)); token = *p++; NEXT;
                CASE(_find_superc)  
                    TOS = sukuna_find_super_class(pop,(sukunaClass*)sp_to_cp(TOS-cellWidth)); token = *p++; NEXT;

                CASE(_selfpp)  push(TOS); TOS = (CELL)cp_to_sp(get_selfp); token = *p++; NEXT;
                CASE(_to_class)  TOS = (CELL)cp_to_sp(to_cclass(TOS)); token = *p++; NEXT;
                CASE(_to_receiver) 
                    Scratch = pop;
                    TOS = (CELL)cp_to_sp(sukuna_to_receiver((sukunaDicEntry*)sp_to_cp(pop-XtOffset(sukunaDicEntry*,Xt)),sp_to_cp(Scratch-cellWidth),sp_to_cp(TOS)));
                    token = *p++;
                    NEXT;

                CASE(_invoke_late)  {
                        CELL* obp;
                        sukunaDicEntry* mtp;
                        push_if_primitive_obj2;
                        obp = (CELL*)pop;
                        Scratch = (CELL)(to_cclass(obp)-1);
                        obp = sp_to_cp(obp);
                        mtp = sukuna_find_method(TOS,(sukunaClass*)Scratch);
                        if(mtp!=SNIL) {
                            TOS = (CELL)cp_to_sp(sukuna_to_receiver(mtp,(CELL*)Scratch,obp));
                            r_push(cp_to_sp(p)); 
                            p = sp_to_cp(*(sp_to_cp(mtp->Xt)));
                            token = *p++;
                            NEXT;
                        }
                        else {
                            sukuna_error_method_not_found((sukunaClass*)Scratch,TOS);
                            NEXTINST(_throw);
                        }
                    }
                CASE(_invoke_latei)  {
                        sukunaDicEntry* mtp;
                        CELL selid;
                        push_if_primitive_obj;
                        Scratch = (CELL)(to_cclass(TOS)-1);
                        TOS = (CELL)sp_to_cp(TOS);
                        selid = *p++;
                        mtp = sukuna_find_method(selid,(sukunaClass*)Scratch);
                        if(mtp!=SNIL) {
                            TOS = (CELL)cp_to_sp(sukuna_to_receiver(mtp,(CELL*)Scratch,(CELL*)TOS));
                            r_push(cp_to_sp(p)); 
                            p = sp_to_cp(*(sp_to_cp(mtp->Xt)));
                            token = *p++;
                            NEXT;
                        }
                        else {
                            sukuna_error_method_not_found((sukunaClass*)Scratch,selid);
                            NEXTINST(_throw);
                        }
                    }
                CASE(_invoke_virtual) 
                    push_if_primitive_obj2;
                    r_push(cp_to_sp(p));
                    p = sp_to_cp(*((CELL*)sp_to_cp(*((CELL*)((CELL)((CELL*)sp_to_cp(*(to_class(((CELL*)(*stackp)))))) + TOS)))));
                    drop;
                    token = *p++;
                    NEXT;
                CASE(_invoke_virtuali) 
                    push_if_primitive_obj;
                    Scratch = *p++;
                    r_push(cp_to_sp(p));
                    p = sp_to_cp(*((CELL*)sp_to_cp(*((CELL*)((CELL)((CELL*)sp_to_cp(*(to_class(((CELL*)(TOS)))))) + Scratch)))));
                    token = *p++;
                    NEXT;
                CASE(_invoke_mixed_virtual)  {
                        CELL mofs;
                        CELL vofs;
                        CELL obp;
                        CELL* clp;
                        CELL vf;
                        mofs = pop;
                        vofs = pop;
                        push_if_primitive_obj2;
                        obp  = pop;
                        clp = ((CELL*)to_cclass(obp));
                        vf = *((CELL*)sp_to_cp(*clp + mofs));
                        if(vf==TOS) {
                            obp = *((CELL*)sp_to_cp(obp + vofs));
                        }
                        TOS = obp;
                        r_push(cp_to_sp(p)); 
                        p = sp_to_cp(*(sp_to_cp(vf)));
                        token = *p++;
                        NEXT;
                    }
                CASE(_invoke_mixed_virtuali)  {
                        CELL xtp  = *p++;
                        CELL mofs = *p++;
                        CELL vofs = *p++;
                        push_if_primitive_obj;
                        Scratch = *((CELL*)sp_to_cp(*to_cclass(TOS) + mofs));
                        if(Scratch==xtp) {
                            TOS = *((CELL*)sp_to_cp(TOS + vofs));
                        }
                        r_push(cp_to_sp(p)); 
                        p = sp_to_cp(*(sp_to_cp(Scratch)));
                        token = *p++;
                        NEXT;
                    }
                CASE(_invoke_mixin)  {
                        CELL mofs;
                        CELL mxp;
                        CELL obp;
                        CELL* clp;
                        CELL vofs;
                        CELL vf;
                        CELL offset;
                        CELL* ofsp;
                        
                        mofs = pop;
                        mxp  = pop;
                        push_if_primitive_obj2;
                        obp  = pop;
                        clp = ((CELL*)to_cclass(obp));
                        ofsp = sukuna_find_mixin_offset((sukunaClass*)mxp, (sukunaClass*)(clp-1));
                        if(ofsp!=SNIL) {
                            offset = *ofsp;
                            vofs = offset >> 16;
                            mofs += (offset & offset_mask);
                            vf = *((CELL*)sp_to_cp(*clp + mofs));
                            if(vf==TOS) {
                                obp = *((CELL*)sp_to_cp(obp + vofs));
                            }
                            TOS = obp;
                            r_push(cp_to_sp(p)); 
                            p = sp_to_cp(*(sp_to_cp(vf)));
                            token = *p++;
                            NEXT;
                        }
                        else {
                            sukuna_error_mixin_not_found((sukunaClass*)(clp-1),mxp);
                            NEXTINST(_throw);
                        }
                    }
                CASE(_invoke_mixini)  {
                        CELL xtp  = *p++;
                        CELL mofs = *p++;
                        CELL mxp  = *p++;
                        CELL* clp;
                        CELL offset;
                        CELL* ofsp;

                        push_if_primitive_obj;
                        clp = to_cclass(TOS);
                        ofsp = sukuna_find_mixin_offset((sukunaClass*)mxp, (sukunaClass*)(clp-1));
                        if(ofsp!=SNIL) {
                            offset = *ofsp;
                            mofs += (offset & offset_mask);
                            Scratch = *((CELL*)sp_to_cp(*clp + mofs));
                            if(Scratch==xtp) {
                                TOS = *((CELL*)sp_to_cp(TOS + (offset >> 16)));
                            }
                            r_push(cp_to_sp(p)); 
                            p = sp_to_cp(*(sp_to_cp(Scratch)));
                            token = *p++;
                            NEXT;
                        }
                        else {
                            sukuna_error_mixin_not_found((sukunaClass*)(clp-1),mxp);
                            NEXTINST(_throw);
                        }
                    }
                CASE(_invoke_stless_mixin)  {
                        CELL mxp;
                        CELL obp;
                        CELL* clp;
                        CELL vofs;
                        CELL vf;
                        CELL offset;
                        CELL* ofsp;
                        
                        mxp  = pop;
                        push_if_primitive_obj2;
                        obp  = pop;
                        clp = ((CELL*)to_cclass(obp));
                        ofsp = sukuna_find_mixin_offset((sukunaClass*)mxp, (sukunaClass*)(clp-1));
                        if(ofsp!=SNIL) {
                            offset = *ofsp;
                            vofs = offset >> 16;
                            TOS += (offset & offset_mask);
                            vf = *((CELL*)sp_to_cp(*clp + TOS));
                            TOS = obp;
                            r_push(cp_to_sp(p)); 
                            p = sp_to_cp(*(sp_to_cp(vf)));
                            token = *p++;
                            NEXT;
                        }
                        else {
                            sukuna_error_mixin_not_found((sukunaClass*)(clp-1),mxp);
                            NEXTINST(_throw);
                        }
                    }
                CASE(_invoke_stless_mixini)  {
                        CELL mofs = *p++;
                        CELL mxp  = *p++;
                        CELL* clp;
                        CELL offset;
                        CELL* ofsp;
                        
                        push_if_primitive_obj;
                        clp = to_cclass(TOS);
                        ofsp = sukuna_find_mixin_offset((sukunaClass*)mxp, (sukunaClass*)(clp-1));
                        if(ofsp!=SNIL) {
                            offset = *ofsp;
                            mofs += (offset & offset_mask);
                            Scratch = *((CELL*)sp_to_cp(*clp + mofs));
                            r_push(cp_to_sp(p)); 
                            p = sp_to_cp(*(sp_to_cp(Scratch)));
                            token = *p++;
                            NEXT;
                        }
                        else {
                            sukuna_error_mixin_not_found((sukunaClass*)(clp-1),mxp);
                            NEXTINST(_throw);
                        }
                    }
                CASE(_invoke_force_self)  {
                        Scratch = pop;
                        TOS = *((CELL*)((CELL)objectp+TOS));
                        push_if_primitive_obj;
                        r_push(cp_to_sp(p)); 
                        p = sp_to_cp(*(sp_to_cp(Scratch)));
                        token = *p++;
                        NEXT;
                    }
                CASE(_invoke_force_selfi)  {
                        Scratch = *p++;
                        push(TOS);
                        TOS = *((CELL*)((CELL)objectp+(*p++)));
                        push_if_primitive_obj;
                        r_push(cp_to_sp(p)); 
                        p = sp_to_cp(*(sp_to_cp(Scratch)));
                        token = *p++;
                        NEXT;
                    }
                CASE(_invoke_force_class)  {
                        CELL vofs;
                        CELL obp;
                        vofs = pop;
                        push_if_primitive_obj2;
                        obp  = pop;
                        Scratch = TOS;                        
                        TOS = *((CELL*)((CELL)sp_to_cp(obp)+vofs));
                        r_push(cp_to_sp(p)); 
                        p = sp_to_cp(*(sp_to_cp(Scratch)));
                        token = *p++;
                        NEXT;
                    }                
                CASE(_invoke_force_classi)  {
                        CELL vofs;
                        CELL obp;
                        Scratch = *p++;                        
                        vofs = *p++;
                        push_if_primitive_obj;
                        obp  = TOS;
                        TOS = *((CELL*)((CELL)sp_to_cp(obp)+vofs));
                        r_push(cp_to_sp(p)); 
                        p = sp_to_cp(*(sp_to_cp(Scratch)));
                        token = *p++;
                        NEXT;
                    }                
                CASE(_invoke_force_mixin)  {
                        CELL mxp;
                        CELL obp;
                        CELL* clp;
                        CELL vofs;
                        CELL offset;
                        CELL* ofsp;
                        
                        mxp  = pop;
                        push_if_primitive_obj2;
                        obp  = pop;
                        clp = ((CELL*)to_cclass(obp));
                        ofsp = sukuna_find_mixin_offset((sukunaClass*)mxp, (sukunaClass*)(clp-1));
                        if(ofsp!=SNIL) {
                            offset = *ofsp;
                            vofs = offset >> 16;
                            Scratch = TOS;
                            TOS = *((CELL*)sp_to_cp(obp + vofs));
                            r_push(cp_to_sp(p)); 
                            p = sp_to_cp(*(sp_to_cp(Scratch)));
                            token = *p++;
                            NEXT;
                        }
                        else {
                            sukuna_error_mixin_not_found((sukunaClass*)(clp-1),mxp);
                            NEXTINST(_throw);
                        }
                    }
                CASE(_invoke_force_mixini)  {
                        CELL mxp;
                        CELL obp;
                        CELL* clp;
                        CELL vofs;
                        CELL offset;
                        CELL* ofsp;
                        
                        Scratch = *p++;
                        mxp  = *p++;
                        push_if_primitive_obj;
                        obp  = TOS;
                        clp = ((CELL*)to_cclass(obp));
                        ofsp = sukuna_find_mixin_offset((sukunaClass*)mxp, (sukunaClass*)(clp-1));
                        if(ofsp!=SNIL) {
                            offset = *ofsp;
                            vofs = offset >> 16;
                            TOS = *((CELL*)sp_to_cp(obp + vofs));
                            r_push(cp_to_sp(p)); 
                            p = sp_to_cp(*(sp_to_cp(Scratch)));
                            token = *p++;
                            NEXT;
                        }
                        else {
                            sukuna_error_mixin_not_found((sukunaClass*)(clp-1),mxp);
                            NEXTINST(_throw);
                        }
                    }
                CASE(_invoke_mixin_super)  {
                        CELL* clp;
                        CELL vofs;
                        clp = to_cclass(cp_to_sp(objectp));
                        TOS = (CELL)sukuna_find_mixin_super((sukunaClass*)TOS,(sukunaClass*)(clp-1),(CELL)cp_to_sp(objectp),(CELL)get_selfp, pop, &vofs);
                        if (TOS == SNIL) {
                            sukuna_error_mixin_not_found((sukunaClass*)(clp-1),TOS);
                            NEXTINST(_throw);
                        }
                        else if(vofs==to_si(0)) {
                            TOS = *((CELL*)sp_to_cp(TOS));
                            NEXTINST(_invoke_static);
                        }
                        else {
                            push(vofs);
                            TOS = *((CELL*)sp_to_cp(TOS));
                            NEXTINST(_invoke_force_class);
                        }
                    }
                CASE(_invoke_mixin_superi)  {
                        CELL* clp;
                        CELL vofs;
                        clp = to_cclass(cp_to_sp(objectp));
                        Scratch = *p++;
                        push(TOS);
                        TOS = (CELL)sukuna_find_mixin_super((sukunaClass*)Scratch,(sukunaClass*)(clp-1),(CELL)cp_to_sp(objectp),(CELL)get_selfp, (*p++), &vofs);
                        if (TOS == SNIL) {
                            sukuna_error_mixin_not_found((sukunaClass*)(clp-1),Scratch);
                            NEXTINST(_throw);
                        }
                        else if(vofs==to_si(0)) {
                            TOS = *((CELL*)sp_to_cp(TOS));
                            NEXTINST(_invoke_static);
                        }
                        else {
                            push(vofs);
                            TOS = *((CELL*)sp_to_cp(TOS));
                            NEXTINST(_invoke_force_class);
                        }
                    }
                CASE(_invoke_static)  {
                        Scratch = TOS;
                        drop;
                        push_if_primitive_obj;
                        r_push(cp_to_sp(p)); 
                        p = sp_to_cp(*(sp_to_cp(Scratch)));
                        token = *p++;
                        NEXT;
                    }
                CASE(_invoke_statici)  {
                        Scratch = *p++;
                        push_if_primitive_obj;
                        r_push(cp_to_sp(p)); 
                        p = sp_to_cp(*(sp_to_cp(Scratch)));
                        token = *p++;
                        NEXT;
                    }

                CASE(_lit_ivar) 
                    push(TOS); TOS = *((CELL*)((CELL)objectp + (CELL)(*p++)));
                    token = *p++; NEXT;
                CASE(_to_ivar) 
                    *((CELL*)((CELL)objectp + (CELL)(*p++))) = TOS;
                    drop;
                    token = *p++; NEXT;
                CASE(_lit_mixed_ivar) 
                    push(TOS);
                    TOS = *((CELL*)((CELL)objectp+*p++));
                    TOS = *((CELL*)((CELL)sp_to_cp(TOS)+*p++));
                    token = *p++; NEXT;
                CASE(_to_mixed_ivar) 
                    Scratch = *((CELL*)((CELL)objectp+*p++));
                    *((CELL*)((CELL)sp_to_cp(Scratch)+*p++)) = TOS;
                    drop;
                    token = *p++; NEXT;
                CASE(_to_p_ivar) 
                    Scratch = ((CELL)objectp+*p++);
                    plusstore((CELL*)Scratch);
                    NEXT;
                CASE(_to_p_mixed_ivar) 
                    Scratch = *((CELL*)((CELL)objectp+*p++));
                    Scratch = ((CELL)sp_to_cp(Scratch)+*p++);
                    plusstore((CELL*)Scratch);
                    NEXT;
                CASE(_to_1p_ivar) 
                    Scratch = ((CELL)objectp+*p++);
                    if(check_si(*((CELL*)Scratch))) {
                        *((CELL*)Scratch) = to_si(to_ci(*((CELL*)Scratch)) + 1);
                        token = *p++; NEXT;
                    } 
                    else {
                        *((CELL*)Scratch) += ptrd(*((CELL*)Scratch));
                        token = *p++; NEXT;
                    }
                CASE(_to_1m_ivar) 
                    Scratch = ((CELL)objectp+*p++);
                    if(check_si(*((CELL*)Scratch))) {
                        *((CELL*)Scratch) = to_si(to_ci(*((CELL*)Scratch)) - 1);
                        token = *p++; NEXT;
                    } 
                    else {
                        *((CELL*)Scratch) -= ptrd(*((CELL*)Scratch));
                        token = *p++; NEXT;
                    }
                CASE(_to_1p_mixed_ivar) 
                    Scratch = *((CELL*)((CELL)objectp+*p++));
                    Scratch = ((CELL)sp_to_cp(Scratch)+*p++);
                    if(check_si(*((CELL*)Scratch))) {
                        *((CELL*)Scratch) = to_si(to_ci(*((CELL*)Scratch)) + 1);
                        token = *p++; NEXT;
                    } 
                    else {
                        *((CELL*)Scratch) += ptrd(*((CELL*)Scratch));
                        token = *p++; NEXT;
                    }
                CASE(_to_1m_mixed_ivar) 
                    Scratch = *((CELL*)((CELL)objectp+*p++));
                    Scratch = ((CELL)sp_to_cp(Scratch)+*p++);
                    if(check_si(*((CELL*)Scratch))) {
                        *((CELL*)Scratch) = to_si(to_ci(*((CELL*)Scratch)) - 1);
                        token = *p++; NEXT;
                    } 
                    else {
                        *((CELL*)Scratch) -= ptrd(*((CELL*)Scratch));
                        token = *p++; NEXT;
                    }

                CASE(_f_dot) 
                    sukuna_float_dot((sukunaFloat*)sp_to_cp(TOS-cellWidth)); drop;
                    token = *p++; NEXT;
                CASE(_f_plus)   fbinop(+); token = *p++; NEXT;
                CASE(_f_minus)  fbinop(-); token = *p++; NEXT;
                CASE(_f_times)  fbinop(*); token = *p++; NEXT;
                CASE(_f_divide) fbinop(/); token = *p++; NEXT;
                CASE(_f_mod)  
                    Scratch = pop; ffunc(fmod(to_cf(Scratch),to_cf(TOS))); token = *p++; NEXT;
                CASE(_f_sin)    ffunc(sin(to_cf(TOS))); token = *p++; NEXT;
                CASE(_f_cos)    ffunc(cos(to_cf(TOS))); token = *p++; NEXT;
                CASE(_f_tan)    ffunc(tan(to_cf(TOS))); token = *p++; NEXT;
                CASE(_f_asin)   ffunc(asin(to_cf(TOS))); token = *p++; NEXT;
                CASE(_f_acos)   ffunc(acos(to_cf(TOS))); token = *p++; NEXT;
                CASE(_f_atan)   ffunc(atan(to_cf(TOS))); token = *p++; NEXT;
                CASE(_f_sinh)   ffunc(sinh(to_cf(TOS))); token = *p++; NEXT;
                CASE(_f_cosh)   ffunc(cosh(to_cf(TOS))); token = *p++; NEXT;
                CASE(_f_tanh)   ffunc(cosh(to_cf(TOS))); token = *p++; NEXT;
                CASE(_f_atan2)  
                    Scratch = pop; ffunc(atan2(to_cf(Scratch),to_cf(TOS))); token = *p++; NEXT;
                CASE(_f_exp)    ffunc(exp(to_cf(TOS))); token = *p++; NEXT;
                CASE(_f_ln)     ffunc(log(to_cf(TOS))); token = *p++; NEXT;
                CASE(_f_log)    ffunc(log10(to_cf(TOS))); token = *p++; NEXT;
                CASE(_f_pow)  
                    Scratch = pop; ffunc(pow(to_cf(Scratch),to_cf(TOS))); token = *p++; NEXT;
                CASE(_f_sqrt)   ffunc(sqrt(to_cf(TOS))); token = *p++; NEXT;
                CASE(_f_ceil)   ffunc(ceil(to_cf(TOS))); token = *p++; NEXT;
                CASE(_f_floor)  ffunc(floor(to_cf(TOS))); token = *p++; NEXT;
                CASE(_f_abs)    ffunc(fabs(to_cf(TOS))); token = *p++; NEXT; 
                CASE(_f_lteqgt)  
                        Scratch = pop;
                        TOS = (to_cf(Scratch)<to_cf(TOS)) ? to_si(-1) : ((to_cf(Scratch)>to_cf(TOS)) ? to_si(1) : to_si(0));
                        token = *p++; NEXT;
                CASE(_f_ameq)  {
                            double e = to_cf(TOS);
                            double v = to_cf(pop);
                            double u = to_cf(pop);
                            TOS = (fabs(u-v)/fabs(u)<=e)&&(fabs(u-v)/fabs(v)<=e) ? STRUE : SFALSE;
                            token = *p++; NEXT;
                        }
                CASE(_f_apeq)  {
                            double e = to_cf(TOS);
                            double v = to_cf(pop);
                            double u = to_cf(pop);
                            TOS = (fabs(u-v)/fabs(u)<=e)||(fabs(u-v)/fabs(v)<=e) ? STRUE : SFALSE;
                            token = *p++; NEXT;
                        }
                CASE(_f_negate) ffunc(-(to_cf(TOS))); token = *p++; NEXT;
                CASE(_f_to_i)   TOS = to_si((CELL)to_cf(TOS)); token = *p++; NEXT;
                CASE(_i_to_f)   ffunc(((double)to_ci(TOS))); token = *p++; NEXT;
                CASE(_f_to_sstr)  {
                        UWORD* fma = (UWORD*)sp_to_cp(pop);
                        CELL n = to_ci(pop);
                        Scratch = pop;
                        TOS = to_si(sukuna_float_to_str((UWORD*)sp_to_cp(Scratch),n,fma,to_ci(TOS),to_cf(pop)));
                        push(Scratch);
                        token = *p++;
                        NEXT;
                    }
                CASE(_sstr_to_f)
                    save_registers;
                    Scratch = (CELL)sukuna_new_float_obj((double)0.0);
                    load_registers;
                    TOS = (sukuna_str_to_float((UWORD*)sp_to_cp(pop),to_ci(TOS),(double*)(Scratch+cellWidth))!=SNIL ? (CELL)cp_to_sp(Scratch+cellWidth) : SNIL);
                    token = *p++;
                    NEXT;

                CASE(_alloc)  
                    save_registers;
                    Scratch = (CELL)cp_to_sp(sukuna_malloc(to_ci(TOS)));
                    load_registers;
                    TOS = Scratch;
                    token = *p++; NEXT;
                CASE(_free)  sukuna_free(sp_to_cp(TOS)); drop; token = *p++; NEXT;

                CASE(_eval)  save_pc; sukuna_evaluate((WORD*)sp_to_cp(pop),to_ci(TOS)); drop;
                    NEXTINST(_halt);
                CASE(_evalcont)  load_pc; token = *p++; NEXT;
                
                CASE(_poptostype) 
                    push(TOS); 
                    if(*retval_stack>0) {
                        TOS = rv_pop; (*retval_stack)--;
                        token = *p++; NEXT;
                    }
                    else {
                        TOS = SNIL;
                        token = *p++; NEXT;
                    }
                CASE(_pushtostype)  rv_push(TOS); (*retval_stack)++; drop;  token = *p++; NEXT;
                CASE(_popargtype) 
                    push(TOS); 
                    if(*argval_stack>0) {
                        TOS = av_pop; (*argval_stack)--;
                        token = *p++; NEXT;
                    }
                    else {
                        TOS = SNIL;
                        token = *p++; NEXT;
                    }
                    
                CASE(_pushargtype)  av_push(TOS); (*argval_stack)++; drop; token = *p++; NEXT;

                CASE(_cfunci)
                    save_registers;
                    Scratch = ((CFunc)gCustomFunctions[to_ci(*p++)])();
                    load_registers;
                    if(Scratch!=SNIL) {
                        token = Scratch;
                        NEXT;
                    }
                    else {
                        token = *p++; 
                        NEXT;
                    }

                CASE(_bye) byeFlag = STRUE; NEXTINST(_halt);               

            }
        ENDDISPATCH
    }
EXIT_VM:
    save_registers;
}

/***********************************************************************************
* Initialize
***********************************************************************************/
void sukuna_init_code_area() {
    *(codep++) = _halt;
}

void sukuna_build_dictionary() {

    sukuna_create_dic_entryC(_noop,"autoexec",normal_flag);
    
    #define DEFINST(inst,wname,flags)  sukuna_create_dic_entryC(inst,wname,flags);
    #include "primitives.h"
    #undef DEFINST

    #define DEFCFUNC(inst,wname,flags,fname)  sukuna_create_cfunc_entryC(inst,wname,flags);
    #include "cfuncs.h"
    #undef DEFCFUNC
    
    sukuna_create_new_labelC("state",&compile_state);
    sukuna_create_new_labelC("has_local",&has_local);
    sukuna_create_new_labelC("gc_limit",&gc_limit);
    sukuna_create_new_labelC("gc_limit_max",&gc_limit_max);
    sukuna_create_new_labelC("gc_debug",&gc_debug);
    sukuna_create_new_labelC("sukuna_version",&(sVariables->version.Body));
    sukuna_create_new_labelC("classdef?",&class_def);
    sukuna_create_new_labelC("methoddef?",&method_def);
    sukuna_create_new_labelC("echo",&echo_flag);
    sukuna_create_new_labelC("^dicentryc",&(sVariables->dicEntryC.Body));
    sukuna_create_new_labelC("^mthentryc",&(sVariables->mthEntryC.Body));
    sukuna_create_new_labelC("^ivrentryc",&(sVariables->ivrEntryC.Body));
    sukuna_create_new_labelC("^locentryc",&(sVariables->locEntryC.Body));
    sukuna_create_new_labelC("^methodclass",&(sVariables->methodClass.Body));
    sukuna_create_new_labelC("^blockclass",&(sVariables->blockClass.Body));
    sukuna_create_new_labelC("^stringclass",&(sVariables->stringClass.Body));
    sukuna_create_new_labelC("^vectorclass",&(sVariables->vectorClass.Body));
    sukuna_create_new_labelC("^floatclass",&(sVariables->floatClass.Body));
    sukuna_create_new_labelC("^systemprim",&(sVariables->systemPrim.Body));
    sukuna_create_new_labelC("^systemint",&(sVariables->systemInt.Body));
    sukuna_create_new_labelC("^systemnil",&(sVariables->systemNil.Body));
    sukuna_create_new_labelC("^systemchar",&(sVariables->systemChar.Body));
    sukuna_create_new_labelC("^systemsymb",&(sVariables->systemSymb.Body));
    sukuna_create_new_labelC("size_of_double",&(sVariables->sizeOfDouble.Body));
    sukuna_create_new_labelC("local_env",&(sVariables->localEnv.Body));
    sukuna_create_new_labelC("quote_value",&(sVariables->quoteValue.Body));
    sukuna_create_new_labelC("^argc",&(sVariables->argc.Body));
    sukuna_create_new_labelC("^argv",&(sVariables->argv.Body));
    sukuna_create_new_labelC("^stdin",&(sVariables->sstdin.Body));
    sukuna_create_new_labelC("^stdout",&(sVariables->sstdout.Body));
    sukuna_create_new_labelC("^stderr",&(sVariables->sstderr.Body));
    sukuna_create_new_labelC("^warning_level",&(sVariables->warnlevel.Body));
    sukuna_create_new_labelC("^opt_level",&(sVariables->optlevel.Body));
    sukuna_create_new_labelC("type_inference",&(sVariables->typeinference.Body));
    sukuna_create_new_labelC("^default_var_type",&(sVariables->typeinference.Body));
    sukuna_create_new_labelC("^exact_self",&(sVariables->exactself.Body));

}

void sukuna_init_data_area() {
    sVariables = (sukunaVariables*)sukuna_dallot(bytes_to_cells(sizeof(sukunaVariables)));
    sVariables->state.Klass = SNIL;
    sVariables->state.Body = SFALSE;
    sVariables->hasLocal.Klass = SNIL;
    sVariables->hasLocal.Body = SFALSE;
    sVariables->gcLimit.Klass = SNIL;
    sVariables->gcLimit.Body = to_si(gGcLimitSize);
    sVariables->gcLimitMax.Klass = SNIL;
    sVariables->gcLimitMax.Body = to_si(gGcLimitMaxSize);
    sVariables->gcDebug.Klass = SNIL;
    sVariables->gcDebug.Body = SFALSE;
    sVariables->version.Klass = SNIL;
    sVariables->version.Body = to_si(sukuna_ver_hex);
    sVariables->classDef.Klass = SNIL;
    sVariables->classDef.Body = SNIL;
    sVariables->methodDef.Klass = SNIL;
    sVariables->methodDef.Body = SNIL;
    sVariables->echoflag.Klass = SNIL;
    sVariables->echoflag.Body = SFALSE;
    
    sVariables->dicEntryC.Klass = SNIL;
    sVariables->dicEntryC.Body = SNIL;
    sVariables->mthEntryC.Klass = SNIL;
    sVariables->mthEntryC.Body = SNIL;
    sVariables->ivrEntryC.Klass = SNIL;
    sVariables->ivrEntryC.Body = SNIL;
    sVariables->locEntryC.Klass = SNIL;
    sVariables->locEntryC.Body = SNIL;

    sVariables->methodClass.Klass = SNIL;
    sVariables->methodClass.Body = SNIL;
    sVariables->blockClass.Klass = SNIL;
    sVariables->blockClass.Body = SNIL;
    sVariables->stringClass.Klass = SNIL;
    sVariables->stringClass.Body = SNIL;
    sVariables->vectorClass.Klass = SNIL;
    sVariables->vectorClass.Body = SNIL;
    sVariables->floatClass.Klass = SNIL;
    sVariables->floatClass.Body = SNIL;
    sVariables->systemPrim.Klass = SNIL;
    sVariables->systemPrim.Body = SNIL;
    sVariables->systemInt.Klass = SNIL;
    sVariables->systemInt.Body = SNIL;
    sVariables->systemNil.Klass = SNIL;
    sVariables->systemNil.Body = SNIL;
    sVariables->systemChar.Klass = SNIL;
    sVariables->systemChar.Body = SNIL;
    sVariables->systemSymb.Klass = SNIL;
    sVariables->systemSymb.Body = SNIL;
    sVariables->sizeOfDouble.Klass = SNIL;
    sVariables->sizeOfDouble.Body = to_si(bytes_to_cells(sizeof(double)));
    sVariables->localEnv.Klass = SNIL;
    sVariables->localEnv.Body = SNIL;
    sVariables->quoteValue.Klass = SNIL;
    sVariables->quoteValue.Body = SFALSE;
    
    sVariables->argc.Klass = SNIL;
    sVariables->argc.Body = SNIL;
    sVariables->argv.Klass = SNIL;
    sVariables->argv.Body = SNIL;

    sVariables->sstdin.Klass = SNIL;
    sVariables->sstdin.Body = (CELL)cp_to_sp(stdin);
    sVariables->sstdout.Klass = SNIL;
    sVariables->sstdout.Body = (CELL)cp_to_sp(stdout);
    sVariables->sstderr.Klass = SNIL;
    sVariables->sstderr.Body = (CELL)cp_to_sp(stderr);

    sVariables->warnlevel.Klass = SNIL;
    sVariables->warnlevel.Body = to_si(1);

    sVariables->optlevel.Klass = SNIL;
    sVariables->optlevel.Body = to_si(1);

    sVariables->typeinference.Klass = SNIL;
    sVariables->typeinference.Body = STRUE;

    sVariables->defaultvartype.Klass = SNIL;
    sVariables->defaultvartype.Body = sukuna_make_default_vtype();

    sVariables->exactself.Klass = SNIL;
    sVariables->exactself.Body = SNIL;
    
    currentTask = (sukunaTask*)sukuna_dallot(bytes_to_cells(sizeof(sukunaTask)));
    topTask = currentTask;
    currentTask->ListLink = SNIL;
    currentTask->sp = &memory[stackTop];
    currentTask->rp = &memory[rStackTop];
    currentTask->op = SNIL;
    currentTask->pc = &memory[0];
    
    currentTask->IN_orig = (WORD*)sukuna_tallot(words_to_cells(buf_size));
    currentTask->IN = currentTask->IN_orig;
    currentTask->cnt = 0;
    currentTask->stream = stdin;
    currentTask->flag = 0;
    currentTask->linenum = 0;
    currentTask->filepath = SNIL;

    sdict = sukuna_create_dictionary(d_area);
    ldict = SNIL;
    sukuna_build_dictionary();
}

void sukuna_init_globals() {
    sukunap = (sukunaGlobals*)&memory[dataTop];
    sukunap->cp = &memory[memTop];
    sukunap->dp = &memory[dataTop]+bytes_to_cells(sizeof(sukunaGlobals));
    sukunap->tp = &memory[tempTop];
    sukunap->bye = SNIL;
    
}

CELL sukuna_init_memory() {
    memSize = codeSize+dataSize; 
    stackTop = memSize+stackSize-1;
    rStackTop = memSize+stackSize+rStackSize-1;
    tempTop = memSize+stackSize+rStackSize;
    memMax = memSize+stackSize+rStackSize+tempSize;
    if((memory = (CELL*)malloc(memMax*cellWidth))==NULL) {
        sukuna_cputs("sukuna: initialize error.\n");
        return SNIL;
    }
    memOffset = (CELL)&memory[0];
    lastHeap = SNIL;
    totalHeapSize = 0;    
    return STRUE;
}

void sukuna_init() {
    sukuna_init_globals();
    sukuna_init_code_area();
    sukuna_init_data_area();
    sukuna_reset_stack();
}

/***********************************************************************************
* Interpreter
***********************************************************************************/
void sukuna_interpret(unsigned char* commandLine) {
    WORD* addr;
    WORD word[buf_size];
    WORD cmd[buf_size];
    CELL n;
    CELL xt;
    CELL count;
    int  i;
    sukunaDicEntry* lfa;
    sukunaDicEntry* latest;
    CELL num;
    double fnum;
    
    if(commandLine!=SNIL) {
        currentTask->flag = STRUE;
        currentTask->IN_orig = cmd;
        currentTask->IN = currentTask->IN_orig;
        currentTask->cnt = cstr_to_sstr(commandLine,(UWORD*)currentTask->IN,buf_size);
    }
    do {
        if(currentTask->ListLink==SNIL) {
            sukuna_msg(">");
        }
        if(currentTask->flag == 0) {
            if(sukuna_fgets(currentTask->IN_orig,buf_size,currentTask->stream)==SNIL) {
                if(currentTask->ListLink!=SNIL) {
                    fclose(currentTask->stream);
                    sukuna_pop_current_stream();
                }
                else {
                    break;
                }
            }
            else {
                currentTask->IN = currentTask->IN_orig;
                currentTask->cnt = to_addr_n(currentTask->IN);
                currentTask->linenum++;
            }
        }
        else {
            if(currentTask->ListLink!=SNIL) {
                sukuna_pop_current_stream();
                sukuna_VM(_evalcont);
            }
            else {
                if(currentTask->cnt<=0) {
                    byeFlag = STRUE;
                }
            }
        }

        if(currentTask->ListLink!=SNIL&&(!(is_false(echo_flag)))) {
            sukuna_type(currentTask->IN,currentTask->cnt);
        }

        while((n = sukuna_next_word(&(currentTask->IN),&(currentTask->cnt),&addr))>0 && byeFlag==SNIL) {
            currentTask->current_word = addr;
            currentTask->current_cnt  = n;

            sukuna_lower(addr,n,word);
            
            count = 0;
            if((lfa=sukuna_findQ(word,n,ldict,&count))!=SNIL) {
                /* Found local variables. */
                if(compile_state==STRUE) {
                    /* compile */
                    if(count==0) {
                        sukuna_compile_localfetch(lfa);
                        if(lfa->Data==func_type) {
                            *(codep++) = _call;
                            sukuna_type_inference_check(lfa);
                        }
                        else if(check_sp(lfa->Data)) {
                            sukuna_compile_localref(lfa);
                        }
                        else {
                            sukuna_literal_type_inference(lfa->Data);
                        }
                    }
                    else {
                        sukuna_compile_localfetch(lfa);
                        for(i=0;i<count;++i) { *(codep++) = _call; }
                        if(count==1) { 
                            sukuna_type_inference_check(lfa); 
                        }
                        else {
                            sukuna_failure_type_check();
                        }
                    }
                }
            }
            else if((lfa=sukuna_find_ivari(word,n,&count))!=SNIL) {
                if(count!=0) { sVariables->quoteValue.Body = STRUE; }
                sukuna_compile_ivarandref(lfa);
                if(count!=0) {
                    sVariables->quoteValue.Body = SFALSE;
                    if(compile_state!=STRUE || ((lfa->Flags&innerfunc_flag!=0)&&((lfa->Flags&immediate_mask)!=0))) {
                        for(i=0;i<count-1;++i) { sukuna_VM(_call); }
                    }
                    else {
                        for(i=0;i<count-1;++i) { *(codep++) = _call; }
                    }
                }
            }
            else if((lfa=sukuna_findQ(word,n,sdict,&count))!=SNIL) {
                /* Found */
                xt = lfa->Xt;
                if(!check_si(xt)) xt = (CELL)cp_to_sp(&(lfa->Xt));
                if((lfa->Flags&immediate_mask)!=0) {
                    /* it's IMMEDIATE ! */
                    if(compile_state==STRUE) {
                        if(count!=0) { sVariables->quoteValue.Body = STRUE; }
                        sukuna_VM(xt);
                        if(count!=0) { 
                            sVariables->quoteValue.Body = SFALSE; 
                            for(i=0;i<count;++i) { sukuna_VM(_call); }
                        }
                        sukuna_type_inference_check(lfa);
                    }
                    else {
                        if(count!=0) { sVariables->quoteValue.Body = STRUE; }
                        sukuna_VM(xt);
                        if(count!=0) { 
                            sVariables->quoteValue.Body = SFALSE;
                            for(i=0;i<count;++i) { sukuna_VM(_call); }
                        }
                    }
                }
                else {
                    if(compile_state==SFALSE) {
                    /* interpret */
                        sukuna_VM(xt);
                        for(i=0;i<count;++i) { sukuna_VM(_call); }
                    }
                    else {
                    /* compile */
                        *(codep++) = xt;
                        latest = (sukunaDicEntry*)sp_to_cp(sdict->LastEntry);
                        if((((CELL)cp_to_sp(&(latest->Xt))) == xt) && (((latest->Flags)&ignore_mask) == 0)) {
                        /* recursive call */
                            *(codep++) = _noop;
                            *(codep++) = _noop;
                        }
                        if(count==0) {
                            sukuna_type_inference_check(lfa);
                        }
                        else {
                            for(i=0;i<count;++i) { *(codep++) = _call; }
                            sukuna_failure_type_check();
                        }
                    }
                }
            }
            else {
                if(sukuna_str_to_num((UWORD*)word,n,&num)==STRUE) {
                    /* number */
                    if(compile_state==SFALSE) {
                        /* interpret */
                        push_data(num);
                    }
                    else {
                        /* compile */
                        *(codep++) = _lit;
                        *(codep++) = num;
                        sukuna_literal_type_inference(sukuna_check_value_type(num));
                    }
                }
                else if(sukuna_str_to_float((UWORD*)word,n,&fnum)==STRUE) {
                    /* float number */
                    if(compile_state==SFALSE) {
                        /* interpret */
                        push_data((CELL)cp_to_sp(&((sukuna_new_float_obj(fnum))->Body)));
                    }
                    else {
                        /* compile */
                        *(codep++) = _lit;
                        *(codep++) = (CELL)cp_to_sp(&((sukuna_static_float_obj(fnum))->Body));                    
                        if(sVariables->floatClass.Body!=SNIL) {
                            sukuna_literal_type_inference(sVariables->floatClass.Body);
                        }
                        else {
                            sukuna_literal_type_inference(float_type);
                        }
                   }
                }
                else if(sukuna_numref_to_char((UWORD*)word,n,&num,(UWORD*)addr)==STRUE) {
                    /* char */
                    if(compile_state==SFALSE) {
                        /* interpret */
                        push_data(num);
                    }
                    else {
                        /* compile */
                        *(codep++) = _lit;
                        *(codep++) = num;                    
                        sukuna_literal_type_inference(char_type);
                    }
                }
                else {
                    /* unrecognized word */
                    sukuna_put_error_info();
                    sukuna_type(word,n);
                    sukuna_cputs(" ? - unrecognized word.\n");
                    
                    sukuna_reset_task();
                    break;
                }
            }
        }
        if(currentTask->ListLink==SNIL) {
            sukuna_msg(" ok\n");
        }
    } while(byeFlag==SNIL);
}

